/*
 * Copyright (C) 2003 Jeppe Nejsum Madsen, nejsum@diku.dk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include "stdafx.h"
#include "boost/bind.hpp"
#include "csplib/variable.hpp"
#include "csplib/relation.hpp"
#include "csp.hpp"

namespace
{
  // Various relation orderings

  struct SizeOrder
  {
    bool operator()(const Relation *r1, const Relation *r2)
    {
      return r1->getSize() < r2->getSize();
    }
  };

  struct TupleCountOrder
  {
    bool operator()(const Relation *r1, const Relation *r2)
    {
      return r1->getNumRows() < r2->getNumRows();
    }
  };

  struct ScopeSizeOrder
  {
    bool operator()(const Relation *r1, const Relation *r2)
    {
      if (r1->getNumColumns() == r2->getNumColumns() && 
          std::equal(r1->getColumns().begin(),  r1->getColumns().end(),
                     r2->getColumns().begin(), ColumnPtrVariableEquality()))
        {
        return r1->getSize() < r2->getSize();
        }
      
      return std::lexicographical_compare(r1->getColumns().begin(),
                                          r1->getColumns().end(),
                                          r2->getColumns().begin(),
                                          r2->getColumns().end(),
                                          ColumnPtrVariableOrder());
    }
  };
}
float JoinSizeCost::operator()(const Relation *r1, const Relation *r2, float lowest_cost) const
  {
  size_t oldsize = lowest_cost == std::numeric_limits<float>::infinity() ? 
    std::numeric_limits<size_t>::max() : static_cast<size_t>(lowest_cost);
  size_t newsize = r1->joinSize(*r2, oldsize);
  return (float)newsize;
  }

float SelectivityCost::operator()(const Relation *r1, const Relation *r2, float) const
  {
  float res;

  res = float(r1->getNumRows()) * float(r2->getNumRows());

  float selectivity;
  size_t common = r1->getNumCommonColumns(*r2);
  selectivity = 1.0f/(common + 1.0f);
  return res*selectivity;
  }

float ADBCost::operator()(const Relation *r1, const Relation *r2, float) const
  {
  size_t commonColumns = r1->getNumCommonColumns(*r2);

  if (commonColumns == 0)
    return std::numeric_limits<float>::infinity();

  size_t  nfree1 = r1->getNumColumns() - commonColumns;
  size_t  nfree2 = r2->getNumColumns() - commonColumns;

  if (nfree1 == 0 || nfree2 == 0)
    return 0.0;

  return (float)nfree1*(float)nfree2*(float)r1->getNumRows()*(float)r2->getNumRows()/(float)commonColumns;
  }

 
float GLMCostFunction::operator()(const Relation *r1, const Relation *r2, float) const
  {
  size_t commonColumns = r1->getNumCommonColumns(*r2);
  
  if (commonColumns == 0)
    return std::numeric_limits<float>::infinity();
  
  size_t  nfree1 = r1->getNumColumns() - commonColumns;
  size_t  nfree2 = r2->getNumColumns() - commonColumns;
  
  if (nfree1 == 0 || nfree2 == 0)
    return 0.0;
  
  return (float)nfree1*(float)nfree2*(float)r1->getNumRows()*(float)r2->getNumRows();
  }

void CSP::joinEqualScope(size_t maxsize)
  {
  std::cout << "Joining relations with same scope. #rels=" << m_Rels.size() << "\n";

  m_Rels.sort(ScopeSizeOrder());

  relcol_t::iterator it = m_Rels.begin();
  RelationPtrScopeEquality  equal_scope;

  while(it != m_Rels.end())
    {
    relcol_t::iterator itcur = it;
    itcur++;

    bool compress = false, joined = false;
    size_t curSize = (*it)->getNumRows();

    int failures = 0;
    while(itcur != m_Rels.end() && equal_scope(*it, *itcur)) 
      {
      try
        {
        joinRels(it, itcur, maxsize);
        }
      catch(std::length_error &)
        {
        if (failures++ < 2)
          {
          ++itcur;
          continue;
          }
        else
          break;
        }

      joined = true;
      m_Rels.erase(itcur);
      std::cout << " #rels=" << m_Rels.size() << "\n";

      Relation *newrel = *it;
      if (newrel->getNumRows() > curSize*3)
        {
        compress = false;
        newrel->compress();
        curSize = newrel->getNumRows();
        }
      else
        compress = true;
      itcur = it;
      ++itcur;
      }
    itcur = it++;

    if(compress)
      (*itcur)->compress();
    }
  }

void CSP::join0(size_t maxsize)
  {
  std::cout << "Joining relations with factor 0. #rels=" << m_Rels.size() << "\n";  

  RelationPtrScopeSubset  is_subset;

  m_Rels.sort(SizeOrder());

  bool found = true;
  while(found)
    {
    relcol_t::iterator it = m_Rels.begin();
    size_t    smallest_result = std::numeric_limits<size_t>::max();
    relcol_t::iterator r1, r2;

    found = false;
    while(it != m_Rels.end())
      {
      relcol_t::iterator itcur = it;
      itcur++;
  
      while(itcur != m_Rels.end())
        {
        Relation *r = *it;
        
        Relation *rcur = *itcur;
        if (it != itcur && (is_subset(r, rcur) || is_subset(rcur, r)))
          {
          found = true;
          size_t result = r->joinSize(*rcur, smallest_result);
          if (result < smallest_result)
            {
            smallest_result = result;
            r1 = it;
            r2 = itcur;
            }
          }
        ++itcur;
        }
      ++it;
      }

    if (found)
      {
      std::cout << "  #rels=" << m_Rels.size() << " ";

      try
        {
        joinRels(r1, r2, maxsize);
        (*r1)->compress();
        m_Rels.erase(r2);
        }
      catch(std::length_error &)
        {
        found = false;
        }
      }
    }
  }

void CSP::joinGOO(size_t maxsize)
  {
  std::cout << "Joining relations using GOO ordering. #rels=" << m_Rels.size() << "\n";

  size_t    num_joins = 0;
  relcol_t  large_rels;
  
  while(m_Rels.size() > 1)
    {
    checkInvariants();
    JoinPair p;
    p = greedySelectJoinPair(ADBCost(), m_Rels.begin(), m_Rels.end());

    if (p.JoinCost == std::numeric_limits<float>::infinity())
      {
      std::cout << "*** Remaining relations are not connected ***\n";
      break;
      }
    std::cout << "  #rels=" << m_Rels.size() << ", JoinCost=" << p.JoinCost << " ";

    try
      {
      joinRels(p.first, p.second, maxsize);
      num_joins++;
      }
    catch(std::length_error &)
      {
      if ((*p.first)->getNumRows() > (*p.second)->getNumRows())
        {
        large_rels.push_back(*p.first);
        m_Rels.erase(p.first);
        }
      else
        {
        large_rels.push_back(*p.second);
        m_Rels.erase(p.second);
        }
      continue;
      }
    m_Rels.erase(p.second);
    
    Relation *joined_rel = *p.first;

    relset_t  changed_relations;
    varset_t  bounded_variables;

    if (joined_rel->identifyUnaryConstraints(&bounded_variables))
      node_consistency(*this, bounded_variables, &changed_relations);

    changed_relations.insert(joined_rel);

    // Only do AC after every 10th join
    if (num_joins++ % 10 == 0)
      {
      relset_t ac_changed;
      weak_arc_consistency(*this, changed_relations, &ac_changed);
      changed_relations.insert(ac_changed.begin(), ac_changed.end());
      }

    // Isolate may remove joined_rel
    changed_relations.erase(joined_rel);
    if(isolateVariables(*joined_rel) == false)
      {
      // Not removed, reinsert
      changed_relations.insert(joined_rel);
      }

    do
      {
      relset_t t;
      
      std::for_each(changed_relations.begin(), changed_relations.end(), boost::bind(&Relation::compress, _1, (const varset_t *)0));
      removeTautologyColumns(changed_relations, t, false);
      std::swap(t, changed_relations);
      } while(!changed_relations.empty());
    }
  m_Rels.splice(m_Rels.end(), large_rels);
  }

void CSP::minimizeSize(bool remove_taut)
  {
  relset_t  changed_relations;
  double    oldsize = getRelSize() ;
  
  std::cout << "Minimize size. size=" << oldsize << "\n";

  bool changed;
  do
    {
    relset_t  rels_to_check;
    varset_t  bounded_variables;

    node_consistency(*this, &rels_to_check);
    changed = weak_arc_consistency(*this, &rels_to_check);

    if (changed && remove_taut)
      removeTautologyColumns(rels_to_check, changed_relations, true);
    checkInvariants();
    changed |= !changed_relations.empty();
    }
  while(changed);
  std::cout << "Minimize size done. deltasize=" << getRelSize()-oldsize << "\n";
  }

void CSP::preprocess(size_t maxs, size_t max0)
  {
  std::cout << "Preprocessing...\n";
  joinEqualScope(maxs);
  join0(max0);
  minimizeSize(true);
  removeRedundancy();
  std::cout << "Preprocessing done\n";
  }

void CSP::postprocess()
  {
  std::cout << "Postprocessing...\n";
  
  // Move isolated rels (if any) back
  m_Rels.splice(m_Rels.end(), m_IsolatedRels);
  buildXRef();
  minimizeSize(false);
  removeRedundancy();
  std::cout << "Postprocessing done\n";
  }

void CSP::build(size_t maxs, size_t max0, size_t maxg)
  {
  varset_t  bound_variables;
  relset_t  changed_relations;

  std::cout << "Building using heuristic\n";

  preprocess(maxs, max0);
  std::cout << "Isolating variables\n";
  relcol_t::iterator it = m_Rels.begin(), pend = m_Rels.end();
  while(it != pend)
    {
    Relation *r = *it;
    ++it;
    // *r might be removed.....
    isolateVariables(*r);
    }

  joinGOO(maxg);
  std::cout << "Heuristic done\n";
  postprocess();
  }

void buildMoeller(CSP &instance, size_t maxsize)
{
  CSP::relcol_t &rels = instance.getRelations();
  
  std::cout << "Joining relations using Mller's heuristic. #rels=" << rels.size() << "\n";

  while(rels.size() > 1)
    {
    JoinPair p = greedySelectJoinPair(GLMCostFunction(), rels.begin(), rels.end());

    if (p.JoinCost == std::numeric_limits<float>::infinity())
      break;

    std::cout << "  #rels=" << rels.size() << ", JoinCost=" << p.JoinCost << " ";

    try
      {
      instance.joinRels(p.first, p.second, maxsize);
      }
    catch(std::length_error &)
      {
      std::cout << "Cannot build instance, join result is larger than " << maxsize << "\n";
      exit(EXIT_FAILURE);
      }
    rels.erase(p.second);
    (*p.first)->compress();
    }
}
