/*
 * Copyright (C) 2003 Jeppe Nejsum Madsen, nejsum@diku.dk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include "stdafx.h"
#include <fstream>
#include "timer.h"
#include "csplib/adb/adbio.hpp"
#include "csplib/relation.hpp"
#include "csplib/graph.hpp"
#include "csp.hpp"

extern void uniform_network(CSP &instance, size_t maxrel);
extern void buildMoeller(CSP &instance, size_t maxsize);

size_t  MaxRelSize0 = 20000;
size_t  MaxRelSizeS = 20000;
size_t  MaxRelSizeG = 500000000;
size_t  MaxRelSizeR = 40000;
size_t  MinSplit    = 2;

void usage()
{
  std::cout << "usage: compilecsp [options] infile outfile\n"
    "  options:\n"
    "   -j0 Join relations with join factor 0 until all rels > MaxRelSize\n"
    "   -js Join all relations with same scope\n"
    "   -jg Join all relations using GOO heuristic\n"
    "\n"
    "   -h  Build using all heuristics\n"
    "   -m  Build using Mller's heuristics\n"
    "   -t  Build using tree clustering\n"
    "   -p  Postprocess tree clustering (create uniform network)\n"
    "   -d  run directional arc consistency\n"
    "\n"
    "   -r0n Set MaxRelSize=n for join with factor 0\n"
    "   -rsn Set MaxRelSize=n for join with same scope\n"
    "   -rgn Set MaxRelSize=n for other joins\n"
    "   -rrn Set MaxRelSize=n for redundancy removal\n"
    "   -xn  Set maximum amount of user CPU usage to n seconds\n"
    "   -yn  Set maximum amount of commited memory to n MB\n"
    "   -sn Set MinSplit=n\n"
    "   -c  Compress file\n"
    "   -u  Uncompress file\n"
    "   -n  Normalize (i.e., uncompress/compress) file\n"
    "   -gd Create dual graph (neaty format)\n"
    "   -gp Create primal graph (neaty format)\n"
    "   -gt Create join tree (neaty format)\n"
    ;
  exit(EXIT_FAILURE);
}

void UncompressFile(const char *infile, const char *outfile)
{
  CSP             instance;
  std::cout << "Uncompressing....\n";
  
  ADB::importADBFile(instance, infile);
  instance.uncompressAll();
  ADB::exportADBFile(instance, outfile);
  }

void CompressFile(const char *infile, const char *outfile)
  {
  CSP             instance;
  
  std::cout << "Compressing....\n";
  ADB::importADBFile(instance, infile);
  instance.compressAll();
  ADB::exportADBFile(instance, outfile);
  }

void normalize(const char *infile, const char *outfile)
  {
  CSP             instance;
  
  std::cout << "Normalizing....\n";
  ADB::importADBFile(instance, infile);
  CSP::relcol_t::iterator it, it_end = instance.getRelations().end();
  for(it = instance.getRelations().begin(); it != it_end; ++it)
    {
    Relation *rel = *it;

    rel->removeRedundancy();
    rel->uncompress();
    rel->compress();
    }
  ADB::exportADBFile(instance, outfile);
  }

void DAC(const char *infile, const char *outfile)
  {
  CSP             instance;
  
  std::cout << "Running DAC....\n";
  ADB::importADBFile(instance, infile);
  instance.postprocess();
  dualgraph_t dg;
  createDualGraph(instance, dg);
  create_join_tree(dg);
  directional_arc_consistency(dg, MaxRelSizeG);
  instance.postprocess();
  ADB::exportADBFile(instance, outfile);
  }

void PostProcess(const char *infile, const char *outfile)
  {
  CSP             instance;
  ADB::importADBFile(instance, infile);
  uniform_network(instance, MaxRelSizeG);
  instance.normalizeLinkVariables();
  ADB::exportADBFile(instance, outfile);
  }

void CreateGraph(const char *infile, const char *outfile, char gtype)
  {
  CSP             instance;
  ADB::importADBFile(instance, infile);
  std::ofstream f(outfile);

  switch(gtype)
    {
    case 'd':
      {
      dualgraph_t g;
      std::cout << "Creating dual graph....\n";
      createDualGraph(instance, g);
      std::cout << "Writing file " << outfile << "....\n";
      f << g;
      }
      break;
    case 't':
      {
      std::cout << "Creating join tree....\n";
      dualgraph_t g;
      createDualGraph(instance, g);
      create_join_tree(g);
      std::cout << "Writing file " << outfile << "....\n";
      f << g;
      }
      break;
    case 'p':
      {
      primalgraph_t g;
      std::cout << "Creating primal graph....\n";
      createPrimalGraph(instance, g);
      std::cout << "Writing file " << outfile << "....\n";
      f << g;
      }
      break;
    default:
      std::cerr << "Unknown graph: " << gtype << "\n";
    }
  }

void Build(const char *infile, const char *outfile)
  {
  CSP             instance;
  ADB::importADBFile(instance, infile);
  instance.build(MaxRelSizeS, MaxRelSize0, MaxRelSizeG);
  instance.normalizeLinkVariables();
  ADB::exportADBFile(instance, outfile);
  }

void BuildTC(const char *infile, const char *outfile)
  {
  CSP             instance;
  ADB::importADBFile(instance, infile);
  tree_clustering(instance, MaxRelSizeS, MaxRelSize0, MaxRelSizeG);
  ADB::exportADBFile(instance, outfile);
  }


void BuildMoeller(const char *infile, const char *outfile)
  {
  CSP             instance;
  ADB::importADBFile(instance, infile);
  buildMoeller(instance, MaxRelSizeG);
  ADB::exportADBFile(instance, outfile);
  }

void JoinFile(const char *infile, const char *outfile, char jointype)
  {
  CSP             instance;
  
  ADB::importADBFile(instance, infile);
  std::cout << "Joining, type=" << jointype
            << ", size=" << instance.getRelSize() 
            << ", #rels=" << instance.getRelations().size() << "....\n";
  switch(jointype)
    {
    case 'f':
      {
      CSP::relcol_t::iterator it = instance.getRelations().begin();
      Relation *r1 = *it;
      it = instance.getRelations().erase(it);
      Relation *r2 = *it;
      instance.getRelations().erase(it);
      std::auto_ptr<Relation> newRel = r1->join(*r2, 42, MaxRelSizeG);
      delete r1;
      delete r2;
      instance.addRelation(newRel);
      }
      break;
    case '0':
      instance.join0(MaxRelSize0);
      break;
    case 's':
      instance.joinEqualScope(MaxRelSizeS);
      break;
    case 'g':
      instance.joinGOO(MaxRelSizeG);
      break;
    default:
      usage();
    }
  instance.normalizeLinkVariables();
  ADB::exportADBFile(instance, outfile);
  }

int main(int , char **argv)
{
  //_CrtSetDbgFlag(_CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF);

  int res = EXIT_SUCCESS;
  JobInfo ji;

  std::cout.setf(std::ios_base::unitbuf, std::ios_base::unitbuf);

  argv++;
  char jointype, timejointype, timecomptype, graph;
  bool norm,dac,tree_clustering, compress, uncompress, build, moeller,post;
  int max_memory_mb = 0, max_cpu_seconds = 0;

  jointype = timejointype = timecomptype = graph = '\0';
  norm = dac = tree_clustering = compress = uncompress = build = moeller = post= false;
  const char *infile = 0, *outfile = 0, *arg;
  while(*argv)
    {
    arg = *argv++;

    if (*arg == '-')
      {
      arg++;
      switch(*arg)
        {
        case 't':
          tree_clustering = true;
          break;
        case 'd':
          dac = true;
          break;
        case 'n':
          norm = true;
          break;
        case 'm':
          moeller = true;
          break;
        case 'p':
          post = true;
          break;
        case 'g':
          arg++;
          graph = *arg;
          break;
        case 'u':
          uncompress = true;
          break;
        case 'c':
          compress = true;
          break;
        case 'h':
          build = true;
          break;
        case 'j':
          arg++;
          jointype = *arg;
          break;
        case 'x':
          arg++;
          max_cpu_seconds = atoi(arg);
          break;
        case 'y':
          arg++;
          max_memory_mb = atoi(arg);
          break;
        case 'r':
          arg++;
          switch(*arg++)
            {
            case '0':
              MaxRelSize0 = atoi(arg);
              break;
            case 's':
              MaxRelSizeS = atoi(arg);
              break;
            case 'g':
              MaxRelSizeG = atoi(arg);
              break;
            case 'r':
              MaxRelSizeR = atoi(arg);
              break;
            default:
              usage();
            }
          break;
        case 's':
          arg++;
          MinSplit = atoi(arg);
          break;        default:
          usage();
        }
      }
    else
      {
      if (infile == 0)
        infile = arg;
      else if (outfile == 0)
        outfile = arg;
      else
        usage();
      }
    }

  if (infile == 0 || outfile == 0)
    usage();

  if (max_cpu_seconds > 0 || max_memory_mb > 0)
    ji.setLimits(max_cpu_seconds, max_memory_mb);

  try
    {
    if (moeller)
      {
      BuildMoeller(infile, outfile);
      }
    else if (post)
      {
      PostProcess(infile, outfile);
      }
    else if (tree_clustering)
      {
      BuildTC(infile, outfile);
      }
    else if (dac)
      {
      DAC(infile, outfile);
      }
    else if (norm)
      {
      normalize(infile, outfile);
      }
    else if (build)
      {
      Build(infile, outfile);
      }
    else if (graph != '\0')
      {
      CreateGraph(infile, outfile, graph);
      }
    else if (compress)
      {
      CompressFile(infile, outfile);
      }
    else if (uncompress)
      {
      UncompressFile(infile, outfile);
      }
    else if (jointype != '\0')
      {
      JoinFile(infile, outfile, jointype);
      }
    else
      usage();
    }
  catch(const std::exception &ex)
    {
    std::cout << "Exception: " << ex.what();
    std::cout << "\nType: " << typeid(ex).name() << "\n" << std::flush;
    res = EXIT_FAILURE;
    }

  catch(...)
    {
    std::cout << "Unknown exception caught\n";
    res = EXIT_FAILURE;
    }
  return res;
  }
