/*
 * Copyright (C) 2003 Jeppe Nejsum Madsen, nejsum@diku.dk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include "stdafx.h"
#include <numeric>
#include "boost/mem_fn.hpp"
#include "boost/bind.hpp"
#include "csplib/uniquerows.hpp"
#include "csplib/utility/formatter.hpp"
#include "csp.hpp"
 
extern size_t MinSplit, MaxRelSizeR;

void CSP::compressAll()
  {
  std::for_each(m_Rels.begin(), m_Rels.end(), boost::bind(&Relation::compress, _1, (const varset_t *)0));
  }

void CSP::uncompressAll()
  {
  std::for_each(m_Rels.begin(), m_Rels.end(), boost::bind(&Relation::uncompress, _1, true));
  }

void CSP::removeRedundancy(relset_t *changed_relations)
  {
  CSP::relcol_t::iterator it, it_end = m_Rels.end();

  Formatter f("Removing redundancy");
  size_t i = 1;
  for(it = m_Rels.begin(); it != it_end; ++it)
    {
    Relation *r = *it;
    f.writeProgress(i++, m_Rels.size(), r->getNumRows());
    if (r->getNumRows() < MaxRelSizeR && r->removeRedundancy() && changed_relations)
      changed_relations->insert(r);
    }
  }

void CSP::pruneInvalidTuples()
  {
  std::for_each(m_Rels.begin(), m_Rels.end(), boost::mem_fn(&Relation::pruneInvalidTuples));
  std::for_each(m_IsolatedRels.begin(), m_IsolatedRels.end(), boost::mem_fn(&Relation::pruneInvalidTuples));
  }

bool CSP::isolateVariables(Relation &r)
  {
  const colvector_t &cols = r.getColumns();
  colvector_t  independent_cols;
  bool         removed = false;

  colvector_t::const_iterator it = cols.begin(), pend = cols.end();
  while(it != pend)
    {
    Variable &var = (*it)->getVariable();
    if (getRelations(var.getId()).size() == 1 ||
        (getRelations(var.getId()).size() == 2 && var.isLinkVar()))
      {
      // Variable is only in this relation or it is a link variable
      // in which case it is safe to isolate
      independent_cols.push_back(*it);
      }
    ++it;
    }

  // If all vars are isolated, don't split the relation but
  // move relation to list of isolated relations
  if (independent_cols.size() == cols.size())
    {
    std::cout << "Isolating relation #" << r.getId() << "\n";
    Relation *rel = removeRelation(&r, false).release();
    rel->compress();
    m_IsolatedRels.push_back(rel);
    removed = true;
    }
  else  if (independent_cols.size() >= MinSplit)
    // Only split if found isolated vars and not all vars are
    // isolated
    {
    std::cout << "Isolating " << independent_cols.size() << " variables in #" << r.getId() << "\n";

    removeRelationXRef(&r);
    Relation *isol = r.split(independent_cols, getNewRelId()).release();

    linkRelations(r, *isol);

    addRelationXRef(&r);
    addRelationXRef(isol);
    m_IsolatedRels.push_back(isol);

    relset_t dummy;
    removed = removeTautologyColumns(r, dummy, true);
    removeTautologyColumns(*isol, dummy, true);
    }

  return removed;
  }

void CSP::linkRelations(Relation &r1, Relation &r2)
  {
  size_t  new_id = getNumVariables();

  std::auto_ptr<Variable> lv = r1.link(r2, new_id);
  if (lv.get() != 0)
    addVariable(lv);
  }

Relation *CSP::joinRels(CSP::relcol_t::iterator rit1, CSP::relcol_t::iterator rit2, size_t maxrows)
  {
  Relation *r1 = *rit1, *r2 = *rit2, *newrel;

  std::cout << "  Join " << r1->getId() << "(" << r1->getNumRows() << ")," << r2->getId()<< "(" << r2->getNumRows() << ")\n";
  try
    {
    newrel = r1->join(*r2, getNewRelId(), maxrows).release();
    }
  catch(std::length_error &)
    {
    std::cout << "** Relation too large **\n";
    throw;
    }

  if (newrel->getNumRows() == 0)
    throw contradiction_error(r1, r2);

  removeRelationXRef(r1);
  removeRelationXRef(r2);
  boost::checked_delete(r1);
  boost::checked_delete(r2);

  addRelationXRef(newrel);

  *rit1 = newrel;
  *rit2 = 0;
  return  newrel;
  }

bool CSP::removeTautologyColumns(relset_t &rels_to_check, relset_t &relset_to_update, bool compress)
  {
  bool any_removed = false;
  relset_t::iterator it = rels_to_check.begin();
  while(it != rels_to_check.end())
    {
    any_removed |= removeTautologyColumns(*(*it), relset_to_update, compress);
    ++it;
    }
  return any_removed;
  }

// Return true if relation is deleted
bool CSP::removeTautologyColumns(Relation &r, relset_t &relset_to_update, bool compress)
  {
  bool removed = false;

  if (compress)
    r.compress();

  if (r.getNumRows() == 1)
    {
    // If only 1 row, remove relation
    std::cout << "Removed redundant relation (1 row) #" << r.getId() << "\n";
    relset_to_update.erase(&r);
    r.identifyUnaryConstraints();
    removeRelation(&r);
    removed = true;
    }
  else
    {
    varset_t tautology_vars;

    if (r.removeTautologyColumns(tautology_vars))
      {
      // Update xref
      varset_t::const_iterator it = tautology_vars.begin(), pend = tautology_vars.end();
      while(it != pend)
        {
        m_VarRels[(*it)->getId()].erase(&r);
        ++it;
        }                                          

      if (r.getNumColumns() < 2)
        {
        removed = true;
        std::cout << "Removed redundant relation (1 column) #" << r.getId() << "\n";
        relset_to_update.erase(&r);
        removeRelation(&r);
        }
      else
        {
        removed = removeTautologyColumns(r, relset_to_update, true);
        }
      }
    }
  return removed;
  }


void CSP::findNewBoundVariables(relset_t &rels_to_check, varset_t &bound_variables)
  {
  std::for_each(rels_to_check.begin(), rels_to_check.end(), 
                boost::bind(&Relation::identifyUnaryConstraints, _1, &bound_variables));
  std::cout << "Found " << bound_variables.size() << " bound vars in " << rels_to_check.size() << " relations\n";
  }

bool CSP::modifyBoundVariableDomain(Variable &var)
  {
  size_t  new_dom_size = var.getValidDomainSize();
  bool    changed = new_dom_size < var.getDomainSize();

  if (changed)
    {
    idvector_t  new_ids(var.getDomainSize());
    
    for(size_t j = 0, i = 0; i < var.getDomainSize(); ++i)
      {
      if (var.isValueValid(i))
        new_ids[i] = j++;
      else
        new_ids[i] = std::numeric_limits<idvector_t::value_type>::max();
      }

    relset_t &var_relations = getRelations(var.getId());
    relset_t::iterator it = var_relations.begin(), pend = var_relations.end();
    while (it != pend)
      {
      Relation *r = *it;
      r->pruneInvalidTuples();
      r->getColumn(var).remapIndicies(new_ids);
      ++it;
      }
    var.removeInvalidValues();
    }

  return changed;
  }

void CSP::normalizeLinkVariables()
  {
  std::vector<Variable *>::iterator it = m_Vars.begin(), pend = m_Vars.end();
  while(it != pend)
    {
    LinkVariable  *lv = dynamic_cast<LinkVariable *>(*it);
    if (lv != 0)
      {
      modifyBoundVariableDomain(*lv);
      }
    ++it;
    }
  }
