/*
 * Copyright (C) 2003 Jeppe Nejsum Madsen, nejsum@diku.dk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#ifndef CSP_H
#define CSP_H

#include "csplib/constraint_network.hpp"

// The CSP Problem instance
class CSP : public ConstraintNetwork
  {
  public:
    CSP(std::string name = "unnamed") : ConstraintNetwork(name) {};
    ~CSP(){};

    // Highlevel global operations

    // (un) compress all relations
    void        compressAll();
    void        uncompressAll();

    // Build using heuristic
    void        build(size_t maxs, size_t max0, size_t maxg);
    void        preprocess(size_t maxs, size_t max0);
    void        postprocess();

    // Split relation if it has isolated variables. Return new relation (if any)
    bool        isolateVariables(Relation &r);

    // Identify any new bound variables
    void        findNewBoundVariables(relset_t &rels_to_check, varset_t &bound_variables);

    // Tighten relations which have bound variables
    bool        reviseRelations(const varset_t &bound_variables, relset_t &revised_relations);

    bool        removeTautologyColumns(Relation &rel, relset_t &rels_to_update, bool compress);
    bool        removeTautologyColumns(relset_t &rels_to_check, relset_t &rels_to_update, bool compress);

    void        minimizeSize(bool remove_taut);

    void        removeRedundancy(relset_t *changed_relations = 0);

    // Remove bound values from link variable domains
    void        normalizeLinkVariables();


    void        joinEqualScope(size_t maxsize);
    void        join0(size_t maxsize);
    void        joinGOO(size_t maxsize);

    Relation *joinRels(relcol_t::iterator rit1, relcol_t::iterator rit2, size_t maxrows);

    // Link two relations with link variable
    void        linkRelations(Relation &r1, Relation &r2);

  private:
    // Remove bound values from variable domain and update relations accordingly
    bool      modifyBoundVariableDomain(Variable &var);
    void      pruneInvalidTuples();
  };


  // Relation Pair with associated join cost
  struct JoinPair : public std::pair<CSP::relcol_t::iterator, CSP::relcol_t::iterator>
    {
    float JoinCost;
    };

  struct JoinPairOrder
    {
    bool operator()(const JoinPair &lhs, const JoinPair &rhs)
      {
      return lhs.JoinCost < rhs.JoinCost;
      }
    };
    template<typename CostFunction>

  JoinPair greedySelectJoinPair(CostFunction f, CSP::relcol_t::iterator begin, CSP::relcol_t::iterator end)
    {
    JoinPair  best;

    best.JoinCost = std::numeric_limits<float>::infinity();

    while(begin != end)
      {
      CSP::relcol_t::iterator it = begin;
      ++it;
      while(it != end)
        {
        float jc =  f(*begin, *it, best.JoinCost);
        if (jc < best.JoinCost)
          {
          best.first = begin;
          best.second = it;
          best.JoinCost = jc;
          }
        ++it;
        }
      ++begin;
      }
    return best;
    }

  struct GLMCostFunction
    {
    float operator()(const Relation *r1, const Relation *r2, float) const;
    };
  struct SelectivityCost
    {
    float operator()(const Relation *r1, const Relation *r2, float) const;
    };
  struct JoinSizeCost
    {
    float operator()(const Relation *r1, const Relation *r2, float) const;
    };
  struct ADBCost
    {;
    float operator()(const Relation *r1, const Relation *r2, float) const;
    };

  
void tree_clustering(CSP &instance,  size_t maxs, size_t max0, size_t max_rel_size);

#endif
