/*
 * Copyright (C) 2003 Jeppe Nejsum Madsen, nejsum@diku.dk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include "stdafx.h"
#include <fstream>
#include "boost/property_map.hpp"
#include "boost/graph/connected_components.hpp"
#include "boost/graph/breadth_first_search.hpp"
#include "boost/graph/minimum_degree_ordering.hpp"
#include "csplib/variable.hpp"
#include "csplib/relation.hpp"
#include "csplib/graph.hpp"
#include "csp.hpp"

using namespace boost;

typedef std::vector<varset_t>                           cliqueset_t;
typedef std::vector<int>                                permutation_t;
typedef graph_traits<primalgraph_t>::vertex_descriptor  Vertex;
typedef graph_traits<primalgraph_t>::out_edge_iterator  EdgeIt;
typedef std::set<Vertex>                                VertexSet;

//#define DUMP(x)
#define DUMP(x) x;

// Convert undirected graph to the equivalent directed graph by
// adding an new edge (v,w) for each edge (w,v) in the undirected graph
template <typename Undirected,
          typename Directed>
void make_directed(const Undirected &in, Directed &out)
  {
  typename graph_traits<Undirected>::edge_iterator ei, ei_end;
  tie(ei,ei_end) = edges(in);
  while(ei != ei_end)
    {
    add_edge(source(*ei,in), target(*ei,in),out);
    add_edge(target(*ei,in), source(*ei,in), out);
    ++ei;
    }
  }

void fill_in(primalgraph_t &G, const permutation_t &perm)
  {
  std::cout << "Triangulating primal graph...\n";

  size_t n = num_vertices(G);
  std::vector<VertexSet>  M(n);

  // Find neighourhoods
  for (size_t i = 0; i < n; ++i)
    {
    primalgraph_t::adjacency_iterator ait, ait_end;
    for(tie(ait, ait_end) = adjacent_vertices(i,G); ait != ait_end; ++ait)
      M[i].insert(*ait);
    }

  for (size_t i = 0; i < n; ++i)
    {
    Vertex            vi = perm[i];

    VertexSet::iterator it, it2, it_end = M[vi].end();
    for(it = M[vi].begin(); it != it_end; ++it)
      {
      it2 = it;
      while(++it2 != it_end)
        {
        primalgraph_t::edge_descriptor e;
        bool found;
        tie(e,found) = edge(*it, *it2, G);
        if (!found)
          {
          add_edge(*it, *it2, G);
          M[*it].insert(*it2);
          M[*it2].insert(*it);
          }
        }
      }
    for(it = M[vi].begin(); it != it_end; ++it)
        {
        M[*it].erase(vi);
        }
    }
  std::cout << "#Vert=" << num_vertices(G) << ", #Edge=" << num_edges(G) << "\n";
  }

void find_max_cliques(primalgraph_t &G, 
                      const permutation_t &perm, 
                      cliqueset_t &max_cliques)
  {
  std::cout << "Finding cliques...\n";

  property_map<primalgraph_t, vertex_variable_t>::type vertexvar = get(vertex_variable_t(), G);
  std::set<VertexSet>  C;
  size_t                n = num_vertices(G);

  boolvector_t        marked(n, false);

  for(size_t i = 0; i < n; ++i)
    {
    Vertex            u = perm[i];
    VertexSet         S;
    primalgraph_t::adjacency_iterator ait, ait_end;

    if (marked[u])
      continue;

    for(tie(ait, ait_end) = adjacent_vertices(u,G); ait != ait_end; ++ait)
      {
      if (!marked[*ait])
        {
        S.insert(*ait);
        }
      }
    S.insert(u);
    std::set<VertexSet>::const_iterator it = C.begin(), it_end = C.end();
    for(; it != it_end; ++it)
      {
      const VertexSet  &vs = *it;
      if(std::includes(vs.begin(), vs.end(), S.begin(), S.end()))
        break;
      }

    if (it == it_end)
      C.insert(S);
    marked[u] = true;
    }

  size_t maxsize = 0;
  max_cliques.clear();

  std::set<VertexSet>::const_iterator it = C.begin(), it_end = C.end();
  for(; it != it_end; ++it)
    {
    const VertexSet  &vs = *it;
    if (vs.size() > maxsize)
      maxsize = vs.size();
    varset_t vars;
    VertexSet::const_iterator vi = vs.begin(), vi_end = vs.end();
    for(; vi != vi_end; ++vi)
      {
      vars.insert(vertexvar[*vi]);
      }
    max_cliques.push_back(vars);
    }
  std::cout << "Found " << max_cliques.size() << " cliques, max clique size is " << maxsize << "\n";
  }

void min_degree_ordering(const primalgraph_t &g, permutation_t &inverse_perm, permutation_t &perm)
  {
  // Boost min_degree needs a directed graph......
  typedef adjacency_list <vecS, vecS, directedS> directed_primalgraph_t;

  std::cout << "Finding a vertex order...\n";

  directed_primalgraph_t dd;
  make_directed(g, dd);

  std::vector<int> degree(num_vertices(dd), 0);
  std::vector<size_t> supernode_sizes(num_vertices(dd), 1);
  property_map<directed_primalgraph_t, vertex_index_t>::type id = get(vertex_index, dd);
  
  inverse_perm = permutation_t(num_vertices(dd), 0);
  perm = permutation_t(num_vertices(dd), 0);

  directed_primalgraph_t::edge_iterator ei, ei_end;

  minimum_degree_ordering
    (dd,
     make_iterator_property_map(&degree[0], id, degree[0]),
     &inverse_perm[0],
     &perm[0],
     make_iterator_property_map(&supernode_sizes[0], id, supernode_sizes[0]), 
     0,
     id);
  }

bool is_clique(const varset_t &clique, const primalgraph_t &G)
  {
  varset_t::const_iterator it, it2, it_end = clique.end();
  for(it = clique.begin(); it != it_end; ++it)
    {
    it2 = it;
    while(++it2 != it_end)
      {
      primalgraph_t::edge_descriptor e;
      bool found;
      tie(e,found) = edge((*it)->getId(), (*it2)->getId(), G);
      if (!found)
        {
        std::cout << "No edge between " << (*it)->getName() << " and " << (*it2)->getName() << " in clique:\n";
        write_set(std::cout, clique) << "\n";
        return false;
        }
      }
    }
  return true;
  }

bool is_max_clique(const cliqueset_t cliques, const varset_t &clique)
  {
  cliqueset_t::const_iterator it, it_end = cliques.end();
  for(it = cliques.begin(); it != it_end; ++it)
    {
    if (*it != clique)
      {
      varset_t diff = clique - *it;
      if (diff.empty())
        {
        std::cout << "Clique "; write_set(std::cout, clique) << " is subset of "; write_set(std::cout, *it) << "\n";
        return false;
        }
      }
    }
  return true;
  }

void check_cliques(const cliqueset_t &cliques, const primalgraph_t &G)
  {
  cliqueset_t::const_iterator it = cliques.begin();

  std::cout << "Checking cliques...\n";
  while(it != cliques.end())
    {
    if (!(is_clique(*it, G) && is_max_clique(cliques, *it)))
      {
      std::cout << "Not max clique !!\n";
      INTERNAL_ERROR;
      }
    ++it;
    }
  }

void augment_rel(Relation &rel, varset_t &vars)
  {
  varset_t tautvars = vars - rel.getVariables();

  varset_t::iterator it, it_end = tautvars.end();
  for(it = tautvars.begin(); it != it_end; ++it)
    {
    rel.addTautologyColumn(*(*it));
    }
  }

void add_loose_clique_rel(CSP &instance, varset_t &vars, size_t max_rel_size, relset_t &clique_rels)
  {
  CSP::relcol_t rels;

  // Find relations contained in vars
  CSP::relcol_t::iterator it, it_end = instance.getRelations().end();
  for(it = instance.getRelations().begin(); it != it_end; ++it)
    {
    Relation *r = *it;
    varset_t diff = r->getVariables() - vars;
    if (diff.empty())
      rels.push_back(r);
    }

  // Remove found relations from instance
  it_end = rels.end();
  for(it = rels.begin(); it != it_end; ++it)
    {
    Relation *r = *it;
    std::auto_ptr<Relation> rel = instance.removeRelation(r);
    rel.release();
    }

  // rels contains the relations to be joined
  while(rels.size() > 1)
    {
    JoinPair p = greedySelectJoinPair(SelectivityCost(), rels.begin(), rels.end());

    std::cout << "  #rels=" << rels.size() << ", JoinCost=" << p.JoinCost << " ";

    Relation                *jr1 = *p.first, *jr2 = *p.second;

    rels.erase(p.second);
    rels.erase(p.first);
 
    std::cout << "size=" << vars.size() << "\n";

    std::auto_ptr<Relation> r = jr1->joinCompress(*jr2, instance.getNewRelId(), max_rel_size);

    delete jr1;
    delete jr2;
    rels.push_back(r.release());
    }

  // Add remaining rels
  if (rels.size() == 0)
    {
    std::auto_ptr<Relation> rel(new Relation(instance.getNewRelId(), "uni"));
    augment_rel(*rel, vars);
    std::cout << "Loose Universal: " << *rel << "\n";
    clique_rels.insert(rel.release());
    }
  else if (rels.size() == 1)
    {
    Relation *rel = *rels.begin();

    if (rel->getVariables().size() != vars.size())
      {
      augment_rel(*rel, vars);
      std::cout << "Loose Augmented: " << *rel << "\n";
      }
    else
      {
      std::cout << "Loose CliqueRel: " << *rel << "\n";
      }
    clique_rels.insert(rel);
    }
  else
    {
    std::cout << "More than 1 rel!!\n";
    INTERNAL_ERROR;
    }
  }

void add_tight_clique_rel(CSP &instance, varset_t &vars, size_t max_rel_size, relset_t &clique_rels)
  {
  relset_t  relset;

  // Find relations which include variables in vars 
  varset_t::iterator it = vars.begin();
  while(it != vars.end())
    {
    relset_t varrels = instance.getRelations((*it)->getId());
    relset.insert(varrels.begin(), varrels.end());
    ++it;
    }

  CSP::relcol_t rels(relset.begin(), relset.end());
  
  // rels contains the relations to be joined
  while(rels.size() > 1)
    {
    JoinPair p = greedySelectJoinPair(SelectivityCost(), rels.begin(), rels.end());

    std::cout << "  #rels=" << rels.size() << ", JoinCost=" << p.JoinCost << " ";

    Relation                *jr1 = *p.first, *jr2 = *p.second;

    rels.erase(p.second);
    rels.erase(p.first);

    // Find union of variables remaining to be joined
    varset_t  varunion(vars);
    CSP::relcol_t::iterator rit = rels.begin();
    while (rit != rels.end())
      {
      varunion.insert((*rit)->getVariables().begin(), (*rit)->getVariables().end());
      ++rit;
      }
    std::cout << "size=" << vars.size() << ", networkvars=" << varunion.size() << "\n";

    std::auto_ptr<Relation> r1, r2, r3;

    varset_t diff1 = jr1->getVariables() - varunion;
    if (!diff1.empty())
      {
      r1 = jr1->project(varunion,instance.getNewRelId());
      r1->compress();
      jr1 = r1.get();
      }
    varset_t diff2 = jr2->getVariables() - varunion;
    if (!diff2.empty())
      {
      r2 = jr2->project(varunion,instance.getNewRelId());
      r2->compress();
      jr2 = r2.get();
      }

    varset_t  remove_vars = jr1->getVariables() + jr1->getVariables();
    std::auto_ptr<Relation> r = jr1->joinCompress(*jr2, instance.getNewRelId(), max_rel_size, remove_vars - varunion);
    rels.push_back(r.release());
    }

  // Add remaining rels
  if (rels.size() == 0)
    {
    std::auto_ptr<Relation> rel(new Relation(instance.getNewRelId(), "uni"));
    augment_rel(*rel, vars);
    std::cout << "Tight Universal: " << *rel << "\n";
    clique_rels.insert(rel.release());
    }
  else if (rels.size() == 1)
    {
    Relation *rel = *rels.begin();

    if (rel->getVariables().size() != vars.size())
      {
      augment_rel(*rel, vars);
      std::cout << "Tight Augmented: " << *rel << "\n";
      }
    else
      {
      std::cout << "Tight CliqueRel: " << *rel << "\n";
      }
    clique_rels.insert(rel);
    }
  else
    {
    std::cout << "More than 1 rel!!\n";
    INTERNAL_ERROR;
    }
  }

void tree_clustering(CSP &instance, size_t maxs, size_t max0, size_t max_rel_size)
  {
  primalgraph_t     G;
  cliqueset_t       max_cliques;
  relset_t          clique_rels;
  permutation_t     perm, inv_perm;

  instance.preprocess(maxs, max0);

  dualgraph_t jt;
  createDualGraph(instance, jt);

  std::vector<int> component(num_vertices(jt));
  int num = connected_components(jt, &component[0]);
  std::cout << "Dual graph has " << num << " connected components\n";

  createPrimalGraph(instance, G);
  min_degree_ordering(G, inv_perm, perm);
  fill_in(G, perm);
  check_chordality(G);

  find_max_cliques(G, perm, max_cliques);
  check_cliques(max_cliques, G);

  cliqueset_t::iterator it = max_cliques.begin();
  size_t                clique_num = 1;

  while(it != max_cliques.end())
    {
    std::cout << "Building clique " << clique_num << " of " << max_cliques.size() << ", ";
    write_set(std::cout, *it);
    try
      {
      add_tight_clique_rel(instance,*it,max_rel_size, clique_rels);
      }
    catch(const std::length_error &)
      {
      add_loose_clique_rel(instance,*it,max_rel_size*1000, clique_rels);
      }

    ++it;
    ++clique_num;
    }

  // Delete relations which are no longer used
  CSP::relcol_t::iterator rit = instance.getRelations().begin(), rit_end = instance.getRelations().end();
  while(rit != rit_end)
    {
    Relation *r = *rit;
    if (clique_rels.find(r) == clique_rels.end())
      delete r;
    ++rit;
    }

  // Insert new relations
  instance.getRelations().clear();
  instance.getRelations().insert(instance.getRelations().begin(), clique_rels.begin(), clique_rels.end());
  instance.buildXRef();
  //instance.postprocess();

  jt.clear();
  createDualGraph(instance, jt);
  component.resize(num_vertices(jt));
  num = connected_components(jt, &component[0]);
  std::cout << "Dual graph has " << num << " connected components\n";
  create_join_tree(jt);
  check_rip(jt);
  instance.postprocess();
  }

void combine_root(CSP &instance, size_t maxrel, dualgraph_t::edge_descriptor edge, dualgraph_t &jt)
  {
  property_map<dualgraph_t, vertex_relation_t>::type vertexrel = get(vertex_relation_t(), jt);

  std::cout << "Combine Root " << edge << "\n";

  Relation *r1 = vertexrel[source(edge, jt)];
  Relation *r2 = vertexrel[target(edge, jt)];

  varset_t edge_vars = intersection(r1->getVariables(), r2->getVariables()), diff1, diff2;
  DUMP(std::cout << "R1 " << *r1 << "\n" << "R2 " << *r2<< " edge=";write_set(std::cout, edge_vars) << "\n");
  
  diff1 = r1->getVariables() - edge_vars;
  if (diff1.size() > 0)
    {
    std::auto_ptr<Relation> splitrel = r1->split(diff1, instance.getNewRelId());
    instance.linkRelations(*r1, *splitrel);

    r1->compress();
    splitrel->compress();
    if (splitrel->getNumRows() < 2)
      {
      // We can remove the relation if we update
      // the domains....
      splitrel->identifyUnaryConstraints();
      std::cout << "Split removed....\n";
      }
    else
      instance.addRelation(splitrel);
    }

  diff2 = r2->getVariables() - edge_vars;
  if (diff2.size() > 0)
    {
    std::auto_ptr<Relation> splitrel = r2->split(diff2, instance.getNewRelId());
    instance.linkRelations(*r2, *splitrel);
    r2->compress();
    splitrel->compress();
    if (splitrel->getNumRows() < 2)
      {
      // We can remove the relation if we update
      // the domains....
      splitrel->identifyUnaryConstraints();
      std::cout << "Split removed....\n";
      }
    else
      instance.addRelation(splitrel);
    }
  try
    {
    std::auto_ptr<Relation> newrel = r1->join(*r2, instance.getNewRelId(), maxrel);
    if (newrel->getNumRows() == 0)
      throw contradiction_error(r1,r2);

    instance.removeRelation(r1);
    instance.removeRelation(r2);

    varset_t keep = newrel->getVariables() - diff1 - diff2;
    if (diff1.size() > 1 || diff2.size() > 1)
      newrel = newrel->project(keep, newrel->getId());
    newrel->compress();
    instance.addRelation(newrel);
    }
  catch(const std::length_error &)
    {
    }
  }

void combine_leaf(CSP &instance, size_t maxrel, dualgraph_t::edge_descriptor edge, dualgraph_t &jt)
  {
  graph_traits<dualgraph_t>::adjacency_iterator ait, ait_end;
  property_map<dualgraph_t, vertex_relation_t>::type vertexrel = get(vertex_relation_t(), jt);

  dualgraph_t::vertex_descriptor u,v,v1,v2;

  std::cout << "Leaf Edge " << edge << " of " << num_edges(jt) << "\n";

  Relation *leaf, *nonleaf;

  u = source(edge,jt);
  v = target(edge,jt);

  if (degree(u, jt) == 1)
    {
    v1 = u;
    v2 = v;
    }
  else
    {
    v1 = v;
    v2 = u;
    }

  leaf = vertexrel[v1];
  nonleaf = vertexrel[v2];

  instance.removeRelationXRef(leaf);

  varset_t edge_vars = intersection(leaf->getVariables(), nonleaf->getVariables());

  DUMP(std::cout << "R1 " << *leaf << "\n" << "R2 " << *nonleaf << " edge=";write_set(std::cout, edge_vars) << "\n");

  // Vars we can split from leaf
  varset_t splitvars = leaf->getVariables() - edge_vars;
  if (splitvars.size() > 1)
    {
    std::auto_ptr<Relation> splitrel = leaf->split(splitvars, instance.getNewRelId());
    DUMP(std::cout << "R1 " << *leaf << "\n");
    instance.linkRelations(*leaf, *splitrel);

    leaf->compress();
    splitrel->compress();
    if (splitrel->getNumRows() < 2)
      {
      // We can remove the relation if we update
      // the domains....
      splitrel->identifyUnaryConstraints();
      std::cout << "Split removed....\n";
      }
    else
      {
      DUMP(std::cout << "Add Split " << *splitrel << "\n");
      instance.addRelation(splitrel);
      }
    }

  try
    {
    instance.removeRelationXRef(nonleaf);

    // Create link relation
    std::auto_ptr<Relation> linkrel = leaf->joinCompress(*nonleaf, instance.getNewRelId(), maxrel);
    if (linkrel->getNumRows() == 0)
      throw contradiction_error(leaf,nonleaf);

    linkrel->compress();

    DUMP(std::cout << "Link Joined " << *linkrel << "\n");

    // We remove from linkrel those variables which are used in edges from nonleaf
    // (rip property ensure vars are not used in other relations)
    varset_t keepvars = linkrel->getVariables();
    for(tie(ait, ait_end) = adjacent_vertices(v2, jt); ait != ait_end; ++ait)
      {
      if (*ait != v1)
        {
        varset_t tmp;
        Relation *r = vertexrel[*ait];
        tmp = keepvars - r->getVariables();
        std::swap(tmp, keepvars);
        }
      }
    if (keepvars.size() > 1 && keepvars.size() < linkrel->getVariables().size())
      {
      std::auto_ptr<Relation> newleaf = linkrel->split(keepvars, leaf->getId());
      instance.linkRelations(*newleaf, *linkrel);
      newleaf->compress();
      if (newleaf->getNumRows() == 1)
        {
        // We can remove the relation if we update
        // the domains....
        newleaf->identifyUnaryConstraints();
        std::cout << "Split removed....\n";
        instance.removeRelation(leaf);
        }
      else
        {
        leaf->assign(*newleaf);
        instance.addRelationXRef(leaf);
        DUMP(std::cout << "Final R1 " << *leaf << "\n");
        }
      }
    else
      {
      std::cout << "R1 removed\n";
      instance.removeRelation(leaf);
      }

    nonleaf->assign(*linkrel);
    if (nonleaf->getNumRows() < 20000)
      nonleaf->removeRedundancy();
    }
  catch(const std::length_error &)
    {
    }
  instance.addRelationXRef(nonleaf);

  DUMP(std::cout << "Final R2" << *nonleaf << "\n");
  }      

void uniform_network(CSP &instance, size_t maxrel)
  {
  dualgraph_t jt;
  createDualGraph(instance, jt);

  std::vector<int> component(num_vertices(jt));
  int num = connected_components(jt, &component[0]);

  std::cout << "Dual graph has " << num << " connected components\n";

  create_join_tree(jt);

  check_rip(jt);

  property_map<dualgraph_t, vertex_relation_t>::type  vertexrel   = get(vertex_relation_t(), jt);
  property_map<dualgraph_t, edge_weight_t>::type      weight_map  = get(edge_weight, jt);
  property_map<dualgraph_t, edge_variables_t>::type   vars_map    = get(edge_variables_t(), jt);
  
  graph_traits<dualgraph_t>::edge_iterator eiter, eiter_end, ei_tmp;
  graph_traits<dualgraph_t>::vertex_descriptor u,v;

  while(num_edges(jt) > 0)
    {
    float best_cost = std::numeric_limits<float>::infinity();
    graph_traits<dualgraph_t>::edge_descriptor best_edge;
  
    for (tie(eiter, eiter_end) = edges(jt); eiter != eiter_end; ) 
      {
      u = source(*eiter,jt);
      v = target(*eiter,jt);
      ei_tmp = eiter++;
     
      if (degree(u, jt) == 1 && degree(v,jt) == 1)
        {
        dualgraph_t jt2;
        instance.buildXRef();
        createDualGraph(instance, jt2);
        create_join_tree(jt2);
        //exit(0);
        combine_root(instance, maxrel, *ei_tmp, jt);
        remove_edge(*ei_tmp, jt);
        }
      else if (degree(u, jt) == 1 || degree(v, jt) == 1)
        {
        Relation *r1 = vertexrel[u], *r2 = vertexrel[v];
        //SelectivityCost  cost;
        //JoinSizeCost  cost;
        ADBCost cost;
        float c = cost(r1,r2, best_cost);
        if (c < best_cost)
          {
          best_cost = c;
          best_edge = *ei_tmp;
          }
        }
      }

    if (best_cost < std::numeric_limits<float>::infinity())
      {
      combine_leaf(instance, maxrel, best_edge, jt);
      remove_edge(best_edge, jt);
      }
    }
  instance.buildXRef();
  instance.postprocess();
  }
