/*
 * Copyright (C) 2003 Jeppe Nejsum Madsen, nejsum@diku.dk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include "stdafx.h"

#pragma warning( push )
#pragma warning( disable : 4100 4267 4512 4702)
#include "Arr/arr.h"
#include "Interval/interval.h"
#include "ArrayLib/adbrt.h"
#pragma warning( pop ) 

#include "../column.hpp"
#include "../relation.hpp"
#include "../constraint_network.hpp"
#include "../utility/formatter.hpp"
#undef max

#pragma warning( disable : 4800)
namespace
  {
  Arr adbvars;

  // Extraction functions
  Variable *ADBVar2Variable(Adb &adb, size_t id)
    {
    Arr &var = adb.m_VARS[(int)id];
    Arr &decl = var[VARX_DECL];
    Arr &tsv = adb.m_TSV[(int)id];

    int nodecls = decl.Len();
    cVarTypeType type = adb.GetVarType((int)id);
    Variable *vp;

    switch(type)
      {
      case VTYPE_BOOL:
        {
        ConcreteVariable<bool> *v = new ConcreteVariable<bool>(id, adb.GetVarName((int)id));
        for (int n = 0; n < nodecls; n++)
          {
          v->addDomainValue(decl.GetIntElem(n) != 0);
          }
        vp = v;
        }
        break;
      case VTYPE_INTG:
        {
        ConcreteVariable<int> *v = new ConcreteVariable<int>(id, adb.GetVarName((int)id));
        for (int n = 0; n < nodecls; n++)
          {
          v->addDomainValue(decl.GetIntElem(n));
          }
        vp = v;
        }
        break;
      case VTYPE_REAL:
        {
        ConcreteVariable<double> *v = new ConcreteVariable<double>(id, adb.GetVarName((int)id));
        for (int n = 0; n < nodecls; n++)
          {
          v->addDomainValue(decl.GetDoubleElem(n));
          }
        vp = v;
        }
        break;
      case VTYPE_ENUM:
        {
        ConcreteVariable<std::string> *v = new ConcreteVariable<std::string>(id, adb.GetVarName((int)id));
        for (int n = 0; n < nodecls; n++)
          {
          char  buf[8192];
          decl[n].Arr2String(buf, sizeof(buf));
          v->addDomainValue(buf);
          }
        vp = v;
        }
        break;
      case VTYPE_LINK:
        {
        LinkVariable *v = new LinkVariable(id, adb.GetVarName((int)id), decl.GetIntElem(0));
        vp = v;
        }
        break;
      default:
        std::cout << "Unsupported variable type: " << type << "\n";
        exit(1);
      }
    assert((size_t)tsv.Len() == vp->getDomainSize());
    for (int i = 0; i < tsv.Len(); i++)
      {
      if (tsv.GetIntElem(i) == 0)
        vp->setValueInvalid((size_t)i);
      }
    return vp;
    }

  void extractVariables(Adb &adb, ConstraintNetwork &csp)
    {
    int varcount = adb.NofVars();

    Formatter f("Extracting variables");
    for (int i = 0; i < varcount; i++)
      {
      f.writeProgress(i+1, varcount);
      csp.addVariable(std::auto_ptr<Variable>(ADBVar2Variable(adb, i)));
      }
    }

  Relation *ADBRel2Relation(ConstraintNetwork &csp, Arr &rel, size_t id)
    {
    int offset = rel.Len()-2;
    Arr &decl = rel[offset];
    Arr &vars = rel[offset+1];

    int novars = vars.Len();
    int nodecls = decl[0].Len();
    char expr[4096];

    if (offset > 0)
      rel[RELX_RELTEXT].Arr2String(expr, sizeof(expr));
    else
      strcpy(expr, "Relation Type");

    Relation *r = new Relation(id, expr);

    for(int x = 0; x < novars; ++x)
      {
      Column *c;

      if (decl[x].Type() == T_PNTR)
        {
        CompressedColumn *cc = CompressedColumn::createNew(csp.getVariable(vars.GetIntElem(x)),nodecls);

        for (int n = 0; n < nodecls; n++)
          {
          idvector_t  vals;
          Arr &vs = decl[x][n];
          for (int j = 0; j < vs.Len(); j++)
            vals.push_back(vs.GetIntElem(j));
          cc->addCell(vals); 
          }
        c = cc;
        }
      else
        {
        UncompressedColumn *uc = UncompressedColumn::createNew(csp.getVariable(vars.GetIntElem(x)),nodecls);

        for (int n = 0; n < nodecls; n++)
          {
          uc->addCell(decl[x].GetIntElem(n)); 
          }
        c = uc;
        }
      r->addColumn(std::auto_ptr<Column>(c));
      }
    return r;
    }

  void extractRelations(Adb &adb, ConstraintNetwork &csp)
    {
    int relcount = adb.NofRels();

    Formatter f("Extracting relations");
    for (int i = 0; i < relcount; i++)
      {
      Arr &rel = adb.m_RB[i];
      f.writeProgress(i+1, relcount);
      csp.addRelation(std::auto_ptr<Relation>(ADBRel2Relation(csp, rel, size_t(i))));
      }
    }

  // Insertion functions
  void variable2ADBVar(Variable *var, Arr &arr)
    {
    arr = Vector(T_PNTR, VARX_NOFELEMS);
    arr[VARX_NAME]    = Arr(var->getName().c_str()).Enclose();
    arr[VARX_IOTYPE]  = IO_INPUT;
    arr[VARX_EXPL]    = "";

    ConcreteVariable<bool> *cvi = dynamic_cast< ConcreteVariable<bool> *>(var);
    if (cvi != 0)
      {
      arr[VARX_TYPE]    = VTYPE_BOOL;
      arr[VARX_DECL]    = Vector(T_BOOL, (int)cvi->getDomainSize());
      arr[VARX_DECLEXPL]= Vector(T_PNTR, (int)cvi->getDomainSize());
      for(int n = 0; n < (int)cvi->getDomainSize(); n++)
        {
        arr[VARX_DECL].Set(cvi->getDomainValue(n), n);
        arr[VARX_DECLEXPL][n] = Arr("");
        }
      }
    else
      {
      ConcreteVariable<int> *cvi = dynamic_cast< ConcreteVariable<int> *>(var);
      if (cvi != 0)
        {
        arr[VARX_TYPE]    = VTYPE_INTG;
        arr[VARX_DECL]    = Vector(T_LONG, (int)cvi->getDomainSize());
        arr[VARX_DECLEXPL]= Vector(T_PNTR, (int)cvi->getDomainSize());
        for(int n = 0; n < (int)cvi->getDomainSize(); n++)
          {
          arr[VARX_DECL].Set(cvi->getDomainValue(n), n);
          arr[VARX_DECLEXPL][n] = Arr("");
          }
        }
    else
      {
      ConcreteVariable<double> *cvd = dynamic_cast< ConcreteVariable<double> *>(var);
      if (cvd != 0)
        {
        arr[VARX_TYPE]    = VTYPE_REAL;
        arr[VARX_DECL]    = Vector(T_LONG, (int)cvd->getDomainSize());
        arr[VARX_DECLEXPL]= Vector(T_PNTR, (int)cvd->getDomainSize());
        for(int n = 0; n < (int)cvd->getDomainSize(); n++)
          {
          arr[VARX_DECL].Set(cvd->getDomainValue(n), n);
          arr[VARX_DECLEXPL][n] = Arr("");
          }
        }
      else
        {
        ConcreteVariable<std::string> *cvs = dynamic_cast< ConcreteVariable<std::string> *>(var);
        if (cvs != 0)
          {
          arr[VARX_TYPE]    = VTYPE_ENUM;
          arr[VARX_DECL]    = Vector(T_PNTR, (int)cvs->getDomainSize());
          arr[VARX_DECLEXPL]= Vector(T_PNTR, (int)cvs->getDomainSize());
          for(int n = 0; n < (int)cvs->getDomainSize(); n++)
            {
            arr[VARX_DECL][n] = Arr(cvs->getDomainValue(n).c_str());
            arr[VARX_DECLEXPL][n] = Arr("");
            }
          }
        else
          {
          LinkVariable *lv = dynamic_cast<LinkVariable *>(var);
          if (lv != 0)
            {
            arr[VARX_TYPE]    = VTYPE_LINK;
            arr[VARX_DECL]    = Vector(T_LONG, 1);
            arr[VARX_DECLEXPL]= Vector(T_PNTR, 1);
            arr[VARX_DECL].Set((int)lv->getDomainSize(), 0);
            arr[VARX_DECLEXPL][0] = Arr("");
            }
          else
            {
            std::cout << "Unknown variable type\n";
            exit(1);
            }
          }
        }
      }
      }
    }


  template<typename T>
    void fillCompressedColumn(Arr &arr, CompressedColumn &c, ArrType arr_type)
    {
    for (int x = 0; x < (int)c.getNumCells(); x++)
      {
      idvector_t  values;
      c.expandCell(x, values);
      arr[x] = Vector(arr_type, values.size());
      for(int j = 0; j < (int)values.size(); j++)
        {
        if (values[j] >= c.getVariable().getDomainSize())
          {
          std::string msg("Relations contains invalid index. Var=");
          msg.append(c.getVariable().getName());
          msg.append(", index=");
          msg.append(boost::lexical_cast<std::string>(values[j]));
          msg.append(", domainSize=");
          msg.append(boost::lexical_cast<std::string>(c.getVariable().getDomainSize()));
          throw std::domain_error(msg);
          }
        arr[x].Set((int)values[j],j);
        }
      }
    }

  template<typename T>
    void fillUncompressedColumn(Arr &arr, UncompressedColumn &c, ArrType arr_type)
    {
    arr = Vector(arr_type, (int)c.getNumCells());

    for (int x = 0; x < (int)c.getNumCells(); x++)
      {
      if (c.getCell(x) >= c.getVariable().getDomainSize())
        {
        std::string msg("Relations contains invalid index. Var=");
        msg.append(c.getVariable().getName());
        msg.append(", index=");
        msg.append(boost::lexical_cast<std::string>(c.getCell(x)));
        msg.append(", domainSize=");
        msg.append(boost::lexical_cast<std::string>(c.getVariable().getDomainSize()));
        throw std::domain_error(msg);
        }
      arr.Set((int)c.getCell(x),x);
      }
    }

  void relation2ADBRel(Relation &rel, Arr &arr, int id)
    {
    arr = Vector(T_PNTR, RELX_NOFELEMS);

    arr[RELX_TYPE]    = REL_SYMBOLIC;
    arr[RELX_RELTEXT] = rel.getExpression().c_str();
    arr[RELX_EXPL]    = boost::lexical_cast<std::string>(id).c_str();;	

    arr[RELX_DECL]    = Vector(T_PNTR, (int)rel.getNumColumns());
    for (int n = 0; n < (int)rel.getNumColumns(); n++)
      {
      Column *col = rel.getColumns()[n];

      size_t max_dom_idx = col->getVariable().getDomainSize()-1;

      if (col->isCompressed())
        {
        arr[RELX_DECL][n] = Vector(T_PNTR, (int)rel.getNumRows());
        CompressedColumn *c = dynamic_cast<CompressedColumn *>(col);
        if (c == 0)
          INTERNAL_ERROR;

        if (max_dom_idx <= 1)
          {
          fillCompressedColumn<bool>(arr[RELX_DECL][n], *c, T_BOOL);
          }
        else if (max_dom_idx <= (size_t)std::numeric_limits<char>::max())
          {
          fillCompressedColumn<int>(arr[RELX_DECL][n], *c, T_BYTE);
          }
        else if (max_dom_idx <= (size_t)std::numeric_limits<short>::max())
          {
          fillCompressedColumn<short>(arr[RELX_DECL][n], *c, T_SHRT);
          }
        else
          {
          fillCompressedColumn<long>(arr[RELX_DECL][n], *c, T_LONG);
          }
        }
      else
        {
        UncompressedColumn *c = dynamic_cast<UncompressedColumn *>(col);
        if (c == 0)
          INTERNAL_ERROR;

        if (max_dom_idx <= 1)
          {
          fillUncompressedColumn<bool>(arr[RELX_DECL][n], *c, T_BOOL);
          }
        else if (max_dom_idx <= (size_t)std::numeric_limits<char>::max())
          {
          fillUncompressedColumn<int>(arr[RELX_DECL][n], *c, T_BYTE);
          }
        else if (max_dom_idx <= (size_t)std::numeric_limits<short>::max())
          {
          fillUncompressedColumn<short>(arr[RELX_DECL][n], *c, T_SHRT);
          }
        else
          {
          fillUncompressedColumn<long>(arr[RELX_DECL][n], *c, T_LONG);
          }
        }
      }

    arr[RELX_PS]      = Vector(T_SHRT, (int)rel.getNumColumns());
    for (int x = 0; x < (int)rel.getNumColumns(); x++)
      arr[RELX_PS].Set((short)rel.getColumns()[x]->getVariable().getId(), x);
    }

  void addTSV(Adb &adb, const ConstraintNetwork &csp)
    {
    adb.m_TSV = Vector(T_PNTR, (int)csp.getNumVariables());

    for (int i = 0; i < (int)csp.getNumVariables(); i++)
      {
      Variable &v = csp.getVariable(i);
      size_t    domsize = v.getDomainSize();

      adb.m_TSV[i] = Vector(T_BOOL, (int)domsize);
      for (int j = 0; j < (int)domsize; j++)
        {
        adb.m_TSV[i].SetInt(v.isValueValid((size_t)j),j);
        }
      }
    }

  void addVariables(Adb &adb, const ConstraintNetwork &csp)
    {
    Arr vars = Vector(T_PNTR, (int)csp.getNumVariables());

    Formatter f("Adding variables");
    for (int i = 0; i < (int)csp.getNumVariables(); i++)
      {
      f.writeProgress(i+1, (int)csp.getNumVariables());
      Arr v;
      Variable &var = csp.getVariable(i);
      variable2ADBVar(&var, v);
      if (i < adbvars.Len())
        v[VARX_NAME] = adbvars[i][VARX_NAME];
      else
        {
        Arr name;
        name = var.getName().c_str();
        v[VARX_NAME] = name.Enclose();
        }
      vars[i] = v;    
      }
    adb.m_VARS = vars;
    }

  void addRelations(Adb &adb, const ConstraintNetwork &csp)
    {
    Formatter f("Adding relations");
    const ConstraintNetwork::relcol_t &rels = csp.getRelations();
    const ConstraintNetwork::relcol_t &isolated_rels = csp.getIsolatedRelations();
    int numRels = rels.size() + isolated_rels.size();

    adb.m_RB = Vector(T_PNTR, numRels);
    ConstraintNetwork::relcol_t::const_iterator it = rels.begin();
    int i = 0;
    while (it != rels.end())
      {
      Relation *rel = *it++;
      f.writeProgress(i+1, numRels, rel->getNumRows());
      relation2ADBRel(*rel, adb.m_RB[i], i);
      i++;
      }
    it = isolated_rels.begin();
    while (it != isolated_rels.end())
      {
      Relation *rel = *it++;
      f.writeProgress(i+1, numRels, rel->getNumRows());
      relation2ADBRel(*rel, adb.m_RB[i],i);
      i++;
      }
    }
  }

namespace ADB
  {
  void importADBFile(ConstraintNetwork &csp, const std::string &adbfile)
    {
    GenericAdbClass adb;
    std::cout << "Loading adb file " << adbfile << "\n";
    adb.Load(adbfile.c_str());
    adbvars = adb.m_VARS;
    extractVariables(adb, csp);
    extractRelations(adb, csp);
    std::cout << "Load finished\n";
    };

  void exportADBFile(const ConstraintNetwork &csp, const std::string &adbfile)
    {
    GenericAdbClass adb;
    std::cout << "Saving adb file " << adbfile << "\n";
    addVariables(adb, csp);
    addRelations(adb, csp);
    addTSV(adb,csp);
    adb.m_INFOS = Vector(T_BOOL, 0);
    adb.m_CUTSETS = Vector(T_BOOL, 0);
    adb.m_VT = Vector(T_BOOL, 0);  
    adb.m_RT = Vector(T_BOOL, 0);
    adb.m_CompleteLevel = COMPL_STATEPROP;
    std::cout << "Writing adb file\n";
    adb.Save(adbfile.c_str());
    std::cout << "Save finished\n";
    }
  }
