/*
 * Copyright (C) 2003 Jeppe Nejsum Madsen, nejsum@diku.dk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include "stdafx.h"
#include "variable.hpp"
#include "detail/column_impl.hpp"

// Create a new ConcreteUncompressedColumn<T> with the T being the
// smallest type large enough to contain all the domain indicies of
// the variable var. If rowCount > 0, it specifies the initial number
// of rows allocated in the new column
UncompressedColumn *UncompressedColumn::createNew(Variable &var, size_t rowCount)
  {
  size_t max_dom_index = var.getDomainSize()-1;

  if (max_dom_index < 8)
    return new ConcreteUncompressedColumn<unsigned char,  BitsetCellTraits<NewBitset<8> > >(var, rowCount);

  if (max_dom_index < 16)
    return new ConcreteUncompressedColumn<unsigned char,  BitsetCellTraits<NewBitset<16> > >(var, rowCount);

  if (max_dom_index < 32)
    return new ConcreteUncompressedColumn<unsigned char,  BitsetCellTraits<NewBitset<32> > >(var, rowCount);
  if (max_dom_index < 64)
    return new ConcreteUncompressedColumn<unsigned char,  BitsetCellTraits<NewBitset<64> > >(var, rowCount);

  if (max_dom_index < 128)
    return new ConcreteUncompressedColumn<unsigned char,  BitsetCellTraits<NewBitset<128> > >(var, rowCount);

  if (max_dom_index < 256)
    return new ConcreteUncompressedColumn<unsigned char,  BitsetCellTraits<NewBitset<256> > >(var, rowCount);

  if (max_dom_index <= std::numeric_limits<unsigned char>::max())
    return new ConcreteUncompressedColumn<unsigned char>(var, rowCount);

  if (max_dom_index <= std::numeric_limits<unsigned short>::max())
    return new ConcreteUncompressedColumn<unsigned short>(var, rowCount);

  if (max_dom_index <= std::numeric_limits<unsigned int>::max())
    return new ConcreteUncompressedColumn<unsigned int>(var, rowCount);
  return new ConcreteUncompressedColumn<unsigned long>(var, rowCount);
  }

// Create a new CompressedColumn from a urvector_t. If the domain of
// the variable is within certain limits, a BitmapCompressedColumn is
// created, otherwise a ConcreteCompressedColumn.
CompressedColumn *CompressedColumn::createNew(Variable &var, const urvector_t &ur)
  {
  size_t max_dom_index = var.getDomainSize()-1;
  if (max_dom_index < 8)
    return new BitmapCompressedColumn<8>(var, ur);

  if (max_dom_index < 16)
    return new BitmapCompressedColumn<16>(var, ur);

  if (max_dom_index < 32)
    return new BitmapCompressedColumn<32>(var, ur);

  if (max_dom_index < 64)
    return new BitmapCompressedColumn<64>(var, ur);
  
  if (max_dom_index < 128)
    return new BitmapCompressedColumn<128>(var, ur);
  
  if (max_dom_index < 256)
    return new BitmapCompressedColumn<256>(var, ur);
  if (max_dom_index <= std::numeric_limits<unsigned char>::max())
    return new ConcreteCompressedColumn<unsigned char>(var, ur);

  if (max_dom_index <= std::numeric_limits<unsigned short>::max())
    return new ConcreteCompressedColumn<unsigned short>(var, ur);

  if (max_dom_index <= std::numeric_limits<unsigned int>::max())
    return new ConcreteCompressedColumn<unsigned int>(var, ur);

  return new ConcreteCompressedColumn<unsigned long>(var, ur);
  }

CompressedColumn *CompressedColumn::createNew(Variable &var, size_t rowCount)
  {
  size_t max_dom_index = var.getDomainSize()-1;

  if (max_dom_index < 8)
    return new BitmapCompressedColumn<8>(var, rowCount);

  if (max_dom_index < 16)
    return new BitmapCompressedColumn<16>(var, rowCount);

  if (max_dom_index < 32)
    return new BitmapCompressedColumn<32>(var, rowCount);

  if (max_dom_index < 64)
    return new BitmapCompressedColumn<64>(var, rowCount);
  
  if (max_dom_index < 128)
    return new BitmapCompressedColumn<128>(var, rowCount);
  
  if (max_dom_index < 256)
    return new BitmapCompressedColumn<256>(var, rowCount);

    if (max_dom_index <= std::numeric_limits<unsigned char>::max())
    return new ConcreteCompressedColumn<unsigned char>(var, rowCount);

  if (max_dom_index <= std::numeric_limits<unsigned short>::max())
    return new ConcreteCompressedColumn<unsigned short>(var, rowCount);

  if (max_dom_index <= std::numeric_limits<unsigned int>::max())
    return new ConcreteCompressedColumn<unsigned int>(var, rowCount);

  return new ConcreteCompressedColumn<unsigned long>(var, rowCount);
  }

std::ostream &operator<<(std::ostream &s, const Column *c)
  {
  return s << c->getVariable().getName();
  }

std::ostream &operator<<(std::ostream &s, const colpair_t &v)
  {
  return s << "(" << (v.first ? v.first->getVariable().getName() : "null") 
           << "," << (v.second ? v.second->getVariable().getName() : "null") << ")";
  }

std::ostream &operator<<(std::ostream &s, const joinindex_t &ji)
  {
  joinindex_t::const_iterator it = ji.begin(), pend = ji.end();
  while (it != pend)
    {
    s << "(" << it->first << "," << it->second << ")\n";
    ++it;
    }
  return s;
  }
