/*
 * Copyright (C) 2003 Jeppe Nejsum Madsen, nejsum@diku.dk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#ifndef COLUMN_H
#define COLUMN_H

#include "variable.hpp"

class UniqueRows;
class UniqueComplement;

// Elements of UniqueRows
// first is the index of the first unique complement row
// second is a vector contianing the union of domain values for rows 
// the having same complement
typedef std::pair<size_t, idvector_t *> urelement_t;
typedef std::vector<urelement_t>        urvector_t;

inline bool operator<(const urelement_t &lhs, const urelement_t &rhs)
  {
  return lhs.first < rhs.first;
  }

// Join index is a set of pair of row numbers which must be 
// joined in the final result
typedef std::deque<idpair_t > joinindex_t;

// Base class for columns
class Column : boost::noncopyable 
  {
  protected:
    Column(Variable &var, bool compressed) 
      : m_Variable(var), m_Compressed(compressed) {};

  public:
    virtual ~Column() {};

    const Variable &getVariable() const   {return m_Variable;}
    Variable       &getVariable()         {return m_Variable;}
    bool            isCompressed() const  {return m_Compressed;}
    void            setIndex(size_t index){m_Index = index;}

    // Virtuals
    virtual void addCellCopy(size_t copyFromIndex) = 0;
  
    // Return number of cells
    virtual size_t  getNumCells() const = 0;

    // Return number of scalars
    virtual size_t  getSize() const = 0;

    // Remove cells which are not listed in the urvector
    virtual void    pruneCells(urvector_t::const_iterator start, urvector_t::const_iterator pend) = 0;

    // Remap indicies to values in new_ids
    virtual void    remapIndicies(const idvector_t &new_ids) = 0;


    virtual void    updateContainmentMask(size_t row, const Column &c, std::vector<bool> &mask) = 0;

    virtual bool    findSupport(const Column &c, std::vector<bool> &mask) = 0;
    virtual void    updateTupleMask(size_t row, const Column &c, std::vector<bool> &mask) = 0;
    
    // xor list of hashvalues with hashvalues for this column
    virtual void    xorHashValues(size_t *hashValues) const = 0;

    // Create the UniqueRows structure for the complement of this column
    virtual void    fillInUniqueRows(UniqueRows &ur) const = 0;

    // Create the UniqueComplement structure for the complement of this column
    virtual void    fillInUniqueComplement(UniqueComplement &ur) const = 0;

    // Compare two cells for equality
    virtual bool    cellsAreEqual(size_t row1, size_t row2) const = 0;

    // Return contents of cell as string
    virtual std::string cellToString(size_t row) const = 0;

    // Create a joinindex based on join between *this and col
    virtual bool    createJoinIndex(joinindex_t &ji, const Column &col, size_t &start_row, size_t max_entries) const = 0;

    // Revise (prune pairs) the joinindex based on join between *this and col
    virtual void    reviseJoinIndex(joinindex_t &ji, const Column &col, size_t start = 0) const = 0;

    // Create new column from *this and col, picking cells listed in joinindex
    virtual Column *createFromJoinIndex(const joinindex_t &ji, const Column &col) const = 0;

    // Clone *this by picking cells listed in joinindex
    virtual Column *cloneFromJoinIndex(const joinindex_t &ji, bool useFirst) const = 0;

    // Clone *this by replicating all cells numNewCells/getNumCells() times
    virtual Column *clone(size_t numNewCells = 0, bool interleave = false) const = 0;

    virtual void    createUniqueMask(const boolvector_t &valid_cell, boolvector_t &mask) const = 0;

    virtual bool    updateValidRowMask(std::vector<bool> &mask) const = 0;
    virtual bool    revise(std::vector<bool> &mask) = 0;
    virtual void    updateRowCounts(size_t start_row, std::vector<double> &counts, double &intersection_count) const = 0;
    virtual void    updateRowCounts(std::vector<double> &counts) const = 0;
    virtual bool findInitialSupport(const Column &col, boolvector_t &rvalid, boolvector_t &svalid, 
                                    std::vector<int> &rsupport, std::vector<size_t> &rsupport_cols, int colcount) = 0;

  protected:
    bool               m_Compressed;
    size_t             m_Index;       // Used for hash table lookup
    Variable          &m_Variable;
  };

typedef std::vector<Column *>         colvector_t;
typedef std::pair<Column *, Column *> colpair_t;
typedef std::vector<colpair_t>        colpairvector_t;

class UncompressedColumn : public Column
  {
  public:
    UncompressedColumn(Variable &var)
      : Column(var,false) {}

    static UncompressedColumn *createNew(Variable &var, size_t rowCount = 0);

    // I/O functions
    virtual void    addCell(size_t val) = 0;
    virtual void    setCell(size_t cell, size_t val) = 0;
    virtual size_t  getCell(size_t index) const = 0;
  };

class CompressedColumn : public Column
  {
  public:
    CompressedColumn(Variable &var)
      : Column(var,true) {}

    static CompressedColumn *createNew(Variable &var, size_t rowCount = 0);
    static CompressedColumn *createNew(Variable &var, const urvector_t &ur);

    virtual void addCell(const idvector_t &values) = 0;
    virtual void expandCell(size_t row, idvector_t &values) const = 0;
  };

// Equality for column ptrs based on variable id
struct ColumnPtrVariableEquality
  {
  bool operator()(const Column *lhs, const Column *rhs)
    {
    return lhs->getVariable() == rhs->getVariable();
    }
  };

// Ordering for columns ptrs, based on variable id
struct ColumnPtrVariableOrder
  {
  typedef Column * colptr_t;
  bool operator()(const colptr_t &c1, const colptr_t &c2)
    {
    return c1->getVariable().getId() < c2->getVariable().getId();
    }
  };


std::ostream &operator<<(std::ostream &s, const Column *c);
std::ostream &operator<<(std::ostream &s, const colpair_t &v);
std::ostream &operator<<(std::ostream &s, const joinindex_t &ji);
#endif
