/*
 * Copyright (C) 2003 Jeppe Nejsum Madsen, nejsum@diku.dk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include "stdafx.h"
#include "csplib/uniquerows.hpp"
#include "csplib/relation.hpp"
#include "csplib/utility/formatter.hpp"

namespace 
{

// Utility types & structs to order columns 
// based on unique tuples in complement
typedef std::pair<size_t, colvector_t::const_iterator>  orderpair_t;
typedef std::vector<orderpair_t>                        comporder_t;
struct ColumnOrdering
  {
  bool operator()(const orderpair_t &p1, const orderpair_t &p2) 
    {
    // If same size, use varid to order
    if (p1.first == p2.first)
      return (*(p1.second))->getVariable() < (*(p2.second))->getVariable();

    return p1.first < p2.first;
    }
  };

// Create the vector order, which contains columns in the order in
// which tey should be compressed
void determineCompressOrder(colvector_t &cols, comporder_t &order, const varset_t *vars_to_compress)
  {
  colvector_t::const_iterator it = cols.begin(), pend = cols.end();
  HashValues                  vals(cols, it);
 
  Formatter f("Compression order");
  int n = 1;
  while(it != pend)
    {
    Column *column = *it;

    if (vars_to_compress != 0 && vars_to_compress->find(&column->getVariable()) == vars_to_compress->end())
      {
      ++it;
      continue;
      }

    f.writeProgress(n++, vars_to_compress ? vars_to_compress->size() : cols.size());
    UniqueRowCount  uc(cols, it, vals);
    
    if (uc.size() < column->getNumCells())
      order.push_back(std::make_pair(uc.size(), it));

    vals.xorColumn(*column);
    if (++it != pend)
      vals.xorColumn(*(*it));
    }
  std::sort(order.begin(), order.end(), ColumnOrdering());
  }

// Compress relation on the specified column 
bool compressColumn(colvector_t &columns, colvector_t::const_iterator compress_col)
  {
  urvector_t urvector;

  UniqueComplement urs(columns, compress_col);

  (*compress_col)->fillInUniqueComplement(urs);

  // Don't compress if there's no benefit
  if (urs.size() >= (*compress_col)->getNumCells())
    return false;
  // Extract info into sorted vector
  UniqueComplement::const_iterator it = urs.begin(), pend = urs.end();
  while (it != pend)
    {
    urvector.push_back(*it);
    it++;
    }
    
  std::sort(urvector.begin(), urvector.end());

  // Create the compressed column from the unique rows
  CompressedColumn *newCol = CompressedColumn::createNew((*compress_col)->getVariable(), urvector);

  // Remove the non-unique tuples of the remaining columns
  colvector_t::iterator colit = columns.begin(), colpend = columns.end();
  while(colit != colpend)
    {
    if (colit == compress_col)
      {
      // Replace the compressed column
      boost::checked_delete(*colit);
      *colit = newCol;
      }
    else
      (*colit)->pruneCells(urvector.begin(), urvector.end());
  
    ++colit;
    }
  return true;
  }

// Uncompress a single row
// begin, end are begin()/end() from new uncompressed columns
// new_col is the current new uncompressed column
// compressed_col is the current column being uncompressed,
// startRow is the row where values should be copied from
// row is the row in comppressed_col to uncompress
// copyPrevious indicates whether 
void uncompressRow(bool uncompress_link, colvector_t::iterator begin, colvector_t::iterator end, 
                   colvector_t::iterator new_col, colvector_t::iterator compressed_col,
                   size_t startRow, size_t row )
  {
  // Terminate recursion
  if (new_col == end)
    return;

  Variable &v = (*compressed_col)->getVariable();
  
  if (!(*compressed_col)->isCompressed() || (v.isLinkVar() && !uncompress_link))
    {
    if ((*compressed_col)->isCompressed())
      {
      idvector_t cellValues;
      CompressedColumn *cc = dynamic_cast<CompressedColumn *>(*compressed_col);
      CompressedColumn *c = dynamic_cast<CompressedColumn *>(*new_col);
      cc->expandCell(row, cellValues);
      c->addCell(cellValues);
      }
    else
      {
      UncompressedColumn *uc = dynamic_cast<UncompressedColumn *>(*compressed_col);
      UncompressedColumn *c = dynamic_cast<UncompressedColumn *>(*new_col);
      c->addCell(uc->getCell(row));
      }

    uncompressRow(uncompress_link, begin, end, boost::next(new_col), boost::next(compressed_col), 
                  (*new_col)->getNumCells()-1, row);

    }
  else
    {
    idvector_t cellValues;
    CompressedColumn *cc = dynamic_cast<CompressedColumn *>(*compressed_col);
    cc->expandCell(row, cellValues);
    
    idvector_t::iterator value;
    bool copyPrevious = false; // Don't copy previus columns for first value
    for(value = cellValues.begin(); value != cellValues.end(); ++value)
      {
      // Copy values from previous columns
      colvector_t::iterator it = begin;

      while(copyPrevious && it != new_col)
        {
        UncompressedColumn *c = dynamic_cast<UncompressedColumn *>(*it);
        assert(c);
        c->addCellCopy(startRow);
        ++it;
        }

      // Copy value from this column
      UncompressedColumn *c = dynamic_cast<UncompressedColumn *>(*new_col);
      assert(c);
      c->addCell(*value);

      // Insert values from following columns
      uncompressRow(uncompress_link, begin, end, boost::next(new_col), boost::next(compressed_col), 
                    c->getNumCells()-1, row);
      copyPrevious = true;
      }
    }
  }
}


// Compress the relation
bool Relation::compress(const varset_t *vars_to_compress)
  {
  bool        compressed = false;
  comporder_t compressOrder;

  determineCompressOrder(m_Columns, compressOrder, vars_to_compress);

  comporder_t::const_iterator it = compressOrder.begin(), pend = compressOrder.end();

  if (it != pend && it->first < getNumRows())
    {
    // Only compress if there's a benefit
    Formatter f("Compressing");
    int n = 1;
    while (it != pend)
      {
      if (compressColumn(m_Columns, it->second))
        compressed = true;
      f.writeProgress(n++, getNumColumns(), getNumRows());
      ++it;
      }
    }
  
  if (compressed)
    rowCountChanged();

  return compressed;
  }

// Uncompress the relation
bool Relation::uncompress(bool uncompress_link)
  {
  bool compressed = false;

  for(size_t i = 0; i < getNumColumns(); ++i)
    compressed |= m_Columns[i]->isCompressed();

  if(!compressed)
    return false;

  colvector_t newCols;

  for(size_t i = 0; i < getNumColumns(); ++i)
    {
    Variable &v = m_Columns[i]->getVariable();
    if (!uncompress_link && v.isLinkVar() && m_Columns[i]->isCompressed())
      newCols.push_back(CompressedColumn::createNew(v));
    else
      newCols.push_back(UncompressedColumn::createNew(v));
    }
      
  colvector_t::iterator it = m_Columns.begin();

  Formatter f("Uncompressing");

  for(size_t row = 0; row < getNumRows(); ++row, false)
    {
    if ((row+1) % 10 == 0)
      f.writeProgress(row+1, getNumRows());

    uncompressRow(uncompress_link, newCols.begin(), newCols.end(), newCols.begin(), it, 0, row);
    }
  
  for_all(m_Columns, delete_element<Column>());
  m_Columns = newCols;
  rowCountChanged();
  return true;
  }
