/*
 * Copyright (C) 2003 Jeppe Nejsum Madsen, nejsum@diku.dk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include "stdafx.h"
#include "boost/graph/undirected_dfs.hpp"
#include "constraint_network.hpp"
#include "graph.hpp"

using namespace boost;

std::vector<int> &Last::last(const Relation &r, const Relation &s)
  {
  idpair_t key = std::make_pair(r.getId(), s.getId());

  lastmap_t::iterator it = m_LastMap.find(key);
  if (it == m_LastMap.end())
    {
    m_LastMap[key] = std::vector<int>(r.getNumRows(), -1);
    return m_LastMap[key];
    }
  else
    return it->second;
  }

void common_columns(const Relation &r, const Relation &s, colpairvector_t &common_columns)
  {
  colvector_t::const_iterator it1 = r.getColumns().begin() , it2 = s.getColumns().begin();
  colvector_t::const_iterator pend1 = r.getColumns().end() , pend2 = s.getColumns().end();

  while (it1 != pend1 && it2 != pend2)
    {
    size_t id1 = (*it1)->getVariable().getId();
    size_t id2 = (*it2)->getVariable().getId();
  
    if (id1 < id2)
      ++it1;
    else if (id1 > id2)
      ++it2;
    else
      common_columns.push_back(std::make_pair(*it1++, *it2++));
    }
  }

// Do a fast revise by scanning each column
bool weak_revise(Relation &r, Relation &s)
  {
  colpairvector_t ccols;
  bool            changed = false;

  common_columns(r,s,ccols);
  colpairvector_t::iterator it, it_end = ccols.end();

  for(it = ccols.begin(); it != it_end; ++it)
    {
    Column *c1 = it->first, *c2 = it->second;
    changed |= c1->findSupport(*c2, r.getValidRows());
    }
  return changed;
  }

bool weak_propagate(ConstraintNetwork &instance, relset_t Q, relset_t &changed_relations)
  {
  bool  changed = false;

  while(!Q.empty())
    {
    Relation *s = *Q.begin();
    Q.erase(Q.begin());

    relset_t rs(instance.getConnectedRelations(*s));
    relset_t::iterator it = rs.begin(), it_end = rs.end();
    for(; it != it_end; ++it)
      {
      Relation *r = *it;
      if (*s != *r && weak_revise(*r, *s))
        {
        Q.insert(r);
        changed_relations.insert(r);
        changed = true;
        }
      }
    }
  return changed;
  }

bool weak_arc_consistency(ConstraintNetwork &instance, const relset_t &Q, relset_t *changed_relations)
  {
  std::cout << "Running AC...\n";
  relset_t  local_changed;

  if (changed_relations == 0)
    changed_relations = &local_changed;

  if (weak_propagate(instance, Q, *changed_relations))
    {
    relset_t::iterator it = changed_relations->begin(), it_end = changed_relations->end();
    for(;it != it_end; ++it)
      {
      Relation *r = *it;
      r->pruneInvalidTuples();
      if (r->getNumRows() == 0)
        throw contradiction_error(r,r);
      }
    }
  std::cout << "AC done\n";
  return !changed_relations->empty();
  }

bool weak_arc_consistency(ConstraintNetwork &instance, relset_t *changed_rels)
  {
  relset_t Q;

  Q.insert(instance.getRelations().begin(), instance.getRelations().end());
  return weak_arc_consistency(instance, Q, changed_rels);
  }

bool weak_arc_consistency(ConstraintNetwork &instance, Relation &r, relset_t *changed_rels)
  {
  relset_t Q;

  Q.insert(&r);
  return weak_arc_consistency(instance, Q, changed_rels);
  }


bool node_consistency(ConstraintNetwork &instance, relset_t *revised_relations)
  {
  varset_t  bounded_variables;
  bool      res; 

  instance.findNewBoundVariables(bounded_variables);

  if (bounded_variables.empty())
    res = false;
  else
    res = node_consistency(instance, bounded_variables, revised_relations);
  return res;
  }

bool node_consistency(ConstraintNetwork &instance, const varset_t &changed_variables, relset_t *revised_relations)
  {
  bool      changed = false;
  varset_t  vars_to_check;
  relset_t  rels_to_check;

  std::cout << "Running NC...\n";

  varset_t::const_iterator varit = changed_variables.begin(), varpend = changed_variables.end();
  while(varit != varpend)
    {
    relset_t& varrels = instance.getRelations((*varit)->getId());
    std::copy(varrels.begin(), varrels.end(), 
              std::insert_iterator<relset_t>(rels_to_check, rels_to_check.begin()));
    ++varit;
    }

  // rels_to_check now contains all relations where node consistency should be enforced
  relset_t::iterator relit = rels_to_check.begin(), relpend = rels_to_check.end();
  while(relit != relpend)
    {
    Relation *r = *relit;

    if (r->revise())
      {
      changed = true;
      r->pruneInvalidTuples();

      if (r->getNumRows() == 0)
        throw contradiction_error(r,r);

      if (revised_relations != 0)
        revised_relations->insert(r);
      
      // Relation has changed, schedule new bound vars for check
      r->identifyUnaryConstraints(&vars_to_check);
      }

    relit++;
    }

  if(!vars_to_check.empty())
    node_consistency(instance, vars_to_check, revised_relations);

  std::cout << "NC Done...\n";

  return changed;
  }

// Graph visitor for DAC
template<typename VertexColorMap>
struct DAC_visitor: public boost::dfs_visitor<>
  {
  DAC_visitor(property_map<dualgraph_t, vertex_relation_t>::type &pmap,
              size_t maxrel,
              const VertexColorMap &cmap) 
    : vertexrel(pmap), max_rel_size(maxrel), vertex_color(cmap) { }

  template <class Edge, class Graph>
  void tree_edge(Edge e, Graph &g) 
    {
    Relation *parent = vertexrel[source(e,g)];
    Relation *child  = vertexrel[target(e,g)];

    try
      {
      if (degree(target(e,g),g) == 1)
        {
        std::cout << "Leaf Tree Edge parent=" << parent->getId() << " child=" << child->getId() << "\n";
        std::auto_ptr<Relation> newrel = child->joinCompress(*parent, 42, max_rel_size);

        if (newrel->getNumRows() == 0)
          throw contradiction_error(parent, child);

        child->assign(*newrel->project(child->getVariables(), child->getId()));
        child->removeRedundancy();
        child->compress();
        parent->assign(*newrel->project(parent->getVariables(), parent->getId()));
        parent->removeRedundancy();
        parent->compress();
        }
      else
        {
        std::cout << "NonLeaf Tree Edge parent=" << parent->getId() << " child=" << child->getId() << "\n";
        std::auto_ptr<Relation> newchild = child->joinCompress(*parent, child->getId(), max_rel_size, parent->getVariables() - child->getVariables());
        if (newchild->getNumRows() == 0)
          throw contradiction_error(parent, child);
        child->assign(*newchild.release());
        }
      }
    catch(std::length_error &)
      {
      // Ignore length errors, just continue with next edge...
      }
    }

  template <class Vertex, class Graph>
    void finish_vertex(Vertex u, Graph &g)
    {
    typedef typename property_traits<VertexColorMap>::value_type ColorValue;
    typedef color_traits<ColorValue> Color;
    typename graph_traits<Graph>::adjacency_iterator it, it_end;

    if (degree(u, g) <= 1)
      return;

    Relation *child  = vertexrel[u];

    for(tie(it,it_end) = adjacent_vertices(u, g); it != it_end; ++it)
      {
      if (get(vertex_color, *it) == Color::gray())
        {
        Relation *parent = vertexrel[*it];
        try
          {
          std::cout << "Back Edge parent=" << parent->getId() << " child=" << child->getId() << "\n";
          std::auto_ptr<Relation> newparent = child->joinCompress(*parent, parent->getId(), max_rel_size, child->getVariables() - parent->getVariables());
          if (newparent->getNumRows() == 0)
            throw contradiction_error(parent, child);

          parent->assign(*newparent.release());
          }
        catch(std::length_error &)
          {
          // Ignore length errors, just continue with next edge...
          }
        }
      }
    }

  protected:
    property_map<dualgraph_t, vertex_relation_t>::type &vertexrel;
    const VertexColorMap &vertex_color;
    size_t          max_rel_size;
  };

bool directional_arc_consistency(dualgraph_t &dg, size_t max_rel_size, Relation *root)
  {
  bool res = true;

  dualgraph_t::vertex_iterator vit, vit_end;
  property_map<dualgraph_t, vertex_relation_t>::type vertexrel = get(vertex_relation_t(), dg);

  for(tie(vit,vit_end) = vertices(dg); vit != vit_end; ++vit)
    {
    Relation *r = vertexrel[*vit];
    if (root == 0 || *r == *root)
      {
      const property_map<dualgraph_t, vertex_color_t>::type &cmap = get(boost::vertex_color, dg);

      DAC_visitor<property_map<dualgraph_t, vertex_color_t>::type> vis(vertexrel, 
                                                                       max_rel_size,
                                                                       cmap);
      boost::undirected_dfs(dg, vis, 
                            get(boost::vertex_color, dg),
                            get(boost::edge_color, dg), 
                            *vit);
      break;
      }
    }
  return true;
  }
