/*
 * Copyright (C) 2003 Jeppe Nejsum Madsen, nejsum@diku.dk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include "stdafx.h"
#include "boost/bind.hpp"
#include "boost/mem_fn.hpp"
#include "constraint_network.hpp"

ConstraintNetwork::ConstraintNetwork(std::string name)
 : m_Name(name), m_NextRelId(100000)
  {
  }

ConstraintNetwork::~ConstraintNetwork()
  {
  for_all(m_Rels, delete_element<Relation>());
  for_all(m_IsolatedRels, delete_element<Relation>());
  for_all(m_Vars, delete_element<Variable>());
  }

void ConstraintNetwork::addVariable(std::auto_ptr<Variable> var)
  {
  Variable *v = var.release();
  
  m_VarRels.resize(v->getId()+1);
  m_Vars.push_back(v);
  }

void ConstraintNetwork::addRelationXRef(Relation *rel)
  {
  colvector_t::const_iterator it = rel->getColumns().begin(), pend = rel->getColumns().end();
  while(it != pend)
    {
    Column *col = *it;
    m_VarRels[col->getVariable().getId()].insert(rel);
    ++it;
    }
  }

void ConstraintNetwork::buildXRef()
  {
  m_VarRels.clear();

  m_VarRels.resize(getNumVariables());

  relcol_t::iterator it = m_Rels.begin(), pend = m_Rels.end();

  while(it != pend)
    {
    addRelationXRef(*it);
    ++it;
    }
  }

void ConstraintNetwork::checkXRef() const
  {
  for(size_t i = 0; i < m_VarRels.size(); i++)
    {
    const relset_t &rels = m_VarRels[i];
    relset_t::const_iterator it = rels.begin(), pend = rels.end();
    while(it != pend)
      {
      const Relation *r = *it;
      try
        {
        r->getColumn(*m_Vars[i]);
        }
      catch(...)
        {
        std::cout << "Var " << m_Vars[i]->getName() << " claims to reference relid " << r->getId() << "\n";
        }
      ++it;
      }
    }

  relcol_t::const_iterator it = m_Rels.begin(), pend = m_Rels.end();
  while(it != pend)
    {
    const Relation *r = *it++;
    for(size_t j = 0; j < r->getNumColumns(); ++j)
      {
      const Variable &var = r->getVariable(j);
      const relset_t &varrels = getRelations(var.getId());
      Relation *r2 = (Relation *)r;
      relset_t::iterator rit = varrels.find(r2);
      if (rit == varrels.end())
        std::cout << "Var " << var.getName() << " doesn't reference relid " << r->getId() << "\n";
      }
    }
  }

void ConstraintNetwork::removeRelationXRef(Relation *rel)
  {
  colvector_t::const_iterator it = rel->getColumns().begin(), pend = rel->getColumns().end();
  while(it != pend)
    {
    Column *col = *it;
    m_VarRels[col->getVariable().getId()].erase(rel);
    ++it;
    }
  }

size_t ConstraintNetwork::getNewRelId() 
  {
  return m_NextRelId++;
  }

void ConstraintNetwork::addRelation(std::auto_ptr<Relation> rel, bool updateXref)
  {
  Relation *r = rel.release();

  if (updateXref)
    addRelationXRef(r);

  m_Rels.push_back(r);
  }

std::auto_ptr<Relation> ConstraintNetwork::removeRelation(Relation *r, bool updateXref)
  {
  if (updateXref)
    removeRelationXRef(r);

  relcol_t::iterator it = std::find(m_Rels.begin(), m_Rels.end(), r);

  if (it != m_Rels.end())
    {
    m_Rels.erase(it);    
    }
  else
    {
    it = std::find(m_IsolatedRels.begin(), m_IsolatedRels.end(), r);
    if (it == m_IsolatedRels.end())
      INTERNAL_ERROR;
    m_IsolatedRels.erase(it);
    }
  return std::auto_ptr<Relation>(r);
  }

relset_t ConstraintNetwork::getConnectedRelations(const Relation &r) const
  {
  relset_t  rels;
  colvector_t::const_iterator it = r.getColumns().begin(), it_end = r.getColumns().end();
  for(; it != it_end; ++it)
    {
    Variable &v = (*it)->getVariable();

    rels.insert(getRelations(v.getId()).begin(), getRelations(v.getId()).end());
    }
  return rels;
  }

void ConstraintNetwork::dumpRels() const
  {
  relcol_t::const_iterator it;
  for (it = m_Rels.begin(); it != m_Rels.end(); ++it)
    {
    Relation *r = *it;
    std::cout << *r << std::endl;
    }
  }

double ConstraintNetwork::getRelSize() const
  {
  double relsize = 0;

  relcol_t::const_iterator it;
  for (it = m_Rels.begin(); it != m_Rels.end(); ++it)
    {
    Relation *r = *it;
    relsize += r->getSize();
    }
  return relsize;
  }

std::ostream &operator<<(std::ostream &s, const ConstraintNetwork::relcol_t &rels)
  {
  s << '{';
  ConstraintNetwork::relcol_t::const_iterator it = rels.begin();
  while (it != rels.end())
    {
    if (it != rels.begin())
      s << ",";
    s << "(" << (*it)->getId() << ",";
    write_set(s, (*it)->getColumns());
    s << ")";
    ++it;
    }
  s << '}';
  return s;
  }

void ConstraintNetwork::checkInvariants() const
  {
  checkXRef();
  std::for_each(m_Rels.begin(), m_Rels.end(), boost::mem_fn(&Relation::checkInvariants));
  std::for_each(m_IsolatedRels.begin(), m_IsolatedRels.end(), boost::mem_fn(&Relation::checkInvariants));
  }

void ConstraintNetwork::findNewBoundVariables(varset_t &bound_variables)
  {
  bound_variables.clear();

  std::for_each(m_Rels.begin(), m_Rels.end(), 
                boost::bind(&Relation::identifyUnaryConstraints, _1, &bound_variables));

  std::for_each(m_IsolatedRels.begin(), m_IsolatedRels.end(),
                boost::bind(&Relation::identifyUnaryConstraints, _1, &bound_variables));
  std::cout << "Found " << bound_variables.size() << " bound vars in all relations\n";
  }
