/*
 * Copyright (C) 2003 Jeppe Nejsum Madsen, nejsum@diku.dk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#ifndef CONSTRAINT_NETWORK_H
#define CONSTRAINT_NETWORK_H

#include "variable.hpp"
#include "relation.hpp"

struct contradiction_error : public std::runtime_error
  {
  const Relation *r1, *r2;
  contradiction_error(const Relation *r1, const Relation *r2)
    : std::runtime_error("Contradiction"), r1(r1), r2(r2) {}
  };

class ConstraintNetwork : boost::noncopyable 
{
public:
  typedef std::list<Relation *>   relcol_t;
  typedef std::vector<Variable *> varcol_t;

  ConstraintNetwork(std::string name = "unnamed");
  virtual ~ConstraintNetwork();
  

  void                    addVariable(std::auto_ptr<Variable> var);
  void                    addRelation(std::auto_ptr<Relation> rel, bool updateXref = true);
  std::auto_ptr<Relation> removeRelation(Relation *r, bool updateXref = true);
  
  const std::string getName()                   const {return m_Name;}
  size_t            getNumVariables()           const {return m_Vars.size();}
  Variable         &getVariable(size_t var_id)  const {return *m_Vars[var_id];}
  const varcol_t   &getVariables()              const {return m_Vars;}
  varcol_t         &getVariables()                    {return m_Vars;}
  size_t            getNumRelations()           const {return m_Rels.size();}
  relcol_t         &getRelations()                    {return m_Rels;}
  const relcol_t   &getRelations()              const {return m_Rels;}
  relset_t         &getRelations(size_t var_id)       {return m_VarRels[var_id];}
  const relcol_t   &getIsolatedRelations()      const {return m_IsolatedRels;}
  const relset_t   &getRelations(size_t var_id) const {return m_VarRels[var_id];}

  relset_t            getConnectedRelations(const Relation &r) const;

    
  double            getRelSize()              const;
  size_t            getNumSolutions()         const;
  size_t            getNewRelId();

  void              findNewBoundVariables(varset_t &bound_variables);
  void              dumpRels() const;
  void              buildXRef();
  void              addRelationXRef(Relation *rel);
  void              removeRelationXRef(Relation *rel);
  void              checkXRef() const;
  void              checkInvariants() const;

  protected:
    varcol_t                m_Vars;
    relcol_t                m_Rels;
    relcol_t                m_IsolatedRels;
    std::vector<relset_t>   m_VarRels;

    std::string             m_Name;
    size_t                  m_NextRelId;
  };

void common_columns(const Relation &r, const Relation &s, colpairvector_t &common_columns);

  class Last
    {
    typedef std::map<idpair_t, std::vector<int> > lastmap_t;
    
    lastmap_t m_LastMap;

    public:
      std::vector<int> &Last::last(const Relation &r, const Relation &s);
    };

bool weak_arc_consistency(ConstraintNetwork &instance, relset_t *modified_rels = 0);
bool weak_arc_consistency(ConstraintNetwork &instance, const relset_t &Q, relset_t *modified_rels = 0);
bool weak_arc_consistency(ConstraintNetwork &instance, Relation &r,relset_t *modified_rels = 0);

bool node_consistency(ConstraintNetwork &instance, const varset_t &changed_variables, relset_t *modified_relations = 0);
bool node_consistency(ConstraintNetwork &instance, relset_t *modified_relations = 0);

std::ostream &operator<<(std::ostream &s, const ConstraintNetwork::relcol_t &rels);

#endif
