/*
 * Copyright (C) 2003 Jeppe Nejsum Madsen, nejsum@diku.dk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include "boost/mpl/if.hpp"
#include "csplib/uniquerows.hpp"

#pragma warning(push)
#pragma warning(disable: 4127)  // conditional expression is constant

namespace detail
{
  // Methods for uncompressed columns
  template<typename Container>
    struct uncompressed_helper
    {
      inline static size_t expandedSize(const Container &cont) 
      {
        return cont.size();
      }
    };
  // Methods for compressed columns
  template<typename Container, typename CellTraits>
    struct compressed_helper
    {
      inline static size_t expandedSize(const Container &cont)
      {
        typename Container::const_iterator it = cont.begin(), pend = cont.end();
        size_t size = 0;
        while(it != pend)
          {
          size += CellTraits::size(*it);
          ++it;
          }
        return size;
      }
    };

  // Method for scalar cells
  template<typename Container>
    struct scalar_helper
    {
      inline static void pruneCells(Container &c,
                                    urvector_t::const_iterator start,
                                    urvector_t::const_iterator pend) 
      {
        typename Container::iterator it = c.begin();
        size_t newSize = pend - start;
        while(start != pend)
          {
          *it++ = c[start->first];
          ++start;
          }
        c.resize(newSize);
      }
    };

  // Method for pointer cells
  template<typename Container>
    struct pointer_helper
    {
      inline static void pruneCells(Container &c,
                                    urvector_t::const_iterator rit,
                                    urvector_t::const_iterator rpend) 
      {
        typename Container::iterator nxt,it = c.begin(), pend = c.end();
        size_t row = 0, newsize = rpend - rit;
        
        nxt = it;
        while(it != pend)
          {
          if (rit != rpend && row == rit->first)
            {
            *nxt = *it;
            ++nxt;
            ++rit;
            }
          else
            boost::checked_delete(*it);
          ++it;
          ++row;
          }
        // TODO: Free memory when resizing
        c.resize(newsize);
      }
    };
}

// A container for column cells
template<typename CellTraits,
         bool IsCompressed,
         typename Container = std::vector<typename CellTraits::value_type> >
struct ColumnContainer : public Container
{
  typedef typename boost::mpl::if_<
    boost::mpl::bool_c<IsCompressed>,
    ::detail::compressed_helper<Container,CellTraits>,
    ::detail::uncompressed_helper<Container> >::type compression_helper;
  
  typedef typename boost::mpl::if_<
    boost::is_pointer<typename CellTraits::value_type>,
    ::detail::pointer_helper<Container>,
    ::detail::scalar_helper<Container> >::type pointer_helper;
  
  typedef Container container_t;
  typedef CellTraits celltraits_t;  
  
  size_t expandedSize() const
  {
    return compression_helper::expandedSize(*this);
  }

  void pruneCells(urvector_t::const_iterator start,
                  urvector_t::const_iterator pend) 
  {
    pointer_helper::pruneCells(*this, start, pend);
  }
    
  // Fill in cells in dest from joinindex and two columns
  template<typename Impl, typename Dest>
  void createFromJoinIndex(const joinindex_t &ji, const Impl &impl, Dest &dest) const
    {
    joinindex_t::const_iterator it = ji.begin(), pend = ji.end();
    while(it != pend)
      {
      size_t row1 = it->first, row2 = it->second;
      dest.push_back(CellTraits::intersection((*this)[row1], impl[row2]));
      ++it;
      }
    }

  // 
  template<typename Impl>
    bool findSupport(std::vector<bool> &mask, const Impl &impl) const
    {
    typename container_t::const_iterator it1 = begin(), pend1 = end();
    typename Impl::container_t::const_iterator pend2 = impl.end();
    
    size_t row = 0;
    bool changed = false;

    while(it1 != pend1)
      {
      if (mask[row])
        {
        typename Impl::container_t::const_iterator it2 = impl.begin();

        while(it2 != pend2)
          {
          if (!CellTraits::intersectionEmpty(*it1, *it2))
            break;
          ++it2;
          }
        if (it2 == pend2)
          {
          mask[row] = false;
          changed = true;
          }
        }
      ++row;
      ++it1;
      }
    return changed;
    }

   template<typename Impl>
    bool findInitialSupport(boolvector_t &rvalid, boolvector_t &svalid, 
                            std::vector<int> &rsupport, std::vector<size_t> &rsupport_cols,
                            size_t colcount, const Impl &impl) const
    {
    typename container_t::const_iterator it1 = begin(), pend1 = end();
    typename Impl::container_t::const_iterator pend2 = impl.end();
    
    size_t  row     = 0;
    bool    changed = false;

    for (row = 0; it1 != pend1; ++row, ++it1)
      {
      if (rvalid[row])
        {
        size_t    row2;
        typename  Impl::container_t::const_iterator it2 = impl.begin();

        int initial_sup = rsupport[row];

        if (initial_sup < 0)
          row2 = 0;
        else
          {
          row2 = (size_t)initial_sup;
          // If we have checked support for all columns and
          // row2 is still valid, no need to check again
          if (rsupport_cols[row] == colcount && svalid[row2])
            continue;

          std::advance(it2, row2);
          }

        while(it2 != pend2)
          {
          if (svalid[row2] && !CellTraits::intersectionEmpty(*it1, *it2))
            break;
          ++it2;
          ++row2;
          }
        if (it2 != pend2)
          {
          if ((int)row2 == initial_sup)
            {
            // We have verified support for this column, increase
            // the number of support columns
            rsupport_cols[row]++;
            }
          else
            {
            // We have found new support, reinitialize support &
            // the number of support columns
            rsupport[row]      = row2;
            rsupport_cols[row] = 1;
            }
          }
        else
          {
          rvalid[row]         = false;
          rsupport_cols[row]  = colcount;
          changed             = true;
          }
        }
      }
    return changed;
    }


  template<typename Impl>
    void updateTupleMask(size_t row, std::vector<bool> &mask, const Impl &impl) const
    {
    typename Impl::container_t::const_iterator it = impl.begin(), pend = impl.end();
    
    size_t row2 = 0;
    while(it != pend)
      {
      if (mask[row2] && CellTraits::intersectionEmpty((*this)[row], *it))
        mask[row2] = false;
      ++row2;
      ++it;
      }
    }

  template<typename Impl>
    void updateContainmentMask(size_t row, std::vector<bool> &mask, const Impl &impl) const
    {
    typename Impl::container_t::const_iterator it = impl.begin(), pend = impl.end();
    
    size_t row2 = row+1;
    std::advance(it, row2);
    while(it != pend)
      {
      if (mask[row2] && !CellTraits::cellContains((*this)[row], *it))
        mask[row2] = false;
      ++row2;
      ++it;
      }
    }

  void updateRowCounts(size_t start_row, std::vector<double> &counts, double &intersection_count) const
    {
    typename container_t::const_iterator it = begin(), pend = end();

    typename CellTraits::value_type cell, tmp_cell;

    if (start_row == 0)
      cell = *it;

    size_t row = start_row+1;
    std::advance(it, row);
    while(it != pend)
      {
      counts[row] *= CellTraits::intersectionSize((*this)[start_row], *it);
      if (start_row == 0)
        {
        if (CellTraits::intersectionEmpty(cell, *it))
          intersection_count = 0.0;
        else
          {
          tmp_cell = CellTraits::intersection(cell, *it);
          std::swap(cell, tmp_cell);
          }
        }
      ++row;
      ++it;
      }
    if (start_row == 0)
      intersection_count *= CellTraits::size(cell);
    }

  void updateRowCounts(std::vector<double> &counts) const
    {
    typename container_t::const_iterator it = begin(), pend = end();

    size_t row = 0;
    while(it != pend)
      {
      counts[row] *= CellTraits::size(*it);
      ++it;
      ++row;
      }
    }

  // Create join index from two columns
  template<typename Impl>
  bool createJoinIndex(joinindex_t &ji, const Impl &impl, size_t &start_row, size_t max_entries) const
    {
    // Impl contains the larger column
    typename container_t::const_iterator it1 = begin(), pend1 = end();
    typename Impl::container_t::const_iterator pend2 = impl.end();
    
    std::advance(it1, start_row);

    while(it1 != pend1)
      {
      size_t row2 = 0;
      typename Impl::container_t::const_iterator it2 = impl.begin();
      while(it2 != pend2)
        {
        if (!CellTraits::intersectionEmpty(*it1, *it2))
          {
          ji.push_back(std::make_pair(start_row, row2));
          }
        ++it2;
        ++row2;
        }
      ++start_row;
      if (ji.size() > max_entries)
        break;
      ++it1;
      }
    //std::cout << "Initial joinindex:" << ji.size() << "\n";
    return it1 == pend1;
    }

  // Revise a joinindex from two columns
  // Only check the pairs contained in the joinindex
  template<typename Impl>
  void reviseJoinIndex(joinindex_t &ji, const Impl &impl, size_t start) const
    {
    joinindex_t::const_iterator it, pend = ji.end();
    joinindex_t::iterator nextVal = ji.begin();

    std::advance(nextVal, start);
    it = nextVal;

    while(it != pend)
      {
      size_t row1 = it->first, row2 = it->second;
      if (!CellTraits::intersectionEmpty((*this)[row1], impl[row2]))
        {
        *nextVal++ = *it;
        }
      ++it;
      }
    ji.resize(std::distance(ji.begin(), nextVal));
    }

  // Clone a column into dest
  void clone(size_t numNewCells, ColumnContainer &dest, bool adjacentCells) const
    {
    size_t numCopies = numNewCells/size(); 

    if (adjacentCells)
      {
      for(size_t j = 0; j < size(); ++j)
        {
        for(size_t i = 0; i < numCopies; ++i)
          {
          dest.push_back(CellTraits::cloneCell((*this)[j]));
          }
        }
      }
    else
      {
      for(size_t i = 0; i < numCopies; ++i)
        {
        for(size_t j = 0; j < size(); ++j)
          {
          dest.push_back(CellTraits::cloneCell((*this)[j]));
          }
        }
      }
    }

  // Clone a column into dest, picking only those cells specified in first/second pair
  // of joinindex
  void cloneFromJoinIndex(const joinindex_t &ji, ColumnContainer &dest, bool useFirst) const
    {
    joinindex_t::const_iterator it = ji.begin(), pend = ji.end();
    while(it != pend)
      {
      if (useFirst)
        dest.push_back(CellTraits::cloneCell((*this)[it->first]));
      else
        dest.push_back(CellTraits::cloneCell((*this)[it->second]));
      ++it;
      }
    }
  
  // Precondition: mask size equals domain, mask is zero filled
   void createUniqueMask(const boolvector_t &valid_cell, boolvector_t &mask) const
    {
    typename container_t::const_iterator it = begin(), pend = end();
    boolvector_t::const_iterator bit = valid_cell.begin();

    while (it != pend)
      {
      if (*bit)
        CellTraits::setMask(*it, mask);
      ++it;
      ++bit;
      }
    }

   // Precondition: mask size equals size()
   bool updateValidRowMask(const Variable &var, std::vector<bool> &mask)  const
    {
    typename container_t::const_iterator it = begin(), pend = end();
    bool    updated = false;

    size_t row = 0;
    while (it != pend)
      {
      if (mask[row] && !CellTraits::cellIsValid(*it, var))
        {
        updated   = true;
        mask[row] = false;
        }
      ++row;
      ++it;
      }
    return updated;
    }

   bool revise(const Variable &var, std::vector<bool> &mask)  
    {
    typename container_t::iterator it = begin(), pend = end();
    bool    updated = false;

    size_t row = 0;
    while (it != pend)
      {
      if (mask[row] && !CellTraits::revise(*it, var))
        {
        updated   = true;
        mask[row] = false;
        }
      ++row;
      ++it;
      }
    return updated;
    }

    std::string cellToString(size_t row, const Variable &var) const 
      {
      return CellTraits::toString((*this)[row], var);
      }

    void remapIndicies(const idvector_t &new_ids)
      {
      typename container_t::iterator it = begin(), pend = end();
      while(it != pend)
        {
        CellTraits::remapCell(*it++, new_ids);
        }
      }

    void expandCell(size_t row, idvector_t &values) const
      {
      CellTraits::expandCell((*this)[row], values);
      }

    bool cellsAreEqual(size_t row1, size_t row2) const
      {
      return CellTraits::equal((*this)[row1], (*this)[row2]);
      }

    void xorHashValues(size_t *hashValues, size_t index, size_t domainBits) const
      {
      typename container_t::const_iterator it = begin(), pend = end();
      while(it != pend)
        {
        CellTraits::hashCell(hashValues++, index, *it++, (unsigned char)domainBits);
        }
      }

    void fillInUniqueComplement(UniqueComplement &ur) const
      {
      typename container_t::const_iterator it = begin(), pend = end();
      size_t row = 0;
      while(it != pend)
        {
        // TODO: Make idvector_t dependent on column type!
        std::pair<UniqueComplement::iterator,bool> p = ur.insert(std::make_pair(row,(idvector_t *)0));

        if (p.second)
          {
          // First occurence of this tuple. Construct new id vector
          // to hold the values from this column in matching tuples
          p.first->second = ur.objectPool.construct();
          //p.first->second->reserve(10);
          }

        // Add the value(s) from this row to the id vector
        CellTraits::expandCell(*it, *(p.first->second));
      
        ++it;
        ++row;
        }
      }

    void fillInUniqueRows(UniqueRows &ur) const
      {
      typename container_t::const_iterator it = begin(), pend = end();
      size_t row = 0;

      while(it != pend)
        {
        std::pair<UniqueRows::iterator,bool> p = ur.insert(std::make_pair(row,(idvector_t *)0));

        if (p.second)
          {
          // First occurence of this tuple. Construct new id vector
          // to hold the row nos of matching tuples
          p.first->second = ur.objectPool.construct();
          }
        else
          // Already have a match for this tuple, add this row to the vector
          p.first->second->push_back(row);
      
        ++it;
        ++row;
        }
      }
  };
#pragma warning(pop)

// Base class for concrete column implementations.  Inherits from type
// specified in ColumnType using CellTraits as traits
// ColumnType2/CellTraits2 specifies the types for the matching
// "inverse" column i.e. compressed/uncompressed respectively
template<typename ColumnType, typename CellTraits,
         typename ColumnType2, typename CellTraits2>
class ColumnBase : public ColumnType
  {
  private:
  struct is_compressed // Is column compressed?
    {
      enum {value = boost::is_same<ColumnType, CompressedColumn>::value};
      int x;
    };
  // Container type for this column
  typedef ColumnContainer<CellTraits, is_compressed::value> container_t;
  // Container type for "inverse" column
  typedef ColumnContainer<CellTraits2, !is_compressed::value> container2_t;

  // Typedefs used if this column is compressed
  struct compressed_types
    {
      typedef container_t compressed_container_t; 
      typedef container2_t uncompressed_container_t;
      typedef ColumnBase<ColumnType, CellTraits, ColumnType2, CellTraits2> compressed_base_t;
      typedef ColumnBase<ColumnType2, CellTraits2, ColumnType, CellTraits> uncompressed_base_t;
    };

  // Typedefs used if this column is uncompressed
  struct uncompressed_types
    {
      typedef container_t uncompressed_container_t;
      typedef container2_t compressed_container_t; 
      typedef  ColumnBase<ColumnType, CellTraits, ColumnType2, CellTraits2> uncompressed_base_t;
      typedef  ColumnBase<ColumnType2, CellTraits2, ColumnType, CellTraits> compressed_base_t;
    };

  // Select appropriate types
  typedef  typename boost::mpl::if_<
          boost::mpl::bool_c<is_compressed::value>
        , compressed_types
        , uncompressed_types
        >::type     helper;
    
  protected:
  
  // Get container member when we _know_ c is same type as us, but compressed
  static const typename helper::compressed_container_t &get_container_c(const Column &c)
      {
      const typename helper::compressed_base_t *cc = dynamic_cast<const typename helper::compressed_base_t *>(&c);
      assert(cc != 0);
      return cc->container;
      }
  static typename helper::compressed_container_t &get_container_c(Column &c)
      {
      typename helper::compressed_base_t *cc = dynamic_cast<typename helper::compressed_base_t *>(&c);
      assert(cc != 0);
      return cc->container;
      }

  // Get container member when we _know_ c is uncompressed same type as us
  static const typename helper::uncompressed_container_t  &get_container_u(const Column &c)
    {
    const typename helper::uncompressed_base_t *uc = dynamic_cast<const typename helper::uncompressed_base_t *>(&c);
    assert(uc != 0);
    return uc->container;
    }

  static typename helper::uncompressed_container_t  &get_container_u(Column &c)
    {
    typename helper::uncompressed_base_t *uc = dynamic_cast<typename helper::uncompressed_base_t *>(&c);
    assert(uc != 0);
    return uc->container;
    }

  public:
    container_t container;

    ColumnBase(Variable &var)
      : ColumnType(var)
      {
      }

  virtual void createUniqueMask(const boolvector_t &valid_cell, boolvector_t &mask) const
    {
    container.createUniqueMask(valid_cell, mask);
    }

  virtual bool updateValidRowMask(std::vector<bool> &mask)  const
    {
    return container.updateValidRowMask(getVariable(), mask);
    }

  virtual bool revise(std::vector<bool> &mask)  
    {
    return container.revise(getVariable(), mask);
    }

  virtual bool findSupport(const Column &col, std::vector<bool> &mask) 
    {
    bool res;

    if (col.isCompressed())
      {
      res = container.findSupport(mask, get_container_c(col));
      }
    else
      {
      res = container.findSupport(mask, get_container_u(col));
      }
    return res;
    }

  virtual void updateTupleMask(size_t row, const Column &col, std::vector<bool> &mask) 
    {
    if (col.isCompressed())
      container.updateTupleMask(row, mask, get_container_c(col));
    else
      container.updateTupleMask(row, mask, get_container_u(col));
    }

  virtual void updateContainmentMask(size_t row, const Column &col, std::vector<bool> &mask) 
    {
    if (col.isCompressed())
      container.updateContainmentMask(row, mask, get_container_c(col));
    else
      container.updateContainmentMask(row, mask, get_container_u(col));
    }

  virtual size_t getNumCells() const 
    {
    return container.size();
    }

  virtual size_t getSize() const 
    {
    return container.expandedSize();
    }

  virtual void xorHashValues(size_t *hashValues) const
    {
    container.xorHashValues(hashValues, m_Index, m_Variable.getDomainBits());
    }
  
  virtual void pruneCells(urvector_t::const_iterator rit, urvector_t::const_iterator rpend) 
    {
    container.pruneCells(rit, rpend);
    }

  virtual void fillInUniqueComplement(UniqueComplement &ur) const
    {
    container.fillInUniqueComplement(ur);
    }

  virtual void fillInUniqueRows(UniqueRows &ur) const
    {
    container.fillInUniqueRows(ur);
    }

  virtual void expandCell(size_t row, idvector_t &values) const
    {
    container.expandCell(row, values);
    }

  virtual bool cellsAreEqual(size_t row1, size_t row2) const
    {
    return container.cellsAreEqual(row1, row2);
    }
  
  virtual std::string cellToString(size_t row) const 
    {
    return container.cellToString(row, getVariable());
    }
  virtual bool findInitialSupport(const Column &col, boolvector_t &rvalid, boolvector_t &svalid, 
                                  std::vector<int> &rsupport, std::vector<size_t> &rsupport_cols, int colcount) 
    {
    bool res;

    if (col.isCompressed())
      {
      res = container.findInitialSupport(rvalid, svalid, rsupport, rsupport_cols, colcount, get_container_c(col));
      }
    else
      {
      res = container.findInitialSupport(rvalid, svalid, rsupport, rsupport_cols, colcount, get_container_u(col));
      }
    return res;
    }

  virtual bool createJoinIndex(joinindex_t &ji, const Column &col, size_t &start_row, size_t max_entries) const
    {
    if (col.isCompressed())
      {
      return container.createJoinIndex(ji, get_container_c(col), start_row, max_entries);
      }
    else
      {
      return container.createJoinIndex(ji, get_container_u(col), start_row, max_entries);
      }
    }

  virtual void  remapIndicies(const idvector_t &new_ids)
    {
    container.remapIndicies(new_ids);
    }

  virtual void updateRowCounts(size_t start_row, std::vector<double> &counts, double &intersection_count) const
    {
    container.updateRowCounts(start_row, counts, intersection_count);
    }

  virtual void updateRowCounts(std::vector<double> &counts) const
    {
    container.updateRowCounts(counts);
    }

  virtual void reviseJoinIndex(joinindex_t &ji, const Column &col, size_t start) const
    {
    if (col.isCompressed())
      {
      container.reviseJoinIndex(ji, get_container_c(col), start);
      }
    else
      {
      container.reviseJoinIndex(ji, get_container_u(col), start);
      }
    }
  };
