/*
 * Copyright (C) 2003 Jeppe Nejsum Madsen, nejsum@diku.dk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <bitset>
#include "boost/bind.hpp"
#include "csplib/utility/newbitset.hpp"

template <typename CellT>
struct BitsetCellTraits
  {
  typedef CellT value_type;

  static std::string toString(const CellT &cellValue, const Variable &v)
    {
    std::stringstream s;
    s << "{";
    bool first = true;
    for(typename CellT::index_type i = 0; i < cellValue.size(); i++)
      {
      if (cellValue.test(i))
        {
        if(!first)
          s << ",";
        first = false;
        s << v.getDomainValueAsString(i);
        }
      }
    s << "}";
    return s.str();
    }

  inline static bool cellIsValid(const CellT &cellValue, const Variable &var)
    {
    for(typename CellT::index_type i = 0; i < cellValue.size(); i++)
      {
      if (cellValue.test(i) && var.isValueValid(i))
        return true;
      }
    return false;
    }

  inline static bool revise(CellT &cellValue, const Variable &var)
    {
    for(typename CellT::index_type i = 0; i < cellValue.size(); i++)
      {
      if (cellValue.test(i) && !var.isValueValid(i))
        {
        cellValue.clear(i);
        }
      }
    return cellValue.any();
    }

  inline static void remapCell(CellT &cellValue, const idvector_t &new_ids)
    {
    for(typename CellT::index_type i = 0; i < cellValue.size(); i++)
      {
      if (cellValue.test(i))
        {
        // Safe, new id is never larger than old id
        cellValue.clear(i);
        if (new_ids[i] > 99999999)
          INTERNAL_ERROR;
        cellValue.set(new_ids[i]);
        }
      }
    }

  inline static void setMask(const CellT &cellValue, std::vector<bool> &mask)
    {
    for(typename CellT::index_type i = 0; i < cellValue.size(); i++)
      {
      if (cellValue.test(i))
        {
        mask[(size_t)i] = true;
        }
      }
    }

  inline static size_t size(const CellT &cellValue)
    {
    return cellValue.count();
    }

  inline static CellT cloneCell(const CellT &cellValue)
    {
    return cellValue;
    }

  inline static bool intersectionEmpty(const CellT &v1, const CellT &v2)
    {
    CellT t(v1);
    t &= v2;
    return t.none();
    }

  inline static bool intersectionEmpty(const CellT &v1, unsigned char v2)
    {
    return !v1.test(v2);
    }

  inline static bool cellContains(const CellT &v1, const CellT &v2)
    {
    for(typename CellT::index_type i = 0; i < v2.size(); i++)
      {
      if (v2.test(i) && !v1.test(i))
        return false;
      }
    return true;
    }

  inline static bool cellContains(const CellT &v1, unsigned char v2)
    {
    return v1.test(v2);
    }

  inline static CellT intersection(const CellT &v1, const CellT &v2)
    {
    CellT t(v1);
    t &= v2;
    return t;
    }

  inline static size_t intersectionSize(const CellT &v1, const CellT &v2)
    {
    CellT t(v1);
    t &= v2;
    return t.count();
    }

  inline static unsigned char intersection(const CellT &, unsigned char v2)
    {
    // We know bitset contains v2
    return v2;
    }

  inline static void expandCell(const CellT &cellValue, idvector_t &values)
    {
    for(size_t bit = 0; bit < cellValue.size(); bit++)
      {
      if(cellValue.test(bit))
        values.push_back(bit);
      }
    }
  
  inline static bool equal(const CellT cellValue1, const CellT cellValue2)
    {
    return cellValue1 == cellValue2;
    }

  inline static void hashCell(size_t *hashValue, size_t index, CellT cellValue, unsigned char)
    {
    *hashValue ^= cellValue.hashValue(index);
    }

  inline static void hashCell(size_t *hashValue, size_t *out, size_t index, CellT cellValue, unsigned char)
    {
    *out = *hashValue ^ cellValue.hashValue(index);
    }

  };

// Generic CellTraits for cells containing containers  of type T
template <typename T>
struct ContainerCellTraits
  {
  typedef std::vector<T>  Container;
  typedef Container *value_type;

  static std::string toString(const value_type &cellValue, const Variable &v)
    {
    std::stringstream s;
    s << '{';
    typename Container::const_iterator it = cellValue->begin();
    while (it != cellValue->end())
      {
      if (it != cellValue->begin())
        s << ",";
      s << v.getDomainValueAsString(*it++);
      }
    s << '}';
    return s.str();
    }

 inline static void setMask(const value_type cellValue, std::vector<bool> &mask)
    {
    typename Container::const_iterator it = cellValue->begin(), pend = cellValue->end();

    while(it != pend)
      {
      mask[*it] = true;
      ++it;
      }
    }

  inline static bool cellIsValid(const value_type cellValue, const Variable &var)
    {
    typename Container::iterator it = cellValue->begin(), pend = cellValue->end(), newend;

    newend = std::stable_partition(it, pend, boost::bind(&Variable::isValueValid, boost::ref(var), _1));

    return newend != it;
    }

  inline static bool revise(value_type cellValue, const Variable &var)
    {
    typename Container::iterator it = cellValue->begin(), pend = cellValue->end(), newend;

    newend = std::stable_partition(it, pend, boost::bind(&Variable::isValueValid, boost::ref(var), _1));
    if (newend != pend)
      {
      cellValue->resize(std::distance(cellValue->begin(), newend));
      }

    return cellValue->size() > 0;
    }

  inline static void remapCell(value_type cellValue, const idvector_t &new_ids)
    {
    typename Container::iterator it = cellValue->begin(), pend = cellValue->end();
    while(it != pend)
      {
      *it = new_ids[*it];
      ++it;
      }
    }

  inline static value_type intersection(const value_type v1, const value_type v2)
    {
    value_type intersect = new Container;
    std::set_intersection(v1->begin(), v1->end(), v2->begin(), v2->end(), std::back_insert_iterator<Container>(*intersect));
    return intersect;
    }

  inline static size_t intersectionSize(const value_type v1, const value_type v2)
    {
    value_type intersect = new Container;
    std::set_intersection(v1->begin(), v1->end(), v2->begin(), v2->end(), std::back_insert_iterator<Container>(*intersect));
    size_t size = intersect->size();
    delete intersect;
    return size;
    }

  
  inline static typename Container::value_type intersection(const value_type, typename Container::value_type v2)
    {
    // We know the container contains v2
    return v2;
    }

  inline static bool cellContains(const value_type v1, const value_type v2)
    {
    return std::includes(v1->begin(), v1->end(), v2->begin(), v2->end());
    }

  inline static bool cellContains(const value_type v1, typename Container::value_type v2)
    {
    return std::binary_search(v1->begin(), v1->end(), v2);
    }

  inline static bool intersectionEmpty(const value_type v1, typename Container::value_type v2)
    {
    // Container v1 is sorted
    return !std::binary_search(v1->begin(), v1->end(), v2);
    }

  inline static bool intersectionEmpty(const value_type v1, const value_type v2)
    {
    typename Container::const_iterator it1 = v1->begin(), pend1 = v1->end();
    typename Container::const_iterator it2 = v2->begin(), pend2 = v2->end();
    
    // Both containers are sorted
    while (it1 != pend1 && it2 != pend2)
      {
      if (*it1 == *it2)
        return false;
      if (*it1 < *it2)
        ++it1;
      else
        ++it2;
      }
    return true;
    }

  inline static size_t size(const value_type v1)
    {
    return v1->size();
    }

  inline static value_type cloneCell(const value_type cellValue)
    {
    return new Container(*cellValue);
    }

  inline static void expandCell(const value_type &cellValue, idvector_t &values)
    {
    values.insert(values.end(), cellValue->begin(), cellValue->end());
    }

  inline static bool equal(const value_type cellValue1, const value_type cellValue2)
    {
    return *cellValue1 == *cellValue2;
    }

  inline static void hashCell(size_t *hashValue, size_t index, value_type cellValue, unsigned char domainBits)
    {
    typename Container::const_iterator it = cellValue->begin(), past_end = cellValue->end();
    size_t hv = 0;
    while (it != past_end)
      {
      hv ^= HashFunction(index++, *it++, domainBits);
      }
    *hashValue ^= hv;
    }

  inline static void hashCell(size_t *hashValue, size_t *out, size_t index, value_type cellValue, unsigned char domainBits)
    {
    typename Container::const_iterator it = cellValue->begin(), past_end = cellValue->end();
    size_t hv = 0;
    while (it != past_end)
      {
      hv ^= HashFunction(index++, *it++, domainBits);
      }
    *out = *hashValue ^ hv;
    }

  };

// Generic CellTraits for cells containing scalars of type T
template <typename T>
struct ScalarCellTraits
  {
    //BOOST_STATIC_ASSERT(boost::is_integral<T>::value);
    
  typedef T value_type;

    inline static bool intersectionEmpty(const T &v1,T v2)
      {
      return v1 != v2;
      }

    inline static bool intersectionEmpty(const T &v1, const std::vector<T> *v2)
      {
      // Container v2 is sorted
      return !std::binary_search(v2->begin(), v2->end(), v1);
      }

    template<size_t Nx>
    inline static bool intersectionEmpty(const T &v1, const NewBitset<Nx> &v2)
      {
      return !v2.test(v1);
      }

  inline static std::string toString(T cellValue, const Variable &v)
    {
    return v.getDomainValueAsString(cellValue);
    //return boost::lexical_cast<std::string>(static_cast<size_t>(cellValue));
    }

  inline static T cloneCell(T cellValue)
    {
    return cellValue;
    }

  
  inline static void  setMask(T cellValue, std::vector<bool> &mask)
    {
    mask[(size_t)cellValue] = true;
    }

  inline static bool cellIsValid(const T cellValue, const Variable &var)
    {
    return var.isValueValid(cellValue);
    }

  inline static bool revise(T cellValue, const Variable &var)
    {
    return var.isValueValid(cellValue);
    }

  inline static void remapCell(T &cellValue, const idvector_t &new_ids)
    {
    if (new_ids[cellValue] > 9999999)
      INTERNAL_ERROR;
    cellValue = new_ids[cellValue];
    }


  inline static size_t intersectionSize(const T v1, const T v2)
    {
    return v1 == v2 ? 1 : 0;
    }

  inline static T intersection(const T v1, const T )
    {
    return v1;
    }

  template<typename Container>
  inline static T intersection(const T v1, const Container &)
    {
    // We know the container contains v1
    return v1;
    }

  inline static bool cellContains(const T v1, T v2)
    {
    return v1 == v2;
    }

  template<size_t Nx>
  inline static bool cellContains(const T v1, const NewBitset<Nx> &v2)
    {
    return v2.size() == 1 && v2.test(v1);
    }

  inline static bool cellContains(const T v1, const std::vector<T> *c)
    {
    return c->size() == 1 && *c->begin() == v1;
    }

  inline static void expandCell(T cellValue, idvector_t &values)
    {
    values.push_back(cellValue);
    }
    
  inline static bool equal(T cellValue1, T cellValue2)
    {
    return cellValue1 == cellValue2;
    }

  inline static size_t size(T )
    {
    return 1;
    }

  inline static void hashCell(size_t *hashValue, size_t index, T cellValue, unsigned char domainBits)
    {
    *hashValue ^= HashFunction(index, cellValue, domainBits);
    }
  };
