/*
 * Copyright (C) 2003 Jeppe Nejsum Madsen, nejsum@diku.dk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include "stdafx.h"
#include <boost/graph/adjacency_list.hpp>
#include "boost/property_map.hpp"
#include <boost/graph/named_function_params.hpp>
#include "boost/graph/undirected_dfs.hpp"
#include <boost/graph/kruskal_min_spanning_tree.hpp>
#include <boost/graph/graphviz.hpp>
#include "graph.hpp"
#include "relation.hpp"

using namespace boost;

// Graph visitor to detect cycles
struct tree_order: public boost::dfs_visitor<>
  {
  tree_order(bool &b) : has_cycle(b) { }

  template <class Edge, class Graph>
  void back_edge(Edge , Graph& ) {
      has_cycle = true;
  }
  protected:
    bool &has_cycle;
  };


bool is_cyclic(dualgraph_t &g)
  {
  bool has_cycle = false;

  tree_order vis(has_cycle);
  boost::undirected_dfs(g, vis, 
                        boost::get(boost::vertex_color, g),
                        boost::get(boost::edge_color, g));
  return has_cycle;
  }

void createDualGraph(ConstraintNetwork &csp, dualgraph_t &g)
  {
  g.clear();

  // Maps relation to vertex
  typedef std::map<Relation *, dualgraph_t::vertex_descriptor, RelationPtrOrder>  relmap_t;
  relmap_t relmap;

  property_map<dualgraph_t, vertex_relation_t>::type vertexrel = get(vertex_relation_t(), g);
  
  ConstraintNetwork::relcol_t::const_iterator it = csp.getRelations().begin(), pend = csp.getRelations().end();
  while(it != pend)
    {
    Relation *r = *it;
    dualgraph_t::vertex_descriptor v = add_vertex(r->getId(), g);
    relmap[r] = v;
    vertexrel[v] = r;
    ++it;
    }
  property_map<dualgraph_t, edge_weight_t>::type weight_map = get(edge_weight, g);
  property_map<dualgraph_t, edge_name_t>::type name_map = get(edge_name, g);
  property_map<dualgraph_t, edge_variables_t>::type vars_map = get(edge_variables_t(), g);
  
  for(size_t i = 0; i < csp.getNumVariables(); i++)
    {
    Variable &v = csp.getVariable(i);
    
    relset_t &rels = csp.getRelations(i);
    relset_t::const_iterator it = rels.begin(), pend = rels.end();
      
    while (it != pend)
      {
      relset_t::const_iterator it2 = it;
      ++it2;
      while(it2 != pend)
        {
        dualgraph_t::edge_descriptor e;
        bool inserted;

        tie(e,inserted) = boost::add_edge(relmap[*it],
                                          relmap[*it2],
                                           lexical_cast<std::string>(v.getId()), 
                                           g);
        if (inserted)
          {
          name_map[e] = v.getName();
          }
        else
          {
          name_map[e].append(" ").append(v.getName());
          }
        ++it2;
        }
      ++it;
      }
    }

  dualgraph_t::edge_iterator eit, eit_end;
  
  for(tie(eit,eit_end) = edges(g); eit != eit_end; ++eit)
    {
    Relation *r1 = vertexrel[source(*eit, g)];
    Relation *r2 = vertexrel[target(*eit, g)];

    vars_map[*eit] = intersection(r1->getVariables(), r2->getVariables());
    weight_map[*eit] = vars_map[*eit].size();
    }
  }

void createPrimalGraph(ConstraintNetwork &csp, primalgraph_t &g)
  {
  g.clear();

  // Maps variable to vertex
  typedef std::map<Variable *, primalgraph_t::vertex_descriptor, VariablePtrOrder>  varmap_t;
  varmap_t varmap;

  property_map<primalgraph_t, vertex_variable_t>::type vertexvar = get(vertex_variable_t(), g);
  
  // Insert vertices (variables)
  for(size_t i = 0; i < csp.getNumVariables(); i++)
    {
    Variable &v = csp.getVariable(i);
    primalgraph_t::vertex_descriptor vd = add_vertex(v.getName(), g);
    varmap[&v] = vd;
    vertexvar[vd] = &v;
    }
  
  // Insert an edge for each two pair of variables in a constraint
  ConstraintNetwork::relcol_t::iterator it = csp.getRelations().begin(), pend = csp.getRelations().end();
  while(it != pend)
    {
    colvector_t &cols = (*it)->getColumns();

    colvector_t::const_iterator it1 = cols.begin();
    while (it1 != cols.end())
      {
      colvector_t::const_iterator it2 = it1;
      ++it2;
      while (it2 != cols.end())
        {
        Variable &v1 = (*it1)->getVariable();
        Variable &v2 = (*it2)->getVariable();

        primalgraph_t::edge_descriptor e;

        bool inserted;

        tie(e,inserted) = boost::add_edge(varmap[&v1],
                                          varmap[&v2],
                                           lexical_cast<std::string>((*it)->getId()), 
                                           g);
        
        ++it2;
        }
      ++it1;
      }
    ++it;
    }
  }

void create_join_tree(dualgraph_t &jt)
  {
  typedef graph_traits<dualgraph_t>::edge_descriptor  Edge;
  typedef std::vector<Edge> EdgeSet;
  
  property_map<dualgraph_t, edge_weight_t>::type weight = get(edge_weight_t(), jt);
  dualgraph_t::edge_iterator eit,eit_end;

  for(tie(eit,eit_end) = edges(jt); eit != eit_end; ++eit)
    {
    weight[*eit] = 10000 - weight[*eit];
    }

  EdgeSet  spanning_tree;
  kruskal_minimum_spanning_tree(jt, std::back_inserter(spanning_tree));

  std::vector<bool>  marked(num_vertices(jt), false);
  EdgeSet::iterator it = spanning_tree.begin(), it_end = spanning_tree.end();
  graph_traits<dualgraph_t>::vertex_iterator viter, viter_end;
  bool error = false;

  for(tie(eit,eit_end) = edges(jt); eit != eit_end; ++eit)
    {
    weight[*eit] = 10000 - weight[*eit];
    }

  property_map<dualgraph_t, vertex_relation_t>::type vertexrel = get(vertex_relation_t(), jt);
  property_map<dualgraph_t, edge_name_t>::type name_map = get(edge_name, jt);

  graph_traits<dualgraph_t>::edge_iterator eiter, eiter_end, ei_tmp;
  for (tie(eiter, eiter_end) = edges(jt); eiter != eiter_end; ) 
    {
    Relation *r1 = vertexrel[source(*eiter, jt)];
    Relation *r2 = vertexrel[target(*eiter, jt)];
    ei_tmp = eiter++;
    if (std::find(spanning_tree.begin(), spanning_tree.end(), *ei_tmp) != spanning_tree.end())
      {
      }
    else
      {
      remove_edge(*ei_tmp, jt);
      }
    }
  if(error)
    INTERNAL_ERROR;
  }

bool is_simplicial(primalgraph_t::vertex_descriptor u, boolvector_t &eliminated, const primalgraph_t &G, bool show = false)
  {
  primalgraph_t::adjacency_iterator ait, ait_end;

  std::vector<size_t> V;

  for(tie(ait, ait_end) = adjacent_vertices(u, G); ait != ait_end; ++ait)
    {
    if (!eliminated[*ait])
      V.push_back(*ait);
    }

  if (V.empty())
    return true;

  for(size_t i = 0; i < V.size(); ++i)
    {
    for(size_t j = i+1; j < V.size(); ++j)
      {
      primalgraph_t::edge_descriptor e;
      bool found;
      tie(e,found) = edge(V[j], V[i], G);
      if (!found)
        return false;
      }
    }

  return true;
  }

void check_chordality(const primalgraph_t &G)
  {
  primalgraph_t::vertex_iterator it, it_end;

  std::cout << "Checking chordality...\n";
  boolvector_t  eliminated(num_vertices(G), false);

  while(std::count(eliminated.begin(), eliminated.end(), false) > 0)
    {
    for(tie(it,it_end) = vertices(G); it != it_end; ++it)
      {
      if (eliminated[*it])
        continue;

      if (is_simplicial(*it, eliminated, G))
        {
        eliminated[*it] = true;
        break;
        }
      }

    if (it == it_end)
      break;
    }

  if (std::count(eliminated.begin(), eliminated.end(), false) > 0)
    {
    std::cout << "Graph is not chordal !!\n";
    INTERNAL_ERROR;
    }
  }

typedef graph_traits<dualgraph_t>::vertex_descriptor Vertex;
typedef std::vector<Vertex> Path;

bool findpath(dualgraph_t::vertex_descriptor u, dualgraph_t::vertex_descriptor v, 
              std::vector<int> &pmap, Path &p, const dualgraph_t &G)
  {
  dualgraph_t::adjacency_iterator it, it_end;
  pmap[u] = 1;
  for(tie(it,it_end) = adjacent_vertices(u,G); it != it_end; ++it)
    {
    if (*it == v)
      {
      p.push_back(*it);
      return true;
      }
    }

  for(tie(it,it_end) = adjacent_vertices(u,G); it != it_end; ++it)
    {
    if (pmap[*it] == 0)
      {
      pmap[*it] = 1;
      if (findpath(*it,v,pmap,p,G))
        {
        pmap[*it] = 2;
        p.push_back(*it);
        return true;
        }
      pmap[*it] = 2;
      }
    }
  return false;
  }

void check_rip(const dualgraph_t &G)
  {
  graph_traits<dualgraph_t>::vertex_iterator it, viter, viter_end;
  property_map<dualgraph_t, vertex_relation_t>::const_type vertexrel = get(vertex_relation_t(), G);

  bool error = false;
  for (tie(viter, viter_end) = vertices(G); !error && viter != viter_end; ++viter) 
    {
    it = viter;
    ++it;
    for(;it != viter_end; ++it)
      {
      Relation *r1 = vertexrel[*viter];
      Relation *r2 = vertexrel[*it];

      varset_t intersect = intersection(r1->getVariables(), r2->getVariables());
      if (!intersect.empty())
        {
        Path path;
        std::vector<int>  pmap(num_vertices(G), 0);
        findpath(*viter, *it, pmap, path, G);
        if (path.empty())
          {
          std::cout << "No path from vertex " << *viter << "(" << *r1 << ") to vertex " << *it << "(" << *r2 << ")\n";
          error = true;
          }
        else 
          {
          for (size_t i = 0; i < path.size(); ++i)
            {
            Relation *r = vertexrel[path[i]];
            varset_t  pathint = intersection(r->getVariables(), intersect);
            if (pathint.size() != intersect.size())
              {
              std::cout << "Path from " << *viter << ":\n" << *r1 << " to " << *it << ":\n" << *r2 << " should contain:\n";
              write_set(std::cout, intersect) << "\nbut vertex " << path[i] << "  " << *r << " only contains:\n";
              write_set(std::cout, pathint) << "\nPath:";
              write_set(std::cout, path) << "\n";
              error = true;
              }
            }
          }
        }
      }
    }
  if (error)
    {
    std::cout << "Dual graph does not have the running intersection property !!\n";
    INTERNAL_ERROR;
    }
  }

std::ostream &operator<<(std::ostream &s, const dualgraph_t &g)
  {
  property_map<dualgraph_t, vertex_name_t>::const_type vertex_map = get(vertex_name, g);
  property_map<dualgraph_t, edge_name_t>::const_type edge_map = get(edge_name, g);
  write_graphviz(s, g,  default_writer(), default_writer());
  return s;
  }

std::ostream &operator<<(std::ostream &s, const primalgraph_t &g)
  {
  property_map<primalgraph_t, vertex_name_t>::const_type vertex_map = get(vertex_name, g);
  property_map<primalgraph_t, edge_name_t>::const_type edge_map = get(edge_name, g);
  write_graphviz(s, g,  default_writer(), default_writer());
  return s;
  }
