/*
 * Copyright (C) 2003 Jeppe Nejsum Madsen, nejsum@diku.dk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include "stdafx.h"
#include "csplib/relation.hpp"
#include "csplib/utility/formatter.hpp"
#include "csplib/uniquerows.hpp"

namespace
{
struct ColumnJoinOrder
  {
  bool operator()(const colpair_t &v1, const colpair_t &v2)
    {
    return v1.first->getVariable().getDomainSize() > v2.first->getVariable().getDomainSize();
    }
  };

// Constructs a joinindex for the common columns of two relations
void joinColumns(const colpairvector_t &columns, joinindex_t &ji, size_t numCommon, size_t maxrows) 
  {
  ji.clear();
  bool done = false;
  size_t  start = 0, start_row = 0;

  while(!done)
    {
    colpairvector_t::const_iterator it = columns.begin(), pend = columns.end();
    Formatter f("Joining columns");
    int n = 1;

    // For the first common column, build initial joinindex
    while (it != pend)
      {
      Column *c1 = it->first, *c2 = it->second;
      ++it;
      // Only join on columns which are in both relations
      if (c1 && c2)
        {
        done = c1->createJoinIndex(ji, *c2, start_row, ji.size() + 5000000);
        f.writeProgress(n++, numCommon, ji.size());
        break;
        }
      }

    // For the remaining common columns (if any), incrementally prune join
    // index
    //size_t new_start = ji.size();

    while(it != pend)
      {
      Column *c1 = it->first, *c2 = it->second;
      // Only join on columns which are in both relations
      if (c1 && c2)
        {
        c1->reviseJoinIndex(ji, *c2, start);
        f.writeProgress(n++, numCommon, ji.size());
        }
      ++it;
      }
    start = ji.size();
    if (start > maxrows)
      throw std::length_error("Join Columns");
    }
  }

// Create a vector containing pair of columns. first points to columns
// in r1, second points to columns in r2. If a variable is in both
// relations the pair will have two non-null fields. Returns the
// number of common columns
size_t unionColumns(const Relation *r1, const Relation *r2, colpairvector_t &columns) 
  {
  size_t numCommon = 0;
  colvector_t::const_iterator it1 = r1->getColumns().begin() , it2 = r2->getColumns().begin();
  colvector_t::const_iterator pend1 = r1->getColumns().end() , pend2 = r2->getColumns().end();

  while (it1 != pend1 && it2 != pend2)
    {
    size_t id1 = (*it1)->getVariable().getId();
    size_t id2 = (*it2)->getVariable().getId();
  
    if (id1 < id2)
      columns.push_back(std::make_pair(*it1++, (Column *)0));
    else if (id1 > id2)
      columns.push_back(std::make_pair((Column *)0, *it2++));
    else
      {
      columns.push_back(std::make_pair(*it1++, *it2++));
      numCommon++;
      }
    }

  while (it1 != pend1)
    columns.push_back(std::make_pair(*it1++, (Column *)0));

  while (it2 != pend2)
    columns.push_back(std::make_pair((Column *)0, *it2++));

  return numCommon;
  }


// Construct output relation based on the join index 
void constructRelation(Relation *newRel, const colpairvector_t &columns, const joinindex_t &ji, const varset_t &vars_to_remove)
  {
  colpairvector_t::const_iterator it;
  std::stringstream ss;
  ss << "Constructing relation " << newRel->getId();

  Formatter f(ss.str());
  int n = 1;
  for(it = columns.begin(); it != columns.end(); ++it)
    {
    colpair_t cp = *it;
    Column   *c;

    f.writeProgress(n++, columns.size());

    Variable &var = cp.first != 0 ?  cp.first->getVariable() : cp.second->getVariable();
    if (vars_to_remove.find(&var) != vars_to_remove.end())
      continue;
    
    if (cp.first == 0 || cp.second == 0)
      {
      // Column is not common, clone the column, picking only the
      // values specified by the join index
      if (cp.first != 0)
        {
        c = cp.first->cloneFromJoinIndex(ji, true);
        }
      else
        {
        c = cp.second->cloneFromJoinIndex(ji, false);
        }
      }
    else
      {
      // Column is common, create a new column based on the
      // intersection of values specified in the join index
      c = cp.first->createFromJoinIndex(ji, *cp.second);
      }
    newRel->addColumn(std::auto_ptr<Column>(c));
    }
  }
}

// Construct cartesian product output relation 
void constructCartesianProduct(Relation *newRel, const colpairvector_t &columns, size_t numNewRows, const varset_t &vars_to_remove)
  {
  colpairvector_t::const_iterator it;

  Formatter f("Constructing cartesian product");
  int n = 1;
  for(it = columns.begin(); it != columns.end(); ++it)
    {
    f.writeProgress(n++, columns.size(), numNewRows);
    colpair_t cp = *it;
    Column   *c;

    Variable &var = cp.first != 0 ?  cp.first->getVariable() : cp.second->getVariable();
    if (vars_to_remove.find(&var) != vars_to_remove.end())
      continue;

    if (cp.first != 0)
      {
      c = cp.first->clone(numNewRows, false);
      }
    else
      {
      c = cp.second->clone(numNewRows, true);
      }
    newRel->addColumn(std::auto_ptr<Column>(c));
    }
}

std::auto_ptr<Relation> Relation::joinCompress(const Relation &r, size_t new_id, size_t maxrows, const varset_t &vars_to_remove) const
  {
  varset_t intersect = intersection(getVariables(), r.getVariables());

  if (intersect.size() == 0)
    return join(r, new_id, maxrows, vars_to_remove);

  varset_t unionvars = getVariables() + r.getVariables();

  unionvars = unionvars - vars_to_remove;

  std::auto_ptr<Relation> split1, split2, newrel;
  std::auto_ptr<Relation> proj1, proj2;
  std::auto_ptr<Variable> lv1, lv2;
  
  const Relation *r1, *r2;

  varset_t splitvars1 = getVariables() - intersect - vars_to_remove;
  if (splitvars1.size() > 0)
    {
    proj1  = project(intersect, 42);
    split1 = project(splitvars1, 43);
    lv1    = proj1->link(*split1, 1000000);
    proj1->compress();
    split1->compress();
    r1 = proj1.get();
    }
  else
    r1 = this;

  varset_t splitvars2 = r.getVariables() - intersect - vars_to_remove;
  if (splitvars2.size() > 0)
    {
    proj2  = r.project(intersect, 44);
    split2 = r.project(splitvars2, 45);
    lv2    = proj2->link(*split2, 1000001);
    proj2->compress();
    split2->compress();
    r2 = proj2.get();
    }
  else
    r2 = &r;

  newrel = r1->join(*r2, new_id, maxrows, vars_to_remove);
  newrel->compress();

  if (split1.get() != 0)
    {
    newrel = newrel->join(*split1, new_id, maxrows, split1->getVariables()-unionvars);
    newrel->compress();
    }
  if (split2.get() != 0)
    {
    newrel = newrel->join(*split2, new_id, maxrows, split2->getVariables()-unionvars);
    newrel->compress();
    }

  return newrel;
  }

// Join a relation with relation r, returning the result.
std::auto_ptr<Relation> Relation::join(const Relation &r, size_t new_id, size_t maxrows, const varset_t &vars_to_remove) const
  {
  std::auto_ptr<Relation> newRel(new Relation(new_id, "JOIN"));

  newRel->m_SourceIds = m_SourceIds;
  newRel->m_SourceIds.insert(newRel->m_SourceIds.end(), r.m_SourceIds.begin(), r.m_SourceIds.end());

  // Find the smallest relation
  const Relation *smallRel, *largeRel;
  if (getNumRows() > r.getNumRows())
    {
    smallRel = &r;
    largeRel = this;
    }
  else
    {
    smallRel = this;
    largeRel = &r;
    }

  colpairvector_t columns;
  size_t numCommon = unionColumns(smallRel, largeRel, columns);
  if (numCommon == 0)
    {
    // No common columns, create cartesian product
    size_t numNewRows = smallRel->getNumRows() * largeRel->getNumRows();
    if (numNewRows > maxrows)
      throw std::length_error("Cartesian product");

    constructCartesianProduct(newRel.get(), columns, numNewRows, vars_to_remove);
    }
  else
    {
    joinindex_t ji;
    joinColumns(columns, ji, numCommon, maxrows);
    constructRelation(newRel.get(), columns, ji, vars_to_remove);
    }

  return newRel;
  }

size_t Relation::joinSize(const Relation &r, size_t maxrows)  const
  {
  size_t  size = 0;
  const Relation *small, *large;

  if (getNumRows() < r.getNumRows())
    {
    small = this;
    large = &r;
    }
  else
    {
    small = &r;
    large = this;
    }

  colpairvector_t colpairs;
  size_t numCommon = unionColumns(small, large, colpairs);

  if (numCommon == 0)
    size = small->getNumRows() * large->getNumRows();
  else
    {
    for(size_t i = 0; i < small->getNumRows() && size < maxrows; ++i)
      {
      boolvector_t  mask(large->getNumRows(), true);
    
      colpairvector_t::const_iterator it = colpairs.begin(), pend = colpairs.end();
      while(it != pend)
        {
        Column *c1 = it->first, *c2 = it->second;
        if (c1 && c2)
          {
          c1->updateTupleMask(i, *c2, mask);
          }
        ++it;
        }
                                
      size += std::count(mask.begin(), mask.end(), true);
      }
    }
  if (size == 0)
    {
    std::cout << *small << *large;
    }

  return size;
  }
