/*
 * Copyright (C) 2003 Jeppe Nejsum Madsen, nejsum@diku.dk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include "stdafx.h"
#include <algorithm>
#include <functional>
#include <numeric>
#include <iomanip>
#include "relation.hpp"
#include "uniquerows.hpp"

Relation::Relation(size_t id, const std::string &expr)
    : m_Id(id),
      m_Expr(expr)
  {
  }

Relation::~Relation()
  {
  for_all(m_Columns, delete_element<Column>());
  }

std::auto_ptr<Variable> Relation::link(Relation &r, size_t new_id)
  {
  std::auto_ptr<Variable> lv;

  if (getNumRows() != r.getNumRows())
    throw std::logic_error("link: Relations does not have same row count");

  UniqueRows  ur1(getColumns());
  (*getColumns().begin())->fillInUniqueRows(ur1);
  
  UniqueRows  ur2(r.getColumns());
  (*r.getColumns().begin())->fillInUniqueRows(ur2);
  
  size_t numArgs = std::min(ur1.size(), ur2.size());

  // Only link relations if there is more than 1 unique argument in either
  // relation
  if (numArgs > 1)
    {
    lv.reset(new LinkVariable(new_id, "Link "+boost::lexical_cast<std::string>(new_id), numArgs));
    
    UncompressedColumn *uc1 = UncompressedColumn::createNew(*lv, getNumRows());
    UncompressedColumn *uc2 = UncompressedColumn::createNew(*lv, r.getNumRows());

    std::cout << lv->getName() << " size=" << numArgs << ", other=" << std::max(ur1.size(), ur2.size()) << "\n";

    UniqueRows::const_iterator it, pend;
    if (ur1.size() < ur2.size())
      {
      it = ur1.begin();
      pend = ur1.end();
      }
    else
      {
      it = ur2.begin();
      pend = ur2.end();
      }

    for(size_t i = 0; i < numArgs; i++)
      {
      size_t cell = it->first;
      uc1->setCell(cell, i);
      uc2->setCell(cell, i);
      idvector_t *idv = it->second;
      idvector_t::const_iterator it1 = idv->begin(), pend1 = idv->end();
      while(it1 != pend1)
        {
        cell = *it1++;
        uc1->setCell(cell, i);
        uc2->setCell(cell, i);
        }
      ++it;
      }
    addColumn(std::auto_ptr<Column>(uc1));
    r.addColumn(std::auto_ptr<Column>(uc2));
    }

  return lv;
  }

void Relation::assign(const Relation &rel)
  {
  for_all(m_Columns, delete_element<Column>());

  m_Columns.clear();
  m_Variables.clear();

  colvector_t::const_iterator it = rel.m_Columns.begin(), pend = rel.m_Columns.end();
  while(it != pend)
    {
    Column   *c = *it;
    addColumn(std::auto_ptr<Column>(c->clone()));
    ++it;
    }
  m_ValidRows = rel.m_ValidRows;
  }

void Relation::addColumn(std::auto_ptr<Column> col)
  {
  // Ensure all columns have the same number of rows
  if (!m_Columns.empty() && col->getNumCells() != getNumRows())
    throw std::logic_error("Columns don't have same rowcount");

  if (m_Columns.empty())
    {
    m_ValidRows = boolvector_t(col->getNumCells(), true);
    }
  
  col->setIndex(getNumColumns());
  m_Variables.insert(&col->getVariable());
  m_Columns.push_back(col.release());
  
  // Ensure all columns are inserted in variable id order
  assert(std::is_sorted(m_Columns.begin(), m_Columns.end(), ColumnPtrVariableOrder()));
  }

void Relation::addTautologyColumn(Variable &var)
  {
  size_t n = m_Columns.empty() ? 1 : getNumRows();

  CompressedColumn *col = CompressedColumn::createNew(var, n);

  idvector_t  domain_values;
  for(size_t i = 0; i < var.getDomainSize(); ++i)
    {
    if (var.isValueValid(i))
      domain_values.push_back(i);
    }

  for(size_t i = 0; i < n; ++i)
    {
    col->addCell(domain_values);
    }
  col->setIndex(getNumColumns());
  if (m_Columns.empty())
    {
    m_ValidRows = boolvector_t(col->getNumCells(), true);
    }

  m_Variables.insert(&col->getVariable());
  m_Columns.push_back(col);
  
  // Ensure all columns are inserted in variable id order
  std::sort(m_Columns.begin(), m_Columns.end(), ColumnPtrVariableOrder());
  }

Column &Relation::getColumn(const Variable &var)
  {
  colvector_t::iterator it = m_Columns.begin(), pend = m_Columns.end();
  while (it != pend)
    {
    if ((*it)->getVariable() == var)
      return *(*it);
    ++it;
    }
  throw std::logic_error("getColumn: searched column not found!");
  }

const Column &Relation::getColumn(const Variable &var) const
  {
  colvector_t::const_iterator it = m_Columns.begin(), pend = m_Columns.end();
  while (it != pend)
    {
    if ((*it)->getVariable() == var)
      return *(*it);
    ++it;
    }
  throw std::logic_error("getColumn: searched column not found!");
  }

double Relation::getNumExpandedTuples() const
  {
  std::vector<double>  row_counts(getNumRows(), 1.0);

  colvector_t::const_iterator it = m_Columns.begin(), pend = m_Columns.end();

  while(it != pend)
    {
    Column *c = *it;
    if (c->isCompressed())
      c->updateRowCounts(row_counts);
    ++it;
    }
  return std::accumulate(row_counts.begin(), row_counts.end(), 0.0);;
  }

double Relation::getSize() const
  {
  double size = 0;
  colvector_t::const_iterator it = m_Columns.begin(), pend = m_Columns.end();
  while(it != pend)
    {
    size += (*it)->getSize();
    ++it;
    }
  return size;
  }

bool Relation::identifyUnaryConstraints(varset_t *changed_variables)
  {
  colvector_t::const_iterator it = m_Columns.begin(), pend = m_Columns.end();
  bool boundsChanged = false;

  while(it != pend)
    {
    Column  *c = *it;
    Variable &var = c->getVariable();
    size_t domSize = var.getDomainSize();
    boolvector_t mask(domSize, false);
    c->createUniqueMask(getValidRows(), mask);

    bool first = true;
    for(size_t i = 0; i < domSize; ++i)
      {
      if (var.isValueValid(i) && !mask[i])
        {
        boundsChanged = true;
        var.setValueInvalid(i);
        if (changed_variables)
          changed_variables->insert(&var);
        }
      }
    ++it;
    }
  return boundsChanged;
  }


bool Relation::revise()
  {
  bool    changed = false;
  colvector_t::iterator it = getColumns().begin(), pend = getColumns().end();

  while(it != pend)
    {
    bool colchange = (*it)->revise(getValidRows());
    if (colchange)
      {
      changed = true;
      }
    it++;
    }

  return changed;
  }

bool Relation::updateValidRowMask()
  {
  bool    changed = false;
  colvector_t::iterator it = getColumns().begin(), pend = getColumns().end();

  while(it != pend)
    {
    changed |= (*it)->updateValidRowMask(getValidRows());
    it++;
    }

  return changed;
  }

size_t Relation::getNumCommonColumns(const Relation &r2) const
  {
  size_t numCommon = 0;
  colvector_t::const_iterator it1 = getColumns().begin() , it2 = r2.getColumns().begin();
  colvector_t::const_iterator pend1 = getColumns().end() , pend2 = r2.getColumns().end();

  while (it1 != pend1 && it2 != pend2)
    {
    size_t id1 = (*it1)->getVariable().getId();
    size_t id2 = (*it2)->getVariable().getId();
  
    if (id1 < id2)
      ++it1;
    else if (id1 > id2)
      ++it2;
    else
      {
      ++it1;
      ++it2;
      numCommon++;
      }
    }
  return numCommon;
  }

std::auto_ptr<Relation> Relation::split(colvector_t &remove_columns, size_t new_id) 
  {
  Relation *newRel = new Relation(new_id, "SPLIT");

  newRel->m_SourceIds   = m_SourceIds;

  colvector_t::iterator it = remove_columns.begin(), pend = remove_columns.end();
  while(it != pend)
    {
    newRel->addColumn(std::auto_ptr<Column>(*it));
    m_Variables.erase(&(*it)->getVariable());
    m_Columns.erase(std::find(m_Columns.begin(), m_Columns.end(), *it));
    ++it;
    }
  return std::auto_ptr<Relation>(newRel);
  }

std::auto_ptr<Relation> Relation::split(varset_t &remove_vars, size_t new_id) 
  {
  Relation *newRel = new Relation(new_id, "SPLIT");

  newRel->m_SourceIds   = m_SourceIds;

  varset_t::iterator it = remove_vars.begin(), pend = remove_vars.end();
  while(it != pend)
    {
    Column *c = &getColumn(*(*it));

    newRel->addColumn(std::auto_ptr<Column>(c));
    m_Variables.erase(*it);
    m_Columns.erase(std::find(m_Columns.begin(), m_Columns.end(), c));
    ++it;
    }
  return std::auto_ptr<Relation>(newRel);
  }

std::auto_ptr<Relation> Relation::project(const varset_t &vars, size_t new_id)  const
  {
  Relation *newRel = new Relation(new_id, "PROJ");
  
  colvector_t::const_iterator it = m_Columns.begin(), pend = m_Columns.end();
  while(it != pend)
    {
    Column   *c = *it;
    Variable &v = c->getVariable();

    if (vars.find(&v) != vars.end())
      {
      newRel->addColumn(std::auto_ptr<Column>(c->clone()));
      }
    ++it;
    }
  newRel->m_ValidRows = this->m_ValidRows;

  return std::auto_ptr<Relation>(newRel);
  }

bool Relation::isCompressed() const
  {
  bool compressed = false;
  colvector_t::const_iterator it = m_Columns.begin(), pend = m_Columns.end();

  for(;it != pend; ++it)
    compressed |= (*it)->isCompressed();

  return compressed;
  }

bool Relation::pruneInvalidTuples()
  {
  size_t      old_row_count = getNumRows();
  urvector_t  ur;
  size_t      numValid = 0;
  bool        pruned = false;

  for(size_t i = 0; i < getNumRows(); ++i)
    {
    if (m_ValidRows[i])
      {
      ur.push_back(std::make_pair(i, (idvector_t*)0));
      numValid++;
      }
    }

  if (numValid < getNumRows())
    {
    pruned = true;
    colvector_t::iterator it = m_Columns.begin(), pend = m_Columns.end();
    while(it != pend)
      {
      (*it)->pruneCells(ur.begin(), ur.end());
      it++;
      }
    rowCountChanged();
    std::cout << "Pruned relation " << m_Id << ": old="<< old_row_count << ", new=" << numValid << "\n";
    }
  return pruned;
  }

bool Relation::removeTautologyColumns(varset_t &tautology_vars)
  {
  colvector_t::iterator it = m_Columns.begin();

  bool first = true;

  while(it != m_Columns.end())
    {
    colvector_t column;
    column.push_back(*it);
    UniqueRows  ur(column);
    (*it)->fillInUniqueRows(ur);
  
    if(ur.size() == 1)
      {
      if (first)
        {
        identifyUnaryConstraints();
        first = false;
        }

      Variable &var = (*it)->getVariable();
      tautology_vars.insert(&var);
      m_Variables.erase(&var);
      delete *it;      
      it = m_Columns.erase(it);
      }
    else
      it++;
    }

  if (tautology_vars.size() > 0)
    {
    std::cout << "Removed " << tautology_vars.size() << " tautology column(s): " << (*tautology_vars.begin())->getName() << "\n";
    }
  return tautology_vars.size() > 0;
  }

void Relation::checkInvariants() const
  {
  bool ok = true;
  if (!std::is_sorted(m_Columns.begin(), m_Columns.end(), ColumnPtrVariableOrder()))
    {
    ok = false;
    std::cerr << "Columns are not sorted\n";
    colvector_t::const_iterator it,it_end;
    for(it = m_Columns.begin(), it_end = m_Columns.end(); it != it_end; ++it)
      {
      std::cerr << (*it)->getVariable().getId() << "\n";
      }
    }

  if (m_ValidRows.size() != getNumRows())
    {
    ok = false;
    std::cerr << "m_ValidRows.size(" << m_ValidRows.size() << ") != getNumRows(" << getNumRows() << ")\n";
    }

  if (!ok)
    throw std::logic_error("Invariants do not hold");
  }

void Relation::rowCountChanged()
  {
  m_ValidRows = boolvector_t(getNumRows(), true);
  }


bool Relation::removeRedundancy()
  {
  colvector_t::iterator it, pend = m_Columns.end();
  bool                  changed = false;

  for(size_t row = 0; row < getNumRows(); ++row)
    {
    boolvector_t  contained(getValidRows());

    for(it = m_Columns.begin(); it != pend; ++it)
      {
      Column      *c = *it;
      c->updateContainmentMask(row, *c, contained);
      }

    for (size_t row2 = row+1; row2 < getNumRows(); ++row2)
      {
      if (contained[row2])
        {
        changed = true;
        setRowValid(row2, false);
        }
      }
    }

  if (changed)
    pruneInvalidTuples();

  return changed;
  }

std::string Relation::getRowAsString(size_t row) const
  {
  std::stringstream s;
  
  s << "[" << row << "]=";
  for(size_t col = 0; col < getNumColumns(); col++)
    {
    s << getColumns()[col]->cellToString(row) << "\t";
    }
  return s.str();
  }

std::ostream &operator<<(std::ostream &s, const Relation &r) 
  {
  size_t  numRows = r.getColumns()[0]->getNumCells();

  s << "Rel #" << r.getId() << "(" << r.getNumRows() << "/" << r.getNumExpandedTuples() << ")" << " ";

  for(size_t col = 0; col < r.getNumColumns(); col++)
    {
    s << r.getColumns()[col]->getVariable().getName();
    if (r.getColumns()[col]->isCompressed())
      s << "+ ";
    else
      s << " ";
    }

  return s;
  }

std::ostream &rel_tuples(std::ostream &s, const Relation &r) 
  {
  size_t  numRows = r.getColumns()[0]->getNumCells();

  std::vector<size_t> colwidths(r.getNumColumns(), 0);
  // Find column widths
  for(size_t col = 0; col < r.getNumColumns(); col++)
    {
    size_t len = r.getColumns()[col]->getVariable().getName().length();
    colwidths[col] = len;
    for(size_t i = 0; i < numRows; ++i)
      {
      len = r.getColumns()[col]->cellToString(i).length();
      if (len > colwidths[col])
        colwidths[col] = len;
      }       
    }
  std::cout << "      ";
  for(size_t col = 0; col < r.getNumColumns(); col++)
    {
    s << std::left << std::setw(colwidths[col])  << r.getColumns()[col]->getVariable().getName();
    if (r.getColumns()[col]->isCompressed())
      s << "+";
    else
      s << " ";
    }

  s << "\n";
  for(size_t i = 0; i < numRows; ++i)
    {
    std::cout << (r.isRowValid(i) ? " " : "-");
    std::cout << std::setw(4) << i << " ";
    for(size_t col = 0; col < r.getNumColumns(); col++)
      {
      s << std::setw(colwidths[col]) << r.getColumns()[col]->cellToString(i) << " ";
      }
    s << "\n";
    }       
  return s;
  }

std::ostream &operator<<(std::ostream &s, const Relation *r) 
  {
  return s << *r;
  }
