/*
 * Copyright (C) 2003 Jeppe Nejsum Madsen, nejsum@diku.dk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#ifndef RELATION_H
#define RELATION_H

#include <map>
#include "variable.hpp"
#include "column.hpp"

class Relation : public boost::noncopyable 
  {
  public:
    Relation(size_t id, const std::string &expr);
    ~Relation();

    inline bool operator==(const Relation &rhs) const
      {
      return m_Id == rhs.m_Id;
      }

     inline bool operator!=(const Relation &rhs) const
      {
      return !(*this == rhs);
      }

    inline bool operator<(const Relation &rhs) const
      {
      return m_Id < rhs.m_Id;
      }

    void               addColumn(std::auto_ptr<Column> col);
    void               addTautologyColumn(Variable &var);
    size_t             getId()                    const {return m_Id;}
    const idvector_t  &getSourceIds()             const {return m_SourceIds;}
    const std::string &getExpression()            const {return m_Expr;}

    size_t             getNumColumns()            const {return m_Columns.size();}
    const colvector_t &getColumns()               const {return m_Columns;}
    colvector_t       &getColumns()                     {return m_Columns;}
    const varset_t    &getVariables()             const {return m_Variables;}
    varset_t          &getVariables()                   {return m_Variables;}

  
    size_t             getNumRows()               const {return m_Columns[0]->getNumCells();}

    
    // Return number of tuples we get if uncompress is called
    double             getNumExpandedTuples()     const;
    
    // Calculate the number of scalar values contained in the
    // relation. For an uncompressed relation, this is numRows*numColumns
    double             getSize()                  const;

    Column            &getColumn(const Variable &var);
    const Column      &getColumn(const Variable &var) const;
    const Variable    &getVariable(size_t column_no)  const {return m_Columns[column_no]->getVariable();}

    size_t             getNumCommonColumns(const Relation &r) const;

    bool               isRowValid(size_t row)     const     {return m_ValidRows[row];}
    void               setRowValid(size_t row, bool valid)  {m_ValidRows[row] = valid;}
    boolvector_t      &getValidRows()                       {return m_ValidRows;}

    size_t             getNumValidRows()          const     {return std::count(m_ValidRows.begin(), m_ValidRows.end(), true);}
    std::string        getRowAsString(size_t row) const;


    // Operations on relation

    // Remove all tuples marked invalid
    bool               pruneInvalidTuples();
    bool               normalize();
    bool               revise();
    bool               updateValidRowMask();
    bool               compress(const varset_t *vars_to_compress = 0);
    bool               uncompress(bool uncompress_link = true);
    bool               isCompressed() const;

    // Return number of rows in join index 
    size_t             joinSize(const Relation &r, size_t maxrows) const;

    std::auto_ptr<Variable> link(Relation &r, size_t new_id = 0);

    // Join the two relations
    std::auto_ptr<Relation> join(const Relation &r, size_t new_id, size_t maxrows, const varset_t &vars_to_remove = varset_t()) const;
    std::auto_ptr<Relation> joinCompress(const Relation &r, size_t new_id, size_t maxrows, const varset_t &vars_to_remove = varset_t()) const;

    // Split the relation, remove the indicated columns
    std::auto_ptr<Relation> split(colvector_t &remove_columns, size_t new_id);
    std::auto_ptr<Relation> split(varset_t &remove_vars, size_t new_id);

    // Create a new relation with a copy of the projected columns
    std::auto_ptr<Relation> project(const varset_t &vars, size_t new_id) const;

    void assign(const Relation &rel);


    bool                identifyUnaryConstraints(varset_t *updated_variables = 0);
    bool                removeTautologyColumns(varset_t &tautology_vars);
    bool                removeRedundancy();
    void                checkInvariants() const; 

  protected:
    void                rowCountChanged();
    bool                m_Changed;
    size_t              m_Id;
    std::string         m_Expr;
    colvector_t         m_Columns;
    varset_t            m_Variables;
    idvector_t          m_SourceIds;
    boolvector_t        m_ValidRows;
};

struct RelationPtrScopeOrder
  {
  inline bool operator()(const Relation *r1, const Relation *r2)
    {
    return std::lexicographical_compare(r1->getColumns().begin(),
                                        r1->getColumns().end(),
                                        r2->getColumns().begin(),
                                        r2->getColumns().end(),ColumnPtrVariableOrder());
    }
  };

struct RelationPtrScopeEquality
  {
  // Precondition: r1 must not be contain more columns than r2
  inline bool operator()(const Relation *r1, const Relation *r2)
    {
    return std::equal(r1->getColumns().begin(),
                      r1->getColumns().end(),
                      r2->getColumns().begin(),
                      ColumnPtrVariableEquality());
    }
  };

struct RelationPtrScopeSubset
  {
  inline bool operator()(const Relation *r1, const Relation *r2)
    {
    return std::includes(r1->getColumns().begin(),
                        r1->getColumns().end(),
                        r2->getColumns().begin(),
                        r2->getColumns().end(),
                        ColumnPtrVariableOrder());
    }
  };

struct RelationPtrOrder
  {
  inline bool operator()(const Relation *r1, const Relation *r2) const
    {
    return *r1 < *r2;
    }
  };

typedef std::set<Relation *, RelationPtrOrder> relset_t;

std::ostream &operator<<(std::ostream &s, const Relation &v);
std::ostream &operator<<(std::ostream &s, const Relation *v);

std::ostream &rel_tuples(std::ostream &s, const Relation &v);

#endif
