/*
 * Copyright (C) 2003 Jeppe Nejsum Madsen, nejsum@diku.dk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#ifndef UNIQUEROWS_H
#define UNIQUEROWS_H

#include <hash_map>
#include <hash_set>
#include "utility/hashfunction.hpp"

namespace detail
  {
  // EqualityComparable for column vectors
  // Compares two rows in a column vector for equality
  // The column specified in colMask is not used in the comparison
  class ColumnComplementEq
    {
    const colvector_t &columns;
    colvector_t::const_iterator colMask;
    
    public:
      ColumnComplementEq(const colvector_t &columns, colvector_t::const_iterator colMask) 
        : columns(columns), colMask(colMask) {}

      inline bool operator()(const size_t &row1, const size_t &row2) const
        {
        if (row1 == row2) 
          return true;

        colvector_t::const_iterator it = columns.begin();
        colvector_t::const_iterator pend = columns.end();
        while(it != pend)
          {
          if (it != colMask)
            {
            if (!(*it)->cellsAreEqual(row1, row2))
              {
              return false;
              }
            }
          ++it;
          }
        return true;
        }
    };

  class ColumnRowEq
    {
    const colvector_t &columns;
    
    public:
      ColumnRowEq(const colvector_t &columns)
        : columns(columns) {}

      inline bool operator()(const size_t &row1, const size_t &row2) const
        {
        if (row1 == row2) 
          return true;

        colvector_t::const_iterator it = columns.begin();
        colvector_t::const_iterator pend = columns.end();
        while(it != pend)
          {
          if (!(*it)->cellsAreEqual(row1, row2))
            {
            return false;
            }
          ++it;
          }
        return true;
        }
    };

  // EqualityComparable that always returns true
  struct ColumnComplementTrueEq
    {
    ColumnComplementTrueEq(const colvector_t &, colvector_t::const_iterator) 
      {}

    inline bool operator()(const size_t, const size_t) const
      {
      return true;
      }
    };
  };

// UniqueComplement is a hashmap that, when constructed,
// contains the indicies of the unique rows in the column
// vector without colMask as the key of the map. The data in the map
// is the indicies in colMask that compares equal on the remaining columns
class UniqueComplement : boost::noncopyable,
                   public std::hash_map<size_t, idvector_t *,
                   Hasher, detail::ColumnComplementEq>
  {
  typedef std::hash_map<size_t, idvector_t *, Hasher, detail::ColumnComplementEq> base_t;
  HashValues  values;

  public:
    boost::object_pool<idvector_t>  objectPool;
    UniqueComplement(const colvector_t &columns,
               const colvector_t::const_iterator colMask);
  };

class UniqueRows: boost::noncopyable,
                  public std::hash_map<size_t, idvector_t *,
                  Hasher, detail::ColumnRowEq>
  {
  typedef std::hash_map<size_t, idvector_t *, Hasher, detail::ColumnRowEq> base_t;
  HashValues  values;

  public:
    boost::object_pool<idvector_t>  objectPool;
    UniqueRows(const colvector_t &columns);
  };

typedef detail::ColumnComplementTrueEq CEQ;
//typedef detail::ColumnComplementEq CEQ;

// UniqueRows is a hashset that, when constructed,
// contains the number of unique rows in the column
// vector without colMask
// If CEQ is ColumnRowMaskEq, the size is correct
// If CEQ is ColumnRowMaskTrueEq the size is an approximation, but it's faster

class UniqueRowCount : boost::noncopyable,
                       public std::hash_set<size_t, Hasher, CEQ>
  {
  typedef std::hash_set<size_t, Hasher, CEQ> base_t;

  public:
    UniqueRowCount(const colvector_t &columns,
                   const colvector_t::const_iterator colMask,
                   const HashValues &vals);
  };


std::ostream &operator<<(std::ostream &s, const UniqueRows &uc);

#endif
