/*
 * Copyright (C) 2003 Jeppe Nejsum Madsen, nejsum@diku.dk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#ifndef HASHFUNCTION_H
#define HASHFUNCTION_H

#include <boost/static_assert.hpp>
#include <boost/type_traits.hpp>
#include "../column.hpp"

// Tables of random values
template<int N>
class RandomTable
  {
  unsigned int tables[256][N];
  static const int NUM_VALUES = 256*N;

  public:
    RandomTable()
      {
      srand(0); // Seed with constant to get consistent results

      for(int i = 0; i < 256; i++)
        {
        for(int j = 0; j < N; j++)
          {
          tables[i][j] = rand();
          }
        }
      }

    // TODO: Check that % N for N=2^n is not using modulo
    inline size_t randomValue(size_t tableNo, size_t index)
      {
      return tables[index][tableNo % N];
      }

    // If max(index) < 256, 2^domainBits specifies max and
    // we view the N tables with 2^8 values as N*2^(8-domainbits) tables,
    // each with 2^domainBits values
    inline size_t randomValue(size_t tableNo, size_t index, unsigned char domainBits)
      {
      size_t offset = ((tableNo << domainBits) + index) % NUM_VALUES;
      return *(&tables[0][0]+offset);
      }
  };

extern RandomTable<8>  randomtable;

// Hash function to use for 1,2 and 4 bytes cells
inline size_t HashFunction(size_t index, unsigned char cell, unsigned char domainBits)
  {
  return randomtable.randomValue(index, cell, domainBits);
  }

inline size_t HashFunction(size_t index, unsigned short cell, unsigned char)
  {
  return randomtable.randomValue(index  , static_cast<unsigned char>(cell & 0xff)) ^ 
         randomtable.randomValue(index+1, static_cast<unsigned char>(cell >> 8));
  }

inline size_t HashFunction(size_t index, unsigned int cell, unsigned char)
  {
  return randomtable.randomValue(index  ,static_cast<unsigned char>(cell & 0xff)) ^ 
         randomtable.randomValue(index+1,static_cast<unsigned char>((cell >> 8) & 0xff)) ^ 
         randomtable.randomValue(index+2,static_cast<unsigned char>((cell >> 16) & 0xff)) ^ 
         randomtable.randomValue(index+3,static_cast<unsigned char>((cell >> 24) & 0xff));
  }

inline size_t HashFunction(size_t index, unsigned long cell, unsigned char x)
  {
  // assume sizeof(int) == sizeof(long)
  BOOST_STATIC_ASSERT(sizeof(int) == sizeof(long));
  return HashFunction(index, (unsigned int)cell, x);
  }

// The precomputed hashvalues, computed columnwise.
class HashValues : boost::noncopyable 
  {
  public:
    // Create vector hashvalues for all columns except colMask
    HashValues(const colvector_t &columns, colvector_t::const_iterator colMask);

    // Create vector hashvalues for all columns
    HashValues(const colvector_t &columns);

    // Create hashvalues for column
    HashValues(const Column &column);
    ~HashValues();

    // xor the vector hashvalues with the specified column
    void xorColumn(const Column &colMask);

    // Debug functions

    // Number of unique hashvalues
    size_t uniqueValues() const;
    // Dump hashtable to file
    void dump(const std::string &filename) const;

    friend class Hasher;
  private:
    size_t  *m_HashValues;
    size_t   m_TableSize;
    void     allocateTable(size_t tableSize);
  };

// HashFunction which uses precomputed hashvalues for each row
// in the column set. For use wiht stl algorithms
class Hasher
  {
  const HashValues &hv;
  public:
    Hasher(const HashValues &v) : hv(v) {}

    inline size_t operator()(const size_t &row) const
      {
      return hv.m_HashValues[row];
      }
  };

#endif
