/*
 * Copyright (C) 2003 Jeppe Nejsum Madsen, nejsum@diku.dk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#ifndef VARIABLE_H
#define VARIABLE_H

// Base class for variables
class Variable : boost::noncopyable
{
public:
  
  Variable(size_t id, const std::string &name, size_t domainSize = 0, bool link_var = false);
    virtual ~Variable() {}

    size_t              getId()                       const {return m_Id;}
    const std::string  &getName()                     const {return m_Name;}
    size_t              getDomainSize()               const {return m_DomainSize;}
    // Get # of bits needed to hold domain
    unsigned char       getDomainBits()               const;
    // Get # of valid domain values
    size_t              getValidDomainSize()          const;
    
    bool                isValueValid(size_t index)    const {return m_ValueValid[index];}
    void                setValueInvalid(size_t index)       {m_ValueValid[index] = false;}
    void                setValueValid(size_t index)         {m_ValueValid[index] = true;}
    bool                isLinkVar()                   const {return m_IsLinkVar;}
    //boolvector_t       &getValidList
    virtual void        removeInvalidValues();

    inline bool operator==(const Variable &rhs) const
      {
      return m_Id == rhs.m_Id;
      }

    inline bool operator<(const Variable &rhs) const
      {
      return m_Id < rhs.m_Id;
      }

  virtual std::string getDomainValueAsString(size_t index) const = 0;

  protected:
    boolvector_t              m_ValueValid;
    size_t                    m_Id;
    size_t                    m_DomainSize;
    mutable unsigned char     m_DomainBits;
    const std::string         m_Name;
    bool                      m_IsLinkVar;
  };

// Special variable type, which holds integers in the range 
// 0 - domainSize-1
class LinkVariable : public Variable
  {
  public:
    LinkVariable(size_t id, const std::string &name, size_t domainSize)
      : Variable(id, name, domainSize, true) {}

    virtual std::string getDomainValueAsString(size_t index) const
      {
      return boost::lexical_cast<std::string>(index);
      }
  };

// Typed variables, with user specified domain
template<typename T>
class ConcreteVariable : public Variable
  {
  public:
    typedef T value_type;

    ConcreteVariable(size_t id, const std::string &name)
      : Variable(id, name) {}

    void addDomainValue(value_type val)             
      {
      m_Domain.push_back(val);
      m_ValueValid.push_back(true);
      m_DomainSize++;
      }

    value_type  getDomainValue(size_t index)  const 
      {
      return m_Domain[index];
      }

    virtual std::string getDomainValueAsString(size_t index) const
      {
      return boost::lexical_cast<std::string>(m_Domain[index]);
      }

    virtual void removeInvalidValues()
      {
      std::vector<value_type>   new_domain(getValidDomainSize());
      typename std::vector<value_type>::iterator it = new_domain.begin();

      for (size_t i = 0; i < m_DomainSize; ++i)
        {
        if (isValueValid(i))
          *it++ = m_Domain[i];
        }
      std::swap(m_Domain, new_domain);

      Variable::removeInvalidValues();
      }

  private:
    std::vector<value_type>   m_Domain;
  };

// Ordering on Variable *, based on contents of ptr
struct VariablePtrOrder
  {
  inline bool operator()(const Variable *v1, const Variable *v2) const
    {
    return *v1 < *v2;
    }
  };

// Variable sets
typedef std::set<Variable *, VariablePtrOrder>  varset_t;

varset_t intersection(const varset_t &u, const varset_t &v);

// Set difference
varset_t operator-(const varset_t &u, const varset_t &v);

// Set union
varset_t operator+(const varset_t &u, const varset_t &v);

std::ostream &operator<<(std::ostream &s, const Variable &v);
std::ostream &operator<<(std::ostream &s, const Variable *v);
#endif
