/*
 * Copyright (C) 2003 Jeppe Nejsum Madsen, nejsum@diku.dk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include "stdafx.h"
#include "timer.h"
#include "csplib/constraint_network.hpp"
#include "csplib/adb/adbio.hpp"
#include "simulatorui.hpp"
#include "runtime_network.hpp"
#include "adbtest.hpp"

size_t NumRuns = 1;

void usage()
{
  std::cout << "usage: cspruntime [options] infile [infile2]\n"
    "  options:\n"
    "   -c  Collect statistics\n"
    "   -nn Set number of iterations to n\n"
    "   -t  Test infile against infile2\n"
    ;
  exit(EXIT_FAILURE);
}

void simulator(const char *infile)
  {
  RuntimeNetwork             instance;
  
  ADB::importADBFile(instance, infile);
  std::cout << "Starting simulation\n";
  SimulatorUI     ui(instance);
  ui.start();
  }

void collect_statistics(const char *infile, size_t num_simulations = 1)
  {
  RuntimeNetwork             rn;
  
  ADB::importADBFile(rn, infile);

  double max_time = 0.0;

  std::cout << "Starting benchmark simulation\n";
  for(size_t num = 0; num < num_simulations; ++num)
    {
    rn.reset();
  
    std::cout << "Run " << num << " of " << num_simulations << "\n";

    // Construct random ordering
    std::vector<size_t> variable_order;
    for(size_t i = 0; i < rn.getNumVariables(); ++i)
      {
      Variable &v = rn.getVariable(i);
      if (!v.isLinkVar())
        variable_order.push_back(v.getId());
      }
    std::random_shuffle(variable_order.begin(), variable_order.end());

    // Select value for each variable
    for(size_t i = 0; i < variable_order.size(); ++i)
      {
      Variable &v = rn.getVariable(variable_order[i]);

      // If only 1 domain val is valid, continue with next
      if (v.getValidDomainSize() > 1)
        {
        // Construct random ordering of valid domain values
        std::vector<size_t> value_order;
        for(size_t i = 0; i < v.getDomainSize(); ++i)
          {
          if (v.isValueValid(i))
            value_order.push_back(i);
          }
        std::random_shuffle(value_order.begin(), value_order.end());
        size_t value = value_order[0];
        std::cout << "SetVar " << v.getName() << "=" << v.getDomainValueAsString(value) << " ";
        HiresTimer t2;
        t2.start();
        if (!rn.add_constraint(v.getId(), value))
          INTERNAL_ERROR;
        t2.stop();
        if (t2.elapsedSecs() > max_time)
          max_time = t2.elapsedSecs();

        std::cout << " elapsed=" << t2.elapsedSecs() << "\n";
        }
      }
    }
  std::cout << "Max " << max_time << "\n";
  }

void test_files(const char *infile1, const char *infile2)
  {
  AdbTest at(infile1, infile2);
  at.test();
  }

int main(int , char* argv[])
{
  std::cout.setf(std::ios_base::unitbuf, std::ios_base::unitbuf);

  int res = EXIT_SUCCESS;
  argv++;
  bool collect, test ;
  collect = test = false;

  const char *infile1 = 0, *infile2 = 0, *arg;

  while(*argv)
    {
    arg = *argv++;

    if (*arg == '-')
      {
      arg++;
      switch(*arg)
        {
        case 'c':
          collect = true;
          break;
        case 't':
          test = true;
          break;
        case 'n':
          arg++;
          NumRuns = atoi(arg);
          break;
        default:
          usage();
        }
      }
    else
      {
      if (infile1 == 0)
        infile1 = arg;
      else if (infile2 == 0)
        infile2 = arg;
      else
        usage();
      }
    }

  if (infile1 == 0)
    usage();

  try
    {
    if (collect)
      {
      collect_statistics(infile1,NumRuns);
      }
    else if (test)
      {
      test_files(infile1, infile2);
      }
    else
      simulator(infile1);
    }
  catch(const std::exception &ex)
    {
    std::cout << "Exception: " << ex.what();
    std::cout << "\nType: " << typeid(ex).name() << "\n" << std::flush;
    res = EXIT_FAILURE;
    }

  catch(...)
    {
    std::cout << "Unknown exception caught\n";
    res = EXIT_FAILURE;
    }
  return res;
}

