/*
 * Copyright (C) 2003 Jeppe Nejsum Madsen, nejsum@diku.dk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include "stdafx.h"
#include "timer.h"
#include "boost/bind.hpp"
#include "boost/graph/undirected_dfs.hpp"
#include "csplib/detail/column_impl.hpp"
#include "csplib/graph.hpp"
#include "runtime_network.hpp"

using namespace boost;

RuntimeNetwork::RuntimeNetwork()
  {}

void RuntimeNetwork::reset()
  {
  relcol_t::iterator it, it2, it_end;

  if (num_vertices(jointree) == 0)
    {
    createDualGraph(*this, jointree);

    property_map<dualgraph_t, edge_weight_t>::type weight = get(edge_weight_t(), jointree);
    property_map<dualgraph_t, vertex_relation_t>::type rels = get(vertex_relation_t(), jointree);
    dualgraph_t::edge_iterator eit,eit_end;

    create_join_tree(jointree);
    check_rip(jointree);

    // See if we have any compressed relations
    compressed = false;
    for(it = getRelations().begin(); it != getRelations().end(); ++it)
      {
      Relation *r = *it;
      compressed |= r->isCompressed();
      }
    std::cout << "Compressed: " << compressed << "\n";

    // See if this is a special case sharing at most 1 var
    maxweight_is_1 = true;
    for(tie(eit,eit_end) = edges(jointree); eit != eit_end; ++eit)
      {
      // Weights are negative....
      if (abs(weight[*eit]) > 1)
        {
        maxweight_is_1 = false;
        std::cout << "Edge " << *eit << " has weight " << abs(weight[*eit]) << "\n";
        std::cout << *rels[source(*eit, jointree)] << "\n" << *rels[target(*eit, jointree)] << "\n";
        break;
        }
      }
    std::cout << "Uniform network: " << maxweight_is_1 << "\n";

    if (maxweight_is_1)
      {
      dualgraph_t dg;
      createDualGraph(*this, dg);

      if (is_cyclic(dg))
        {
        std::cout << "Dualgraph is cyclic, cannot simulate !!\n";
        INTERNAL_ERROR;
        }
      }

    arc_consistency();
    initial_last = last;

    update_valid_domains();

    for(it = getRelations().begin(); it != getRelations().end(); ++it)
      {
      Relation *r = *it;
      std::auto_ptr<Relation> rel = r->project(r->getVariables(), r->getId());
      initrels.push_back(rel.release());
      }

    tsv.resize(getNumVariables());
    for(size_t i = 0; i < getNumVariables(); i++)
      {
      Variable &v = getVariable(i);
      tsv[i] = boolvector_t(v.getDomainSize());
      for(size_t j = 0; j < v.getDomainSize(); ++j)
        {
        tsv[i][j] = v.isValueValid(j);
        }
      }
    }
  last = initial_last;
  H.clear();
  // Reinit valid domain values
  for(size_t i = 0; i < getNumVariables(); i++)
    {
    Variable &v = getVariable(i);
    boolvector_t &tsv_i = tsv[i];

    for(size_t j = 0; j < v.getDomainSize(); ++j)
      {
      if(tsv_i[j])
        v.setValueValid(j);
      else
        v.setValueInvalid(j);
      }
    }

  // Reinit relations
  it2 = initrels.begin();
  for(it = getRelations().begin(); it != getRelations().end(); ++it, ++it2)
    {
    Relation *rel = *it;
    Relation *initrel = *it2;
    rel->assign(*initrel);
    }
  }

bool RuntimeNetwork::add_constraint(size_t var_index, size_t domain_index)
  {
  bool res = false;

  if (revise_dual(var_index, domain_index))
    {
    H.push_back(std::make_pair(var_index, domain_index));
    update_valid_domains();
    res = true;
    }

  return res;
  }

bool RuntimeNetwork::revise_dual(size_t var_index, size_t domain_index)
  {
  relset_t &varrels = getRelations(var_index);
  Variable &var = getVariable(var_index);

  if (maxweight_is_1 || varrels.empty())
    {
    // No relations constrains this var, update domain directly
    for (size_t i = 0; i < var.getDomainSize(); ++i)
      {
      if (i == domain_index)
        var.setValueValid(i);
      else
        var.setValueInvalid(i);    
      }

    if (varrels.empty())
      return true;
    }

  Relation *rel = *varrels.begin();

  if (maxweight_is_1)
    {
    bool contract = false;

    relset_t changed_rels;

    relset_t::iterator it, it_end = varrels.end();
    for(it = varrels.begin(); !contract && it != it_end; ++it)
      {
      Relation *r = *it;
      if(revise(*r, *r))
        {
        if (r->getNumValidRows() == 0)
          {
          std::cout << "Contradiction in Revise-Dual\n";
          contract = true;
          }
        relset_t cr = getConnectedRelations(*r);
        changed_rels.insert(cr.begin(), cr.end());
        }
      }
    while(!contract && !changed_rels.empty())
      {
      Relation *r = *changed_rels.begin();
      changed_rels.erase(r);
      if (revise(*r,*r))
        {
        if (r->getNumValidRows() == 0)
          {
          std::cout << "Contradiction in Revise-Dual\n";
          contract = true;
          }
        relset_t cr = getConnectedRelations(*r);
        changed_rels.insert(cr.begin(), cr.end());
        }
      }
    return !contract;
    }
  else
    {
    Column   *col = &rel->getColumn(var);

    std::auto_ptr<UncompressedColumn> statecol(UncompressedColumn::createNew(var));
    statecol->addCell(domain_index);

    if (compressed)
      {
      Relation  r(0, "");

      r.addColumn(std::auto_ptr<Column>(statecol));
      std::auto_ptr<Relation> joinedrel = r.join(*rel, rel->getId(), 500000);
      rel->assign(*joinedrel);
      }
    else
      {
      col->findSupport(*statecol, rel->getValidRows());
      }
    }

  bool res;
  if (rel->getNumValidRows() == 0)
    {
    std::cout << "Contradiction in Revise-Dual\n";
    res = false;
    }
  else
    {
    res = arc_consistency(rel);
    }

  return res;
  }

void RuntimeNetwork::remove_constraint(size_t )
  {
  }

void RuntimeNetwork::update_valid_domains()
  {
  if (maxweight_is_1)
    // Domains are already updated
    return;

  varset_t vars(getVariables().begin(), getVariables().end());

  while(!vars.empty())
    {
    Variable *var = *vars.begin();
    relset_t &varrels = getRelations(var->getId());

    if (varrels.empty())
      vars.erase(var);
    else
      {
      Relation *rel = *varrels.begin();
      varset_t  tmp;

      tmp = vars - rel->getVariables();
      std::swap(vars, tmp);
      rel->identifyUnaryConstraints();
      }
    }
  }

bool RuntimeNetwork::revise(Relation &r, Relation &s)
  {
  // Handle special cases
  if (maxweight_is_1)
    {
    if (r.updateValidRowMask())
      return r.identifyUnaryConstraints();
    else
      return false;
    }

  // Normal, uncompressed, revise
  bool            changed = false;

  colpairvector_t ccols;
  common_columns(r,s,ccols);
  colpairvector_t::iterator it, it_end = ccols.end();

  std::vector<int> &support = last.last(r,s);

  // Do a column-wise arc-consistency
  std::vector<size_t> support_cols(r.getNumRows(),  ccols.size());
  size_t              num_finished = 0;
  
  while(num_finished != r.getNumRows())
    {
    for(it = ccols.begin(); num_finished != r.getNumRows() && it != it_end;)
      {
      Column *c1 = it->first;
      Column *c2 = it->second;
      ++it;
      changed |= c1->findInitialSupport(*c2, r.getValidRows(), s.getValidRows(), 
                                        support, support_cols, ccols.size());
      num_finished = std::count(support_cols.begin(), support_cols.end(), ccols.size());
      }
    }
  return changed;
  }

void RuntimeNetwork::propagate(relset_t Q)
  {
  property_map<dualgraph_t, vertex_relation_t>::type vertexrel = get(vertex_relation_t(), jointree);

  while(!Q.empty())
    {
    Relation *s = *Q.begin();
    Q.erase(Q.begin());

    dualgraph_t::adjacency_iterator it, it_end;
    for(tie(it,it_end) = adjacent_vertices(s->getId(), jointree); it != it_end; ++it)
      {
      Relation *r = vertexrel[*it];
      if (*s != *r && revise(*r, *s))
        {
        if (r->getNumValidRows() == 0)
          {
          std::cout << "Contradiction in AC\n";
          rel_tuples(std::cout, *r);

          throw contradiction_error(r,s);
          }
        Q.insert(r);
        }
      }
    }
  }

bool RuntimeNetwork::arc_consistency()
  {
  if (compressed && !maxweight_is_1)
    return arc_consistency(*getRelations().begin());

  bool res = true;
  try
    {
    relset_t Q;
    dualgraph_t::adjacency_iterator it, it_end;
    dualgraph_t::vertex_iterator vit, vit_end;
    property_map<dualgraph_t, vertex_relation_t>::type vertexrel = get(vertex_relation_t(), jointree);

    for(tie(vit,vit_end) = vertices(jointree); vit != vit_end; ++vit)
      {
      Relation *r = vertexrel[*vit];
      for(tie(it,it_end) = adjacent_vertices(*vit, jointree); it != it_end; ++it)
        {
        Relation *s = vertexrel[*it];
        if (*s != *r && revise(*r, *s))
          {
          if (r->getNumValidRows() == 0)
            {
            std::cout << "Contradiction in AC\n";
            rel_tuples(std::cout, *r);
            throw contradiction_error(r,s);
            }
          Q.insert(r);
          }
        }
      }
    propagate(Q);
    }
  catch(contradiction_error &)
    {
    res = false;
    }
  return res;
  }


bool RuntimeNetwork::arc_consistency(Relation *rel)
  {
  if (compressed && !maxweight_is_1)
    return directional_arc_consistency(jointree, 10000000, rel);

  bool res = true;
  try
    {
    relset_t Q;
    dualgraph_t::adjacency_iterator it, it_end;
    dualgraph_t::vertex_iterator vit, vit_end;
    property_map<dualgraph_t, vertex_relation_t>::type vertexrel = get(vertex_relation_t(), jointree);

    for(tie(vit,vit_end) = vertices(jointree); vit != vit_end; ++vit)
      {
      Relation *r = vertexrel[*vit];
      if (*r == *rel)
        {
        for(tie(it,it_end) = adjacent_vertices(*vit, jointree); it != it_end; ++it)
          {
          Relation *s = vertexrel[*it];
          if (*s != *r && revise(*s,*r))
            {
            if (s->getNumValidRows() == 0)
              {
              std::cout << "Contradiction in AC\n";
              throw contradiction_error(r,s);
              }
            Q.insert(s);
            }
          }
        }
      }
    propagate(Q);
    }
  catch(contradiction_error &)
    {
    res = false;
    }
  return res;
  }
